/*
 * Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of an abstract graph.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif	/* HAVE_CONFIG_H */

#include <math.h>
#include <regex.h>

#include <complib/cl_heap.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_GRAPH_C
#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_graph.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_unhealthy_ports.h>
#include <opensm/osm_helper.h>

#define	MAX_PLANE	IB_MAX_NUM_OF_PLANES_XDR
#define	ALL_PLANES	(((1 << MAX_PLANE) - 1) << 1)
#define	NO_PLANE	0


typedef struct {
	osm_ucast_mgr_t		*p_mgr;
	osm_graph_t		*p_graph;
	void			*p_data;
} graph_context_t;


static boolean_t		graph_debug = FALSE;

/* -------------------------------------------------------------------------------------------------
 * Debug section.
 */

#define v2_snprintf(format, ...)	bufused += snprintf(buffer+bufused, bufsize-bufused, format, ## __VA_ARGS__)

#define print_or_log(BUFFER) {									\
	if (graph_debug) {									\
		printf("%s", BUFFER);								\
	}											\
												\
	if (p_log != NULL) {									\
		OSM_LOG(p_log, OSM_LOG_DEBUG, "%s", BUFFER);					\
	}											\
												\
	bufused = 0;										\
}

__attribute__((unused))
static void graph_print(osm_graph_t *p_graph)
{
	size_t			bufsize;
	size_t			bufused;
	char			*buffer;
	osm_log_t		*p_log;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_remote_sw;
	osm_vertex_t		*p_vertex;
	osm_vertex_t		*p_remote_vertex;
	osm_edge_t		*p_edge;

	p_log = osm_log_is_active_v2(p_graph->p_log, OSM_LOG_DEBUG, FILE_ID) ? p_graph->p_log : NULL;
	if (!graph_debug && (p_log == NULL)) {
		return;
	}

	bufused = 0;
	bufsize = 65536;
	if ((buffer = malloc(bufsize)) == NULL) {
		return;
	}

	v2_snprintf("\n============================================================================================\n");
	v2_snprintf("GRAPH: Plane %d   Vertices %ld   Edges %ld\n",
		p_graph->plane,
		p_graph->num_vertices,
		p_graph->num_edges);
	print_or_log(buffer);

	foreach_graph_vertex(p_graph, p_vertex) {
		p_sw = osm_graph_vertex_to_switch(p_vertex);
		v2_snprintf("\n\t%3d   %-30s    %3ld edges    plane=%x   hcas=%d\n",
			p_vertex->index,
			osm_vertex_name(p_vertex),
			p_vertex->num_edges,
			p_sw->plane_mask,
			p_vertex->num_hcas);

		foreach_graph_vertex_edge(p_vertex, p_edge) {
			p_remote_vertex = osm_graph_index_to_vertex(p_graph, p_edge->remote_vertex_index);
			p_remote_sw = osm_graph_vertex_to_switch(p_remote_vertex);

			v2_snprintf("\t\t%4d : %3d -> (%4d,%3d)  %s  (W=%d)\n",
				p_edge->index,
				p_edge->port_num,
				p_edge->remote_vertex_index,
				p_edge->remote_port_num,
				osm_switch_name(p_remote_sw),
				p_edge->weight);
		}

		print_or_log(buffer);
	}

	free(buffer);
}

/* -------------------------------------------------------------------------------------------------
 * Utility section.
 */

static inline int graph_switch_to_plane_mask(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw)
{

	/*
	 * A plane mask of 0 on a planarized switch is an error.  Hence, we don't include it in any graph.
	 * (This is number 4 below.)
	 *
	 * The order of the tests is:
	 *	1) if this is NOT a planarized subnet, then return plane 0.
	 *	2) if this is a prism switch, then return ALL_PLANES.
	 *	3) if the plane_mask is non-zero, then return it.
	 *	4) if the plane_mask is 0 and it is a planarized system, return NO_PLANE.
	 *	5) if the plane_mask is 0 and it is NOT planarized system, return ALL_PLANES (legacy switch).
	 *
	 */
	if (!p_subn->is_planarized) {
		return 1;
	} else if (osm_switch_is_prism_switch(p_sw, NULL)) {
		return ALL_PLANES;
	} else if (p_sw->plane_mask != 0) {
		return p_sw->plane_mask;
	} else if (osm_switch_get_planarized_system(p_subn, p_sw)) {
		return NO_PLANE;
	} else {
		return ALL_PLANES;
	}
}

static inline uint8_t graph_switch_to_plane_number(IN osm_switch_t *p_sw)
{
	uint8_t			plane_number = -1;
	int			plane_mask = p_sw->plane_mask;

	/*
	 * Legacy switches have plane mask set to 0.
	 */
	if (plane_mask == 0) {
		return ALL_PLANES;
	}

	/*
	 * Prism switches have plane mask set to ALL_PLANES.
	 */
	if (plane_mask == ALL_PLANES) {
		return ALL_PLANES;
	}

	/*
	 * Planarized switches have plane mask with only 1 bit set.
	 */
	while (plane_mask != 0) {
		plane_mask >>= 1;
		plane_number++;
	}

	return plane_number;
}

static inline osm_graph_t *switch_to_graph(IN osm_ucast_mgr_t *p_mgr, IN osm_switch_t *p_sw)
{
	uint8_t			plane_number = graph_switch_to_plane_number(p_sw);

	return p_mgr->graphs[plane_number];
}

static inline boolean_t graph_link_is_in_plane(IN osm_switch_t *p_sw, IN uint32_t port_num, IN uint8_t graph_plane)
{
	uint8_t			remote_port_num = 0;
	osm_node_t		*p_remote_node = osm_node_get_remote_node(p_sw->p_node, port_num,
									  &remote_port_num);

	if (!p_remote_node)
		return FALSE;

	return  osm_planar_port_routable_in_plane(p_sw->p_node, port_num, graph_plane) &&
		osm_planar_port_routable_in_plane(p_remote_node, remote_port_num, graph_plane);
}

/* -------------------------------------------------------------------------------------------------
 * Weight section.
 */

static int set_edge_weight(void *context, uint64_t guid, char *port_weight)
{
	osm_node_t		*p_node;
	osm_graph_t		*p_graph;
	osm_vertex_t		*p_vertex;
	osm_edge_t		*p_edge;
	osm_switch_t		*p_sw;
	osm_ucast_mgr_t		*p_mgr;
	osm_subn_t		*p_subn;
	regex_t			*p_regex;
	regmatch_t		match[3];
	uint32_t		port;
	uint32_t		weight;
	uint32_t		graph_mask;
	uint32_t		plane_mask;
	graph_context_t		*p_context;

	p_context = (graph_context_t *)context;
	p_mgr = p_context->p_mgr;
	p_subn = p_mgr->p_subn;
	p_graph = p_context->p_graph;
	p_regex = p_context->p_data;
	graph_mask = 1 << p_graph->plane;

	/*
	 * Find the node for this guid.
	 */
	p_node = osm_get_node_by_guid(p_subn, cl_hton64(guid));
	if ((p_node == NULL) || ((p_sw = p_node->sw) == NULL)) {
		OSM_LOG(p_graph->p_log, OSM_LOG_DEBUG, "switch with guid 0x%016lx was not found\n", guid);
		return 0;
	}

	plane_mask = graph_switch_to_plane_mask(p_subn, p_sw);
	if ((graph_mask & plane_mask) == 0) {		// TODO: make a functions
		return 0;
	}

	/*
	 * Parse the rest of the line and verify the values.
	 */
	if (regexec(p_regex, port_weight, 3, match, 0) != 0) {
		OSM_LOG(p_graph->p_log, OSM_LOG_DEBUG, "invalid hop_weight line for guid 0x%016lx : [%s]\n", guid, port_weight);
		return 0;
	}

	port = strtoul(port_weight + match[1].rm_so, NULL, 0);
	weight = strtoul(port_weight + match[2].rm_so, NULL, 0);

	if (port >= p_sw->num_ports) {
		OSM_LOG(p_graph->p_log, OSM_LOG_DEBUG, "bad port specified for guid 0x%016lx\n", guid);
		return 0;
	}

	if (weight >= 0x100) {
		OSM_LOG(p_graph->p_log, OSM_LOG_DEBUG, "bad hop weight specified for guid 0x%016lx port %u\n", guid, port);
		return 0;
	}

	/*
	 * Set the weight.
	 */
	p_vertex = osm_graph_switch_to_vertex(p_graph, p_sw);
	if (p_vertex != NULL) {
		foreach_graph_vertex_edge(p_vertex, p_edge) {
			if (p_edge->port_num == port) {
				p_edge->weight = weight;

				if (weight > p_graph->max_weight) {
					p_graph->max_weight = weight;
				}

				break;
			}
		}
	}

	return 0;
}

void set_edge_hop_weights(osm_ucast_mgr_t *p_mgr, osm_graph_t *p_graph)
{
	regex_t			hop_weight_regex;
	osm_subn_t		*p_subn;
	graph_context_t		context;

	context.p_mgr = p_mgr;
	context.p_graph = p_graph;
	context.p_data = &hop_weight_regex;

	p_subn = p_mgr->p_subn;

	/*
	 * If no file, just return.
	 */
	if (p_subn->opt.hop_weights_file == NULL) {
		return;
	}

	/*
	 * Create the regex for parsing the file.
	 */
	if (regcomp(&hop_weight_regex, "[ \t]*([0-9]+)[ \t]+([0-9]+)", REG_EXTENDED) != 0) {
		OSM_LOG(p_graph->p_log, OSM_LOG_ERROR, "ERR BC04: no regex for hop weights file - assuming constant weights\n");
		return;
	}

	/*
	 * Parse the file.
	 */
	OSM_LOG(p_graph->p_log, OSM_LOG_DEBUG, "parsing hop_weights_file [%s]\n", p_subn->opt.hop_weights_file);

	if (parse_node_map(p_subn->opt.hop_weights_file, set_edge_weight, &context)) {
		OSM_LOG(p_graph->p_log, OSM_LOG_ERROR, "ERR BC05: can't parse hop_weights_file [%s]\n", p_subn->opt.hop_weights_file);
	}

	regfree(&hop_weight_regex);
}

/* -------------------------------------------------------------------------------------------------
 * Internal section.
 */

__attribute__((unused))
static inline int remote_type(osm_switch_t *p_sw, uint32_t port_num)
{
	uint8_t			remote_port_num = 0;
	osm_node_t		*p_node = p_sw->p_node;
	osm_node_t		*p_remote_node = osm_node_get_remote_node(p_node, port_num, &remote_port_num);

	if (p_remote_node == NULL) {
		return 0;
	}

	return osm_node_get_type(p_remote_node);
}

__attribute__((unused))
static inline boolean_t remote_type_is_hca(osm_switch_t *p_sw, uint32_t port_num)
{
	return (remote_type(p_sw, port_num) == IB_NODE_TYPE_CA);
}

__attribute__((unused))
static inline boolean_t remote_type_is_switch(osm_switch_t *p_sw, uint32_t port_num)
{
	return (remote_type(p_sw, port_num) == IB_NODE_TYPE_SWITCH);
}

__attribute__((unused))
static inline boolean_t remote_type_is_router(osm_switch_t *p_sw, uint32_t port_num)
{
	return (remote_type(p_sw, port_num) == IB_NODE_TYPE_ROUTER);
}

__attribute__((unused))
static inline osm_switch_t *remote_switch(osm_node_t *p_node, uint32_t port_num, uint32_t *p_remote_port_num)
{
	uint8_t			remote_port_num = 0;
	osm_node_t		*p_remote_node = osm_node_get_remote_node(p_node, port_num, &remote_port_num);
	osm_switch_t		*p_remote_sw = NULL;

	if ((p_remote_node != NULL) && (p_remote_node != p_node) && (p_remote_node->sw != NULL)) {
		p_remote_sw = p_remote_node->sw;
	}

	if (p_remote_port_num != NULL) {
		*p_remote_port_num = remote_port_num;
	}

	return p_remote_sw;
}

__attribute__((unused))
static boolean_t switch_is_isolated(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw)
{
	return lookup_held_back_switch(p_subn, p_sw->p_node);
}

__attribute__((unused))
static boolean_t port_is_connected(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw, IN uint32_t port_num)
{
	osm_node_t		*p_remote_node = osm_node_get_remote_node(p_sw->p_node, port_num, NULL);
	osm_physp_t		*p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);

	if ((p_remote_node == NULL) || (p_remote_node == p_sw->p_node)) {
		return FALSE;
	} else if ((p_remote_node->sw != NULL) ||
		   (osm_physp_get_special_port_type(p_physp->p_remote_physp) == IB_SPECIAL_PORT_TYPE_RESERVED)) {
		if (p_subn->is_planarized) {
			return osm_planar_port_connectivity_correct(p_sw->p_node, port_num);
		} else {
			return TRUE;
		}
	} else {
		return FALSE;
	}
}

__attribute__((unused))
static boolean_t port_is_isolated(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw, IN uint32_t port_num)
{
	if (lookup_isolated_port(&p_subn->p_osm->sm, p_sw->p_node, (uint8_t)port_num)) {
		return TRUE;
	}

	return FALSE;
}

__attribute__((unused))
static boolean_t port_is_fnm(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw, IN uint32_t port_num)
{
	osm_physp_t		*p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);

	if (!p_subn->is_planarized) {
		return FALSE;
	} else if (p_physp == NULL) {
		return FALSE;
	} else {
		return osm_physp_is_fnm(p_subn, p_physp);
	}
}

__attribute__((unused))
static boolean_t port_is_asymmetric(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw, IN uint32_t port_num)
{
	osm_physp_t		*p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);

	if (!p_subn->is_planarized) {
		return FALSE;
	} else if (p_physp == NULL) {
		return FALSE;
	} else {
		return osm_planar_is_link_asymmetric(p_subn, p_physp, NULL);
	}
}


__attribute__((unused))
static boolean_t port_is_active(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw, IN uint32_t port_num)
{
	return   port_is_connected(p_subn, p_sw, port_num) && 
		!port_is_isolated(p_subn, p_sw, port_num) &&
		!port_is_fnm(p_subn, p_sw, port_num) &&
		!port_is_asymmetric(p_subn, p_sw, port_num);
}

__attribute__((unused))
static boolean_t port_is_not_active(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw, IN uint32_t port_num)
{
	return !port_is_active(p_subn, p_sw, port_num);
}

__attribute__((unused))
static boolean_t port_connects_to_switch(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw, IN uint32_t port_num, IN uint8_t plane)
{
	uint32_t		remote_port_num;
	osm_switch_t		*p_remote_sw;
	uint32_t		plane_mask, graph_mask;

	p_remote_sw = remote_switch(p_sw->p_node, port_num, &remote_port_num);

	if (p_remote_sw == NULL) {
		return FALSE;
	} else if (switch_is_isolated(p_subn, p_remote_sw)) {
		return FALSE;
	}

	/*
	 * Check that the remote switch is in the plane.
	 */
	graph_mask = 1 << plane;
	plane_mask = graph_switch_to_plane_mask(p_subn, p_remote_sw);

	if ((graph_mask & plane_mask) == 0) {
		return FALSE;
	}

	/*
	 * In order to have this port connect to a switch, the following must be true:
	 *	1) the switch port must be in the plane
	 *	2) the remote switch port must be in the plane
	 *	3) the port must be active
	 *	4) the remote node type must be a switch
	 */
	return  graph_link_is_in_plane(p_sw, port_num, plane) &&
		port_is_active(p_subn, p_sw, port_num) &&
		remote_type_is_switch(p_sw, port_num);
}

__attribute__((unused))
static boolean_t port_connects_to_hca(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw, IN uint32_t port_num, IN uint8_t plane)
{
	return  graph_link_is_in_plane(p_sw, port_num, plane) &&
		port_is_active(p_subn, p_sw, port_num) &&
		remote_type_is_hca(p_sw, port_num);
}

/* -------------------------------------------------------------------------------------------------
 * API section.
 */

void osm_graph_destroy(IN osm_graph_t *p_graph)
{
	if (p_graph != NULL) {
		free(p_graph->vertices);
		free(p_graph->edges);
		free(p_graph);
	}
}

osm_graph_t *osm_graph_create(IN osm_ucast_mgr_t *p_mgr, IN uint8_t plane)
{
	osm_log_t		*p_log;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_remote_sw;
	osm_vertex_t		*p_vertex;
	osm_vertex_t		*p_remote_vertex;
	osm_edge_t		*p_edge;
	osm_subn_t		*p_subn;
	osm_graph_t		*p_graph;
	osm_graph_t		*ret;
	osm_vertex_t		*p_vertices;
	osm_edge_t		*p_edges;
	osm_switch_t		**p_sw_list;

	uint32_t		port_num;
	uint32_t		remote_port_num;
	uint32_t		graph_mask;
	uint32_t		plane_mask;
	uint32_t		num_vertices;
	uint32_t		num_edges;
	uint32_t		vertex_edges;
	uint32_t		v_index;

	OSM_LOG_ENTER(p_mgr->p_log);

	p_subn = p_mgr->p_subn;
	p_log = &p_subn->p_osm->log;
	ret = NULL;

	graph_mask = 1 << plane;

	/*
	 * Count the number of vertices and edges.
	 */
	if (cl_qmap_count(&p_subn->sw_guid_tbl) == 0) {
		OSM_LOG(p_log, OSM_LOG_DEBUG, "topology has no switches\n");
		goto Exit;
	}

	p_sw_list = calloc(cl_qmap_count(&p_subn->sw_guid_tbl), sizeof(osm_switch_t *));
	if (p_sw_list == NULL) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BC02: graph allocation failed\n");
		goto Exit;
	}

	num_vertices = 0;
	num_edges = 0;

	foreach_switch(p_subn, p_sw) {
		p_sw->graph_index[plane] = -1;
		if (switch_is_isolated(p_subn, p_sw)) {
			continue;
		}

		plane_mask = graph_switch_to_plane_mask(p_subn, p_sw);
		if ((graph_mask & plane_mask) != 0) {
			vertex_edges = 0;
			for (port_num = 1; port_num < p_sw->num_ports; port_num++) {
				if (port_connects_to_switch(p_subn, p_sw, port_num, plane)) {
					vertex_edges++;
					num_edges++;
				}
			}

			if (vertex_edges != 0) {
				p_sw_list[num_vertices++] = p_sw;
			} else {
				OSM_LOG(p_log, OSM_LOG_DEBUG, "Switch 0x%016lx has no edges\n", // TODO: maybe INFO?
					osm_node_get_node_guid_ho(p_sw->p_node));
			}
		}

	}

	if (num_vertices == 0) {
		OSM_LOG(p_log, OSM_LOG_DEBUG, "graph plane %u has no switches\n", plane);
		free(p_sw_list);
		goto Exit;
	}

	/*
	 * Allocate the memory for the graph.
	 */
	p_graph     = calloc(1, sizeof(osm_graph_t));
	p_vertices  = calloc(num_vertices, sizeof(osm_vertex_t));
	p_edges     = calloc(num_edges, sizeof(osm_edge_t));

	if ((p_graph == NULL) || (p_vertices == NULL) || (p_edges == NULL)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BC01: graph allocation failed\n");
		free(p_graph);
		free(p_vertices);
		free(p_edges);
		free(p_sw_list);

		goto Exit;
	}

	p_graph->p_subn = p_subn;
	p_graph->p_log = &p_subn->p_osm->log;
	p_graph->plane = plane;
	p_graph->num_vertices = num_vertices;
	p_graph->num_edges = num_edges;
	p_graph->max_weight = 1;
	p_graph->vertices = p_vertices;
	p_graph->edges = p_edges;

	/*
	 * Fill in the graph.
	 */
	for (v_index = 0; v_index < num_vertices; v_index++) {
		p_sw = p_sw_list[v_index];

		p_sw->graph_index[plane] = v_index;
		p_sw->re_priv = NULL;

		p_vertex = osm_graph_index_to_vertex(p_graph, v_index);
		p_vertex->index = v_index;
		p_vertex->p_sw = p_sw;
	}

	free(p_sw_list);

	p_edge = p_graph->edges;

	foreach_graph_vertex(p_graph, p_vertex) {
		p_sw = osm_graph_vertex_to_switch(p_vertex);
		p_vertex->edges = p_edge;
		p_vertex->num_edges = 0;

		for (port_num = 1; port_num < p_sw->num_ports; port_num++) {
			if (port_connects_to_switch(p_subn, p_sw, port_num, plane)) {
				p_remote_sw = remote_switch(p_sw->p_node, port_num, &remote_port_num);
				if (p_remote_sw == NULL) {
					continue;
				}

				if (p_remote_sw->graph_index[plane] < 0) {
					continue;
				}

				p_remote_vertex = osm_graph_switch_to_vertex(p_graph, p_remote_sw);

				p_edge->index = p_edge - p_graph->edges;
				p_edge->vertex_index = p_vertex->index;
				p_edge->port_num = port_num;
				p_edge->remote_vertex_index = p_remote_vertex->index;
				p_edge->remote_port_num = remote_port_num;
				p_edge->weight = 1;
				p_edge++;

				p_vertex->num_edges++;
			} else if (port_connects_to_hca(p_subn, p_sw, port_num, plane)) {
				p_vertex->num_hcas++;
			}
		}

		if (p_vertex->num_edges > p_graph->max_vertex_edges) {
			p_graph->max_vertex_edges = p_vertex->num_edges;
		}
	}

	/*
	 * Add the edge weights.
	 */
	set_edge_hop_weights(p_mgr, p_graph);

	graph_print(p_graph);
	ret = p_graph;

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}
