/*
 * Copyright (c) 2022-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    OSM Key Manager for key configuration of multiple classes
 *
 * Author:
 *    Or Nechemia, Mellanox
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <iba/ib_types.h>
#include <complib/cl_debug.h>
#include <complib/cl_sha512.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_KEY_MGR_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_subnet.h>
#include <opensm/osm_congestion_control.h>
#include <opensm/osm_vendor_specific.h>
#include <opensm/osm_key_mgr.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_db_pack.h>
#include <opensm/osm_inform.h>

int fill_random_bytes(uint8_t * buffer, size_t buffer_size);
int osm_subn_refresh_key_seed(osm_sm_t * p_sm);

static inline ib_net64_t key_mgr_get_seed_picker(const osm_subn_opt_t * p_opt)
{
	return p_opt->key_mgr_seed ? p_opt->key_mgr_seed : p_opt->m_key;
}

static inline void key_mgr_get_key_seed(osm_key_mgr_t * p_key_mgr, uint8_t seed[64], size_t * seed_size)
{
	ib_net64_t seed_picker;
	const void * src;
	const osm_subn_opt_t * p_opt = p_key_mgr->p_opt;

	seed_picker = key_mgr_get_seed_picker(p_opt);
	if (seed_picker == OSM_KEY_SEED_RANDOM) {
		src = p_key_mgr->key_seed;
		*seed_size = sizeof(p_key_mgr->key_seed);
	} else if (p_opt->key_mgr_seed) {
		src = &p_opt->key_mgr_seed;
		*seed_size = sizeof(p_opt->key_mgr_seed);
	} else {
		src = &p_opt->m_key;
		*seed_size = sizeof(p_opt->m_key);
	}

	if (seed_picker == OSM_KEY_SEED_RANDOM)
		CL_PLOCK_ACQUIRE(&p_key_mgr->seed_lock);

	memcpy(seed, src, *seed_size);

	if (seed_picker == OSM_KEY_SEED_RANDOM)
		CL_PLOCK_RELEASE(&p_key_mgr->seed_lock);
}

static ib_net64_t key_mgr_calc_key(uint8_t * seed,
				   size_t seed_size,
				   ib_net64_t guid,
				   uint8_t mgmt_class,
				   uint8_t key_type)
{
	uint8_t sha512_result[SHA512_DIGEST_LENGTH];
	cl_sha512_ctxt_t sha512_ctxt;

	/* For unknown GUIDs, use key 0 */
	if (!guid)
		return 0;

	cl_sha512_init(&sha512_ctxt);
	cl_sha512_update(&sha512_ctxt, seed, seed_size);
	cl_sha512_update(&sha512_ctxt, (uint8_t *)&guid, sizeof(guid));
	cl_sha512_update(&sha512_ctxt, (uint8_t *)&mgmt_class, sizeof(mgmt_class));
	if (key_type)
		cl_sha512_update(&sha512_ctxt, (uint8_t *)&key_type, sizeof(key_type));
	cl_sha512_digest(sha512_result, &sha512_ctxt);
	return ((ib_net64_t *)sha512_result)[1];
}

static inline ib_net64_t osm_key_mgr_calculate_key(osm_key_mgr_t * p_key_mgr,
						   ib_net64_t guid,
						   uint8_t mclass,
						   uint8_t key_type)
{
	uint8_t seed[64];
	size_t seed_size;

	key_mgr_get_key_seed(p_key_mgr, seed, &seed_size);

	return key_mgr_calc_key(seed, seed_size, guid, mclass, key_type);
}

static ib_net64_t * key_mgr_get_port_key(osm_port_t * p_port, uint8_t mgmt_class,
					 uint8_t key_type)
{
	switch (mgmt_class) {
	case IB_MCLASS_CC:
		return &p_port->cc_key;
	case IB_MCLASS_VS:
		return &p_port->vs_key;
	case IB_MCLASS_N2N:
		if (key_type)
			return &p_port->n2n_data.n2n_key;
		return &p_port->n2n_data.n2n_mgr_key;
	default:
		return NULL;
	}
}

static inline boolean_t key_mgr_is_using_random_seed(const osm_key_mgr_t * p_key_mgr)
{
	return key_mgr_get_seed_picker(p_key_mgr->p_opt) == OSM_KEY_SEED_RANDOM;
}

static int key_mgr_refresh_seed(osm_key_mgr_t * p_key_mgr)
{
	int ret = 0;
	boolean_t using_random_seed;

	using_random_seed = key_mgr_is_using_random_seed(p_key_mgr);
	OSM_ASSERT(using_random_seed);

	CL_PLOCK_EXCL_ACQUIRE(&p_key_mgr->seed_lock);
	if (fill_random_bytes(p_key_mgr->key_seed, sizeof(p_key_mgr->key_seed))) {
		OSM_LOG(p_key_mgr->p_log, OSM_LOG_ERROR,
			"ERR 9700: Failed to read seed from /dev/urandom\n");
		ret = 1;
		goto Exit;
	}

Exit:
	CL_PLOCK_RELEASE(&p_key_mgr->seed_lock);
	return ret;
}

static cl_status_t osm_key_mgr_restart_periodic_update_timer(osm_key_mgr_t * p_key_mgr)
{
	cl_status_t cl_status;
	uint32_t factor = MINUTES_TO_MILLISEC;

	if (getenv("OSM_DEBUG_PERIODIC_UPDATE_INTERVAL_SECONDS"))
		factor = SECOND_TO_MILLIS;

	cl_status = cl_timer_start(&p_key_mgr->periodic_key_update_timer,
				   p_key_mgr->p_opt->periodic_key_update * factor);
	if (cl_status != CL_SUCCESS)
		OSM_LOG(p_key_mgr->p_log, OSM_LOG_ERROR, "ERR 9701: Failed to restart periodic key update timer\n");

	return cl_status;
}

static int osm_key_mgr_refresh_seed_signal_update(osm_key_mgr_t * p_key_mgr)
{
	int ret = 0;

	OSM_LOG_ENTER(p_key_mgr->p_log);

	if (key_mgr_refresh_seed(p_key_mgr)) {
		ret = 1;
		goto Exit;
	}

	if (osm_subn_refresh_key_seed(p_key_mgr->p_sm)) {
		ret = 1;
		goto Exit;
	}

	/*
	 * We set a timer to fire again as a backup. This is because,
	 * if SM is not in master state, it will ignore the signal (below)
	 * and our flow will be interrupted (we will never update keys again).
	 * In a normal flow (where SM is master), this timer will be restarted
	 * again from osm_key_mgr_restart_periodic_update.
	 */
	(void)osm_key_mgr_restart_periodic_update_timer(p_key_mgr);

	p_key_mgr->restart_timer = TRUE;
	osm_sm_signal(p_key_mgr->p_sm, OSM_SIGNAL_UPDATE_KEYS);

Exit:
	OSM_LOG_EXIT(p_key_mgr->p_log);
	return ret;
}

cl_status_t osm_key_mgr_restart_periodic_update(osm_key_mgr_t * p_key_mgr, boolean_t force_restart)
{
	cl_status_t cl_status = CL_SUCCESS;

	if (force_restart || p_key_mgr->restart_timer) {
		cl_status = osm_key_mgr_restart_periodic_update_timer(p_key_mgr);
		if (cl_status != CL_SUCCESS)
			p_key_mgr->periodic_update_active = FALSE;
		else
			p_key_mgr->periodic_update_active = TRUE;

		p_key_mgr->restart_timer = FALSE;
	}

	return cl_status;
}

void osm_key_mgr_stop_periodic_update(osm_key_mgr_t * p_key_mgr)
{
	cl_timer_stop(&p_key_mgr->periodic_key_update_timer);
	p_key_mgr->periodic_update_active = FALSE;
}

void osm_key_mgr_intiate_key_update(osm_sm_t * p_sm)
{
	p_sm->key_mgr.recalc_keys_reason = OSM_KEY_MGR_RECALC_KEYS_REASON_PERIODIC_UPDATE;

	/*
	 * What we need from the sweep:
	 * 1. key change function invocation
	 * 2. key db flush
	 * 3. periodic update timer restart
	 */
	p_sm->p_subn->force_heavy_sweep = TRUE;
	OSM_LOG(p_sm->p_log, OSM_LOG_INFO, "Periodic key update: invoking heavy sweep\n");
	osm_sm_sweep(p_sm);
}

void osm_key_mgr_key_update_preprocess(osm_sm_t * p_sm)
{
	if (p_sm->key_mgr.recalc_keys_reason != OSM_KEY_MGR_RECALC_KEYS_REASON_PERIODIC_UPDATE)
		return;

	// TODO - this should be an api of osm_[vs|cc]_*, should we synchronize access to it?
	p_sm->p_subn->p_osm->vs.recalculate_keys = TRUE;
	p_sm->p_subn->p_osm->cc.recalculate_keys = TRUE;
	p_sm->p_subn->p_osm->n2n.recalculate_keys = TRUE;

	p_sm->key_mgr.num_mkeys_configured = 0;
}

void osm_key_mgr_key_update_postprocess(osm_sm_t * p_sm)
{
	if (p_sm->key_mgr.recalc_keys_reason != OSM_KEY_MGR_RECALC_KEYS_REASON_PERIODIC_UPDATE)
		return;

	p_sm->key_mgr.recalc_keys_reason = 0;

	OSM_LOG(p_sm->p_log, OSM_LOG_INFO, "Periodic key update cycle completed\n");
	OSM_LOG(p_sm->p_log, OSM_LOG_VERBOSE,
		"%u out of %u ports were successfully configured for M_Key\n",
		p_sm->key_mgr.num_mkeys_configured, cl_qmap_count(&p_sm->p_subn->port_guid_tbl));
	osm_report_notice_1322(p_sm->p_subn);

	osm_key_mgr_restart_periodic_update(&p_sm->key_mgr, FALSE);
}

static void osm_key_mgr_timer_callback(void * context)
{
	osm_key_mgr_refresh_seed_signal_update((osm_key_mgr_t *)context);
}

ib_api_status_t osm_key_mgr_init(osm_key_mgr_t * p_key_mgr, struct osm_sm * p_sm)
{
	ib_api_status_t status = IB_SUCCESS;
	cl_timer_t * p_timer = &p_key_mgr->periodic_key_update_timer;

	OSM_ASSERT(p_key_mgr);
	OSM_ASSERT(p_sm);

	memset(p_key_mgr, 0, sizeof(*p_key_mgr));
	p_key_mgr->p_sm = p_sm;
	p_key_mgr->p_opt = &p_sm->p_subn->opt;
	p_key_mgr->p_log = p_sm->p_log;
	p_key_mgr->periodic_update_active = FALSE;

	if ((status = cl_timer_init(p_timer, osm_key_mgr_timer_callback, p_key_mgr)) != CL_SUCCESS)
		goto Exit;

	if ((status = cl_plock_init(&p_key_mgr->seed_lock)) != CL_SUCCESS)
		goto Exit_timer;

	if (key_mgr_is_using_random_seed(p_key_mgr) &&
	    ((key_mgr_refresh_seed(p_key_mgr) ||
	      ((p_key_mgr->p_opt->periodic_key_update > 0) &&
	       (osm_key_mgr_restart_periodic_update(p_key_mgr, TRUE) != CL_SUCCESS))))) {
		status = IB_INSUFFICIENT_RESOURCES;
		goto Exit_timer_lock;
	}

Exit_timer_lock:
	if (status != IB_SUCCESS)
		cl_plock_destroy(&p_key_mgr->seed_lock);

Exit_timer:
	if (status != IB_SUCCESS)
		cl_timer_destroy(p_timer);
Exit:
	return status;
}

void osm_key_mgr_destroy(osm_key_mgr_t * p_key_mgr)
{
	osm_key_mgr_stop_periodic_update(p_key_mgr);
	cl_plock_destroy(&p_key_mgr->seed_lock);
	cl_timer_destroy(&p_key_mgr->periodic_key_update_timer);
}

osm_db_domain_t * key_mgr_get_key_db(osm_opensm_t * p_osm, uint8_t mgmt_class, uint8_t key_type)
{
	osm_subn_t *p_subn = &p_osm->subn;

	switch (mgmt_class) {
	case IB_MCLASS_CC:
		return p_subn->p_g2cckey;
	case IB_MCLASS_VS:
		return p_subn->p_g2vskey;
	case IB_MCLASS_N2N:
		if (key_type)
			return p_subn->p_g2n2nkey;
		return p_subn->p_g2n2n_mgrkey;
	default:
		return NULL;
	}
}

boolean_t osm_key_mgr_db_is_initialized(osm_opensm_t * p_osm, uint8_t mgmt_class, uint8_t key_type)
{
	return (key_mgr_get_key_db(p_osm, mgmt_class, key_type) ? TRUE : FALSE);
}

int osm_key_mgr_flush_db_to_disk(osm_opensm_t * p_osm, uint8_t mgmt_class, uint8_t key_type)
{
	return osm_db_store(key_mgr_get_key_db(p_osm, mgmt_class, key_type),
			    p_osm->subn.opt.fsync_high_avail_files);
}

// TODO - create an array of all classes? s.t. adding one be only to the array and not in tons of places
boolean_t key_mgr_is_dual_key_supported(const osm_port_t * p_port, uint8_t mgmt_class)
{
	struct osm_node * p_node;

	switch (mgmt_class) {
	case IB_MCLASS_CC:
		p_node = osm_port_get_node(p_port);
		return ib_cc_enh_info_is_dual_key_supported(&p_node->cc_enh_info);
	case IB_MCLASS_VS:
		/* Currently, we assume that dual key is supported for VS class if it's supported for SMP class */
		return ib_mlnx_epi_supports_dual_mkey(&osm_port_get_physp(p_port)->ext_port_info);
	case IB_MCLASS_N2N:
		return ib_n2n_cpi_is_dual_key_supported(&p_port->n2n_data.cpi);
	default:
		return FALSE;
	}
}

ib_net64_t osm_key_mgr_generate_key(osm_opensm_t * p_osm, osm_port_t * p_port,
				    uint8_t mgmt_class, uint8_t key_type)
{
	osm_subn_t *p_subn = &p_osm->subn;
	ib_net64_t key = 0;
	ib_net64_t guid = p_port->guid;
	uint8_t key_enable;
	osm_db_domain_t *p_g2key =
	    key_mgr_get_key_db(p_osm, mgmt_class, key_type);
	boolean_t recalculate_keys;
	osm_key_mgr_t * p_key_mgr = &p_osm->sm.key_mgr;

	/* In case SM got trap before key DB was initialized */
	if (!p_g2key)
		goto Exit;

	switch (mgmt_class) {
	case IB_MCLASS_CC:
		key_enable = p_subn->opt.cc_key_enable;
		recalculate_keys = p_osm->cc.recalculate_keys;
		break;
	case IB_MCLASS_VS:
		key_enable = p_subn->opt.vs_key_enable;
		recalculate_keys = p_osm->vs.recalculate_keys;
		break;
	case IB_MCLASS_N2N:
		key_enable = p_subn->opt.n2n_key_enable;
		recalculate_keys = p_osm->n2n.recalculate_keys;
		break;
	default:
		goto Exit;
	}

	if (key_enable == OSM_CAPABILITY_DISABLED) {
		osm_db_guid2key_set(p_g2key, cl_ntoh64(guid),
				     cl_ntoh64(key));
		goto Exit;
	}

	/*
	 * Before periodic update, recalculate_keys was true only when seed changed.
	 * Currently, we set recalculate_keys to TRUE each time to do the periodic update.
	 * Therefore, we have to recalculate keys only when:
	 * 1. key_mgr_seed is changed (as before), or
	 * 2. when periodic update is active and device supports dual key
	 */
	if (p_key_mgr->recalc_keys_reason == OSM_KEY_MGR_RECALC_KEYS_REASON_PERIODIC_UPDATE)
		recalculate_keys = key_mgr_is_dual_key_supported(p_port, mgmt_class);
	else if ((p_key_mgr->recalc_keys_reason == OSM_KEY_MGR_RECALC_KEYS_REASON_OTHER) &&
		 p_key_mgr->periodic_update_active)
		/*
		 * If periodic update is active but recalc reason is reason is not periodic update,
		 * then this is the first time we are configuring keys (in periodic update).
		 * In this case we have to wait one cycle and not update keys.
		 */
		recalculate_keys = FALSE;

	if (recalculate_keys || osm_db_guid2key_get(p_g2key, cl_ntoh64(guid), &key)) {
		key = osm_key_mgr_calculate_key(p_key_mgr, guid, mgmt_class, key_type);
		osm_db_guid2key_set(p_g2key, cl_ntoh64(guid),
				     cl_ntoh64(key));
	} else {
		key = cl_hton64(key);
	}

Exit:
	return key;
}

void osm_key_mgr_get_cached_key(osm_opensm_t * p_osm, osm_port_t * p_port,
				uint8_t mgmt_class, uint8_t key_type)
{
	ib_net64_t *p_port_key =
	    key_mgr_get_port_key(p_port, mgmt_class, key_type);
	osm_db_domain_t *p_g2key =
	    key_mgr_get_key_db(p_osm, mgmt_class, key_type);
	ib_net64_t key = 0;
	ib_net64_t guid = p_port->guid;

	(void)osm_db_guid2key_get(p_g2key, cl_ntoh64(guid), &key);
	*p_port_key = cl_hton64(key);
}

void osm_key_mgr_set_cached_key(osm_opensm_t * p_osm, osm_port_t * p_port,
				uint8_t mgmt_class, uint8_t key_type)
{
	ib_net64_t *p_port_key =
	    key_mgr_get_port_key(p_port, mgmt_class, key_type);
	osm_db_domain_t *p_g2key =
	    key_mgr_get_key_db(p_osm, mgmt_class, key_type);
	ib_net64_t guid = p_port->guid;

	/* In case SM got trap before key DB was initialized */
	if (!p_g2key || !p_port_key)
		return;

	osm_db_guid2key_set(p_g2key, cl_ntoh64(guid), cl_ntoh64(*p_port_key));
}

void osm_key_mgr_set_key_if_not_present_in_db(
	struct osm_opensm * p_osm, osm_port_t * p_port,
	uint8_t mgmt_class, uint8_t key_type,
	uint64_t key)
{
	const uint64_t guid = osm_port_get_guid_ho(p_port);
	osm_db_domain_t * p_g2key = key_mgr_get_key_db(p_osm, mgmt_class, key_type);

	if (osm_db_guid2key_get(p_g2key, guid, NULL))
		osm_db_guid2key_set(p_g2key, guid, key);
}
