/*
 * Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *	Module for Link as a Service configuration for NVLink systems
 *
 * Author:
 *	Or Nechemia, NVIDIA
 */
#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_NVLINK_PRTN_C
#include <opensm/osm_opensm.h>
#include <opensm/osm_partition.h>
#include <opensm/osm_node.h>
#include <opensm/osm_nvlink.h>
#include <opensm/osm_nvlink_prtn.h>
#include <opensm/osm_hierarchy_info.h>
#include <opensm/osm_db_pack.h>
#include <opensm/osm_db.h>
#include <iba/ib_types.h>

#define OSM_NVL_PORT2NVL_FILE	"port2nvl_prtn"


extern int wait_for_pending_transactions(osm_opensm_t * p_osm);

static ib_api_status_t nvl_prtn_add_remote_switch_to_laas(osm_subn_t * p_subn,
							  osm_nvlink_prtn_t * p,
							  osm_physp_t * p_physp);

static void nvl_prtn_remove_port_from_prtn(osm_subn_t * p_subn, osm_nvlink_prtn_t * p_prtn,
					   osm_physp_t * p_physp);

uint8_t osm_nvl_prtn_get_port_subn_admin_state(osm_nvlink_mgr_t * p_mgr, uint64_t node_guid, uint8_t port_num)
{
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	
	p_node = osm_get_node_by_guid(p_mgr->p_subn, cl_hton64(node_guid));
	if (!p_node)
		return IB_NMX_ADMINSTATE_NA;

	p_physp = osm_node_get_physp_ptr(p_node, port_num);
	if (!p_physp || !p_physp->ext_port_info_updated)
		return IB_NMX_ADMINSTATE_NA;

	return ib_mlnx_epi_get_nmx_admin_state(&p_physp->ext_port_info);
}

static uint8_t nvl_prtn_get_port_req_admin_state(osm_nvlink_mgr_t * p_mgr, osm_physp_t *p_physp)
{
	osm_nvl_admin_state_map_item_t *p_admin_state_item;
	cl_fmap_item_t *p_item;
	osm_sw_port_t sw_port_key;

	sw_port_key.guid = osm_node_get_node_guid_ho(p_physp->p_node);
	sw_port_key.port_num = p_physp->port_num;

	p_item = cl_fmap_get(&p_mgr->nvl_admin_states, &sw_port_key);
	if (p_item == cl_fmap_end(&p_mgr->nvl_admin_states))
		return IB_NMX_ADMINSTATE_NA;

	p_admin_state_item = (osm_nvl_admin_state_map_item_t *)p_item;

	return p_admin_state_item->admin_state;
}

static osm_nvlink_prtn_t * nvl_prtn_get_nvl_prtn_by_pkey(osm_nvlink_mgr_t * p_mgr, ib_net16_t pkey)
{
	cl_map_item_t *p_prtn_item;

	p_prtn_item = cl_qmap_get(&p_mgr->nvl_prtn_tbl, pkey);
	return (p_prtn_item == cl_qmap_end(&p_mgr->nvl_prtn_tbl)) ?
	    NULL : (osm_nvlink_prtn_t *)p_prtn_item;
}

static void nvl_prtn_set_mlnx_extended_port_info(osm_nvlink_mgr_t * p_mgr, osm_physp_t * p_physp)
{
	osm_sm_t *sm = &p_mgr->p_subn->p_osm->sm;
	osm_madw_context_t mad_context;
	osm_node_t *p_node;
	osm_dr_path_t *p_path;
	ib_api_status_t status = IB_NOT_DONE;
	ib_mlnx_ext_port_info_t mepi;
	uint8_t admin_state_req;

	OSM_LOG_ENTER(sm->p_log);

	p_node = p_physp->p_node;

	OSM_ASSERT(p_node && p_physp);

	/* This MEPI sets Admin State for switch trunk ports only */
	if (osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH)
		goto Exit;

	/* If not responded to latest MEPI, will not be configured to new admin state */
	if (!is_mepi_supported(sm, p_node) ||
	    !ib_general_info_is_cap_enabled(&p_node->general_info,
					    IB_GENERAL_INFO_IS_NMX_ADMINSTATE_SUPPORTED) ||
	    !p_physp->ext_port_info_updated)
		goto Exit;

	p_path = osm_physp_get_dr_path_ptr(sm, p_physp);
	mepi = p_physp->ext_port_info;
	admin_state_req = nvl_prtn_get_port_req_admin_state(p_mgr, p_physp);

	/* Parameter was not updated - no need to send MEPI SET */
	if (admin_state_req == IB_NMX_ADMINSTATE_NA ||
	    (ib_mlnx_epi_get_nmx_admin_state(&mepi) == admin_state_req &&
	     !p_physp->need_update))
		goto Exit;

	ib_mlnx_epi_set_state_change_enable_nmx_admin_state(&mepi);
	ib_mlnx_epi_set_nmx_admin_state(&mepi, admin_state_req);

	memset(&mad_context, 0, sizeof(osm_madw_context_t));
	mad_context.pi_context.node_guid = osm_node_get_node_guid(p_node);
	mad_context.pi_context.port_guid = osm_physp_get_port_guid(p_physp);
	mad_context.pi_context.set_method = TRUE;

	OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
		 "Sending MLNXExtendedPortInfo attribute for switch 0x%" PRIx64 " port %u "
		 "changing admin state from %u to %u\n",
		cl_ntoh64(osm_node_get_node_guid(p_node)), p_physp->port_num,
		admin_state_req, ib_mlnx_epi_get_nmx_admin_state(&mepi));

	p_physp->ext_port_info_updated = FALSE;
	status = osm_req_set(sm, p_path, p_physp, (uint8_t *) &mepi,
			     sizeof(mepi), IB_MAD_ATTR_MLNX_EXTENDED_PORT_INFO,
			     cl_hton32(p_physp->port_num), FALSE, ib_port_info_get_m_key(&p_physp->port_info),
			     0, CL_DISP_MSGID_NONE, &mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR BB0A: "
			"Sending MLNXExtendedPortInfo attribute for switch 0x%" PRIx64 " port %u failed (%s)\n",
			cl_ntoh64(osm_node_get_node_guid(p_node)), p_physp->port_num,
			ib_get_err_str(status));

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

char * osm_nvl_prtn_get_admin_state_str(uint8_t admin_state)
{
	switch(admin_state) {
	case IB_NMX_ADMINSTATE_NA:
		return "N/A";
	case IB_NMX_ADMINSTATE_UP:
		return "UP";
	case IB_NMX_ADMINSTATE_DOWN:
		return "DOWN";
	case IB_NMX_ADMINSTATE_DIAG:
		return "DIAG";
	default:
		return "UNKNOWN";
	}
}

static void nvl_prtn_process_admin_state_ports(cl_map_item_t * obj, IN void * context)

{
	osm_switch_t *p_sw = (osm_switch_t *) obj;
	osm_nvlink_mgr_t *p_mgr = context;
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num;

	OSM_ASSERT(p_sw);

	num_ports = osm_node_get_num_physp(p_sw->p_node);
	for (port_num = 1; port_num < num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
		if (!p_physp)
			continue;

		/* Configure Admin State to ports */
		nvl_prtn_set_mlnx_extended_port_info(p_mgr, p_physp);

		/*
		 * NOTE: SM sets Admin State values to subnet and it might cause partition degradation,
		 * 	 GFM is responsible for sending reroute request if necessary
		 */
	}
}

ib_api_status_t nvl_prtn_add_to_admin_state_map(osm_nvlink_mgr_t * p_mgr,
						uint64_t guid,
						uint8_t port_num,
						uint8_t admin_state)
{
	osm_nvl_admin_state_map_item_t *p_admin_state_item;
	cl_fmap_item_t *p_item;
	osm_sw_port_t sw_port_key;
	osm_sw_port_t *p_sw_port_key;

	sw_port_key.guid = guid;
	sw_port_key.port_num = port_num;
	p_item = cl_fmap_get(&p_mgr->nvl_admin_states, &sw_port_key);
	if (p_item != cl_fmap_end(&p_mgr->nvl_admin_states)) {
		p_admin_state_item = (osm_nvl_admin_state_map_item_t *)p_item;
		p_admin_state_item->admin_state = admin_state;
		return IB_SUCCESS;
	}

	p_sw_port_key = (osm_sw_port_t *)calloc(1, sizeof(osm_sw_port_t));
	if (!p_sw_port_key) {
		OSM_LOG(&p_mgr->p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB0B: "
			"Could not allocate memory for admin_state_map key\n");
		return IB_INSUFFICIENT_MEMORY;
	}

	p_sw_port_key->guid = guid;
	p_sw_port_key->port_num = port_num;

	/* Switch port is not part of admin state table. Create item */
	p_admin_state_item = (osm_nvl_admin_state_map_item_t *)calloc(1, sizeof(osm_nvl_admin_state_map_item_t));
	if (!p_admin_state_item) {
		OSM_LOG(&p_mgr->p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB0C: "
			"Could not allocate memory for admin_state_map item\n");
		free(p_sw_port_key);
		return IB_INSUFFICIENT_MEMORY;
	}

	p_admin_state_item->admin_state = admin_state;
	/* if already exist, need to update previous item */
	p_item = cl_fmap_insert(&p_mgr->nvl_admin_states, p_sw_port_key, &p_admin_state_item->fmap_item);
	if (p_item != &p_admin_state_item->fmap_item) {
		OSM_LOG(&p_mgr->p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB0D: "
			"Failed to insert 0x%"PRIx64" port %u to admin state map\n",
			guid, port_num);

		free(p_sw_port_key);
		free(p_admin_state_item);
		return IB_ERROR;
	}

	return IB_SUCCESS;
}

void nvl_prtn_remove_from_admin_state_map(osm_nvlink_mgr_t * p_mgr,
					  uint64_t guid,
					  uint8_t port_num)
{
	cl_fmap_item_t *p_item;
	osm_sw_port_t sw_port_key;

	sw_port_key.guid = guid;
	sw_port_key.port_num = port_num;

	p_item = cl_fmap_get(&p_mgr->nvl_admin_states, &sw_port_key);
	if (p_item == cl_fmap_end(&p_mgr->nvl_admin_states))
		return;

	cl_fmap_remove_item(&p_mgr->nvl_admin_states, p_item);
	free((osm_sw_port_t *)cl_fmap_key(p_item));
	free(p_item);
}

ib_api_status_t osm_nvl_prtn_set_port_admin_state(osm_nvlink_mgr_t * p_mgr,
						  uint64_t guid,
						  uint8_t port_num,
						  uint8_t admin_state)
{
	ib_api_status_t status = IB_SUCCESS;

	if (admin_state > IB_NMX_ADMINSTATE_DIAG) {
		status = IB_INVALID_PARAMETER;
		goto Exit;
	}

	/* admin state map should hold all subnet switch ports for UP/DOWN/DIAG states */
	if (admin_state == IB_NMX_ADMINSTATE_NA)
		nvl_prtn_remove_from_admin_state_map(p_mgr, guid, port_num);
	else
		status = nvl_prtn_add_to_admin_state_map(p_mgr, guid, port_num, admin_state);

Exit:
	return status;
}

static void nvl_prtn_remove_all_nvl_sw_tbl(osm_nvlink_mgr_t * p_mgr, osm_nvlink_prtn_t * p_prtn)
{
	cl_map_t *p_tbl = &p_prtn->sw_tbl;
	cl_map_iterator_t p_next, p_item;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;
	osm_physp_t *p_physp;
	osm_switch_t *p_sw;
	uint8_t port_num;
	ib_net64_t sw_guid;

	p_next = cl_map_head(p_tbl);
	while (p_next != cl_map_end(p_tbl)) {
		p_item = p_next;
		sw_guid = cl_map_key(p_next);
		p_next = cl_map_next(p_item);

		p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_obj(p_item);
		free(p_prtn_sw_data);

		cl_map_remove_item(p_tbl, p_item);

		p_sw = osm_get_switch_by_guid(p_mgr->p_subn, sw_guid);
		if (!p_sw)
			continue;

		for (port_num = 1; port_num < osm_node_get_num_physp(p_sw->p_node); port_num++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
			if (!p_physp)
				continue;

			if (p_physp->nvl_laas_pkey != p_prtn->pkey)
				continue;

			nvl_prtn_remove_port_from_prtn(p_mgr->p_subn, p_prtn, p_physp);
		}
	}
}

static osm_nvlink_prtn_t * nvl_prtn_new(osm_nvlink_mgr_t * p_mgr, ib_net16_t pkey)
{
	osm_nvlink_prtn_t *p_nvl_prtn = NULL;
	cl_map_item_t *p_check;
	osm_log_t *p_log = &p_mgr->p_subn->p_osm->log;

	pkey &= cl_hton16((uint16_t) ~ 0x8000);
	if (!pkey) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BB14: "
			"Unable to create NVL partition with pkey 0x%04x\n",
			cl_ntoh16(pkey));
		return NULL;
	}

	p_nvl_prtn = malloc(sizeof(osm_nvlink_prtn_t));
	if (!p_nvl_prtn) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BB15: "
			"Unable to create NVL partition 0x%04x\n",
			cl_ntoh16(pkey));
		return NULL;
	}

	memset(p_nvl_prtn, 0, sizeof(osm_nvlink_prtn_t));
	p_nvl_prtn->pkey = pkey;
	cl_map_construct(&p_nvl_prtn->sw_tbl);
	cl_map_init(&p_nvl_prtn->sw_tbl, 2);

	p_check = cl_qmap_insert(&p_mgr->nvl_prtn_tbl, p_nvl_prtn->pkey, &p_nvl_prtn->map_item);
	if (p_check == &p_nvl_prtn->map_item) {
		OSM_LOG(p_log, OSM_LOG_VERBOSE, "Create new NVL partition 0x%04x\n",
			cl_ntoh16(pkey));
	} else {
		OSM_LOG(p_log, OSM_LOG_VERBOSE, "Duplicated NVL partition"
			" definition: 0x%04x. Skips creating new partition\n",
			cl_ntoh16(pkey));

		osm_nvl_prtn_destroy_prtn(p_mgr, p_nvl_prtn);
		p_nvl_prtn = (osm_nvlink_prtn_t *)p_check;
	}

	return p_nvl_prtn;
}

osm_nvlink_prtn_t * osm_nvl_prtn_make_new(IN osm_nvlink_mgr_t * p_mgr,
					  IN const char *name,
					  IN ib_net16_t pkey,
					  IN uint8_t max_trunk_links,
					  IN boolean_t is_reroute_req,
					  OUT osm_prtn_t ** pp_prtn)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_log_t *p_log = &p_mgr->p_subn->p_osm->log;
	osm_nvlink_prtn_t *p_nvl_prtn = NULL;
	boolean_t is_prtn_new = FALSE;

	/* This function shall be used only for NVLink configuration */
	if (!osm_subn_is_nvlink_enabled(p_subn))
		return p_nvl_prtn;

	if (!osm_prtn_find_by_pkey(p_subn, pkey))
		is_prtn_new = TRUE;

	if (pp_prtn) {
		*pp_prtn = osm_prtn_make_new(p_log, p_subn, name, pkey);
		if (!*pp_prtn)
			return NULL;
	}

	/* Create new NVLink partition */
	p_nvl_prtn = nvl_prtn_new(p_mgr, pkey);
	if (!p_nvl_prtn)
		return NULL;

	/*
	 * In case partition already exists, reroute request requires to set partition ports
	 * to Contain and Drain before rerouting is allowed. However in case the partition is new,
	 * no need for SM to set partition ports to Contain and Drain first.
	 */
	if (is_reroute_req)
		p_nvl_prtn->state = is_prtn_new ? NVL_PRTN_STATE_REROUTE : NVL_PRTN_STATE_REROUTE_REQUESTED;

	p_nvl_prtn->max_trunk_links = max_trunk_links;

	return p_nvl_prtn;
}

int osm_nvl_prtn_add_gpu(osm_nvlink_mgr_t *p_mgr, osm_nvlink_prtn_t *p_nvl_prtn, ib_net64_t node_guid)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_sm_t *sm = &p_subn->p_osm->sm;
	osm_node_t *p_node;
	osm_gpu_t *p_gpu;
	osm_physp_t *p_physp;
	int port_num;
	uint8_t plane;
	osm_prtn_t *p_prtn = NULL;

	p_node = osm_get_node_by_guid(p_subn, node_guid);
	if (!p_node) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Add GPU GUID 0x%" PRIx64 " to partition 0x%x failed - GUID does not exists\n",
			cl_ntoh64(node_guid),
			cl_ntoh16(p_nvl_prtn->pkey));
		return 1;
	}

	p_gpu = p_node->p_gpu;

	if (!p_gpu) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR BB21: "
			"Add GPU GUID 0x%" PRIx64 " to partition 0x%x failed - node is not GPU\n",
			cl_ntoh64(node_guid),
			cl_ntoh16(p_nvl_prtn->pkey));
		return 1;
	}

	cl_hashmap_insert(&p_mgr->gpu_to_pkey_map, node_guid, p_nvl_prtn->pkey);

	/*
	 * At this point, no support for delete GPU from one partition and add GPU to another
	 * at the same sweep. Hence, if marked as member, assume it is the same prtn and no action
	 * is required on this routing.
	 */
	p_gpu->last_sweep_pkey = p_nvl_prtn->pkey;

	p_prtn = osm_prtn_find_by_pkey(p_subn, p_nvl_prtn->pkey);
	if (!p_prtn) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR BB22: "
			"Add GPU GUID 0x%" PRIx64 " to partition 0x%x failed - partition not found\n",
			cl_ntoh64(node_guid),
			cl_ntoh16(p_nvl_prtn->pkey));
		return 1;
	}

	/* Add all ports of the node to the partition */
	for (port_num = 1; port_num < osm_node_get_num_physp(p_node); port_num++) {
		p_physp = osm_node_get_physp_ptr(p_node, port_num);
		if (!p_physp)
			continue;

		/* All ports need to be FULL member */
		osm_insert_port_to_partition(sm, p_prtn,
					     OSM_PRTN_MEMBER_TYPE_FULL, p_physp->port_guid, FALSE);

		nvl_prtn_add_remote_switch_to_laas(p_subn, p_nvl_prtn, p_physp);

		/* Planes are already configured at this point */
		plane = osm_gpu_get_port_plane(p_gpu, port_num);
		if (p_subn->first_time_master_sweep) {
			/*
			 * On SM restart, should avoid changing the routing of GPUs that were already
			 * configured to the partitions (on ACTIVE/ARMED state), but add routing
			 * towards GPUs that are part of the partition but discovered for the first
			 * time (on INIT state).
			 */
			if (ib_port_info_get_port_state(&p_physp->port_info) == IB_LINK_INIT) {
				p_gpu->prtn_member_status[plane] = NVL_PRTN_GPU_STATUS_NEW;
			} else if (ib_port_info_get_port_state(&p_physp->port_info) == IB_LINK_ARMED ||
				   ib_port_info_get_port_state(&p_physp->port_info) == IB_LINK_ACTIVE) {
				p_gpu->prtn_member_status[plane] = NVL_PRTN_GPU_STATUS_MEMBER;
			}
			/*
			 * In any other case, status is kept as NON_MEMBER, as initialized,
			 * until added explicitly to the partition.
			*/
		} else if (p_gpu->prtn_member_status[plane] == NVL_PRTN_GPU_STATUS_NON_MEMBER) {
			p_gpu->prtn_member_status[plane] = NVL_PRTN_GPU_STATUS_NEW;
		}
	}

	return 0;
}

ib_net16_t osm_nvl_prtn_get_gpu_prtn(osm_nvlink_mgr_t *p_mgr, osm_gpu_t * p_gpu)
{
	uintptr_t key;
	uintptr_t value;

	OSM_ASSERT(p_gpu && p_gpu->p_node);

	key = osm_node_get_node_guid(p_gpu->p_node);
	value = cl_hashmap_find(&p_mgr->gpu_to_pkey_map, key);

	return (ib_net16_t)value;
}

/*
 * Shall be called right after pkey sweep, before rerouting, so GPU prtn_member_status
 * is updated for reroute phase.
 */
void nvl_prtn_update_deleted_gpu_member_status(cl_map_item_t * item, void * context)
{
	osm_nvlink_mgr_t *p_mgr = (osm_nvlink_mgr_t *)context;
	osm_gpu_t *p_gpu = (osm_gpu_t *)item;
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_log_t *p_log = &p_subn->p_osm->log;
	cl_map_item_t *p_prtn_item;
	osm_nvlink_prtn_t *p_prtn;
	ib_net16_t pkey = p_gpu->last_sweep_pkey;
	osm_physp_t *p_physp;
	uint8_t port_num;

	/*
	 * Update status of removed GPUs - that were members but no longer part of GPU to PKey DB.
	 * Added GPUs are updated by osm_nvl_prtn_add_gpu
	 */
	if (osm_nvl_prtn_get_gpu_prtn(p_mgr, p_gpu))
		return;

	OSM_LOG(p_log, OSM_LOG_DEBUG,
		"GPU 0x%"PRIx64" is marked as deleted from partition 0x%x\n",
		osm_node_get_node_guid_ho(p_gpu->p_node), cl_ntoh16(p_gpu->last_sweep_pkey));

	memset(p_gpu->prtn_member_status, NVL_PRTN_GPU_STATUS_NON_MEMBER, sizeof(p_gpu->prtn_member_status));
	p_gpu->last_sweep_pkey = 0x0;

	/* Remove GPU from rail filter of the partition it was removed from */
	p_prtn_item = cl_qmap_get(&p_mgr->nvl_prtn_tbl, pkey);
	if (p_prtn_item != cl_qmap_end(&p_mgr->nvl_prtn_tbl)) {
		p_prtn = (osm_nvlink_prtn_t *)p_prtn_item;

		for (port_num = 1; port_num < osm_node_get_num_physp(p_gpu->p_node); port_num++) {
			p_physp = osm_node_get_physp_ptr(p_gpu->p_node, port_num);
			if (!p_physp)
				continue;

			nvl_prtn_remove_port_from_prtn(p_subn, p_prtn, p_physp);
		}
	}
}

static ib_api_status_t nvl_prtn_validate_sw_tbl_cnd_info(osm_subn_t * p_subn,
							 osm_nvlink_prtn_t * p_prtn)
{
	cl_map_iterator_t p_next;
	ib_net64_t sw_guid;
	osm_switch_t *p_sw;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	p_next = cl_map_head(&p_prtn->sw_tbl);
	while (p_next != cl_map_end(&p_prtn->sw_tbl)) {
		sw_guid = cl_map_key(p_next);
		p_next = cl_map_next(p_next);

		p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
		if (!p_sw)
			continue;

		/* Switch CnDInfo configured successfully */
		if (p_sw->nvlink_sw_data.cnd_info_updated)
			continue;

		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB23: "
			"ContainAndDrainInfo was not set successfully for switch"
			" 0x%"PRIx64". Set partition 0x%x to error state\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid),
			cl_ntoh16(p_prtn->pkey));

		p_prtn->state = NVL_PRTN_STATE_ERROR;
		status = IB_ERROR;
	}

	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return status;
}

static void nvl_prtn_update_new_gpu_to_member_status(cl_map_item_t * item, void * context)
{
	osm_gpu_t *p_gpu = (osm_gpu_t *)item;
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num, plane;

	/*
	 * GPUs are considered new to partition from the sweep they were added until
	 * added to the plane.
	 * This functions is called by reroute_finish, hence all GPU known ports are set with planes.
	 */
	num_ports = osm_node_get_num_physp(p_gpu->p_node);
	for (port_num = 1; port_num < num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_gpu->p_node, port_num);
		if (!p_physp)
			continue;

		plane = osm_gpu_get_port_plane(p_gpu, p_physp->port_num);
		if (!plane)
			continue;

		if (p_gpu->prtn_member_status[plane] == NVL_PRTN_GPU_STATUS_NEW)
			p_gpu->prtn_member_status[plane] = NVL_PRTN_GPU_STATUS_MEMBER;
	}
}

boolean_t osm_nvl_prtn_get_gpu_reroute_allowed(osm_nvlink_mgr_t *p_mgr, osm_gpu_t * p_gpu,
					       uint8_t plane)
{
	cl_map_item_t *p_item;
	osm_nvlink_prtn_t *p_prtn;
	ib_net16_t pkey = osm_nvl_prtn_get_gpu_prtn(p_mgr, p_gpu);

	OSM_ASSERT(plane && plane <= OSM_NVLINK_MAX_PLANES);

	if (!pkey)
		return FALSE;

	p_item = cl_qmap_get(&p_mgr->nvl_prtn_tbl, pkey);
	if (p_item == cl_qmap_end(&p_mgr->nvl_prtn_tbl))
		return FALSE;

	p_prtn = (osm_nvlink_prtn_t *)p_item;
	if (p_gpu->prtn_member_status[plane] != NVL_PRTN_GPU_STATUS_NEW &&
	    p_prtn->state != NVL_PRTN_STATE_REROUTE)
		return FALSE;

	return TRUE;
}

/*
 * Check if specified pkey is in array of pkeys
 * Returns 1 if exist, otherwise, returns 0.
 */
static int nvl_ptrn_check_pkey_in_list(IN const uint16_t * pkeys, IN int num_pkeys, uint16_t pkey)
{
	int i;

	for (i = 0; i < num_pkeys; i++) {
		if (pkey == pkeys[i])
			return 1;
	}

	return 0;
}

void osm_nvl_prtn_destroy_prtn(osm_nvlink_mgr_t * p_mgr, osm_nvlink_prtn_t * p_prtn)
{
	nvl_prtn_remove_all_nvl_sw_tbl(p_mgr, p_prtn);
	cl_map_destroy(&p_prtn->sw_tbl);

	free(p_prtn);
}

void osm_nvl_prtn_destroy_prtn_by_pkey(osm_nvlink_mgr_t * p_mgr, ib_net16_t pkey)
{
	osm_nvlink_prtn_t * p_prtn;
	cl_map_item_t *p_prtn_item;

	p_prtn_item = cl_qmap_remove(&p_mgr->nvl_prtn_tbl, pkey);
	if (p_prtn_item == cl_qmap_end(&p_mgr->nvl_prtn_tbl))
		return;

	p_prtn = (osm_nvlink_prtn_t *)p_prtn_item;
	osm_nvl_prtn_destroy_prtn(p_mgr, p_prtn);
}

void nvl_prtn_validate_port2nvl_prtn_db(osm_nvlink_mgr_t * p_mgr, uint16_t * pkeys, int num_pkeys)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_log_t *p_log = &p_subn->p_osm->log;
	cl_qlist_t guids;
	osm_db_guid_portnum_elem_t *p_item;
	uint16_t pkey;
	boolean_t valid_entry = TRUE;
	osm_switch_t *p_sw;
	cl_map_item_t *p_sw_item;
	osm_physp_t *p_physp;
	osm_physp_t *p_remote_physp;
	uint64_t guid;
	uint32_t port;
	ib_net16_t pkey_no;

	OSM_LOG_ENTER(p_log);

	cl_qlist_init(&guids);

	if (osm_db_guid_portnum_keys_list(p_mgr->p_port2nvl_prtn, &guids)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BB10: "
			"could not get port list\n");
		goto Exit;
	}

	while ((p_item = (osm_db_guid_portnum_elem_t *) cl_qlist_remove_head(&guids))
	       != (osm_db_guid_portnum_elem_t *) cl_qlist_end(&guids)) {
		if (osm_db_port2pkey_get(p_mgr->p_port2nvl_prtn, p_item->guid, p_item->portnum, &pkey)) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BB11: "
				"could not get pkey for guid:0x%016" PRIx64 " port:%u\n",
				p_item->guid, p_item->portnum);
			free(p_item);
			continue;
		}

		valid_entry = TRUE;
		guid = p_item->guid;
		port = p_item->portnum;

		/* Check that the switch present */
		p_sw_item = cl_qmap_get(&p_subn->sw_guid_tbl, cl_hton64(guid));
		p_sw = (osm_switch_t *)p_sw_item;
		if (p_sw_item == cl_qmap_end(&p_subn->sw_guid_tbl)) {
			OSM_LOG(p_log, OSM_LOG_VERBOSE,
				"Switch with GUID 0x%lx not found\n",
				guid);
			valid_entry = FALSE;
		} else if (p_sw->num_ports < port) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BB42: "
				"Invalid port number %d for switch with GUID 0x%lx\n",
				port, guid);
			valid_entry = FALSE;
		} else if (!pkey || pkey > 0x7fff) {
			/* Check pkey value */
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BB43: "
				"Invalid pkey 0x%x for switch 0x%lx port number %d\n",
				pkey, guid, port);
			valid_entry = FALSE;
		} else if (!nvl_ptrn_check_pkey_in_list(pkeys, num_pkeys, pkey)) {
			/* Check pkey value */
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"Invalid pkey 0x%x of switch 0x%lx port number %d - partition does not exist\n",
				pkey, guid, port);
			valid_entry = FALSE;
		} else {
			/* Check pkey on physp */
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port);
			if (!p_physp) {
				OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BB44: "
					"Failed to get port object for switch GUID 0x%lx port number %d\n",
					guid, port);
				valid_entry = FALSE;
			} else if (osm_physp_is_fnm(p_subn, p_physp) ||
				   (p_physp->p_remote_physp &&
				    osm_node_get_type(p_physp->p_remote_physp->p_node) != IB_NODE_TYPE_SWITCH)) {
				/*
				 * DB holds only trunk ports (might be down), hence remove any FNMs
				 * and access ports.
				 */
				OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BB46: "
					"Invalid switch 0x%lx port number %d, not a trunk port\n",
					 guid, port);
				valid_entry = FALSE;
			} else if (p_physp->nvl_laas_pkey && p_physp->nvl_laas_pkey != cl_hton16(pkey)) {
				OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BB45: "
					"Contradicting pkeys 0x%x and 0x%x for switch 0x%lx port number %d\n",
					pkey, cl_ntoh16(p_physp->nvl_laas_pkey), guid, port);
				valid_entry = FALSE;
			} else {
				/* VALID ENTRY */
				pkey_no = cl_hton16(pkey);

				/*
				 * SM sets the pkey also on disconnected trunk ports to identify
				 * failed ports that were previously assigned to partition, and mark that
				 * this partition needs rerouting.
				 *
				 * After marking the partition for rerouting, the pkey should be
				 * removed from the port.
				 */
				p_physp->nvl_laas_pkey = pkey_no;

				/* Skip disconnected ports */
				p_remote_physp = osm_physp_get_remote(p_physp);
				if (p_remote_physp)
					p_remote_physp->nvl_laas_pkey = pkey_no;
			}
		}

		if (!valid_entry &&
		    osm_db_port2pkey_delete(p_mgr->p_port2nvl_prtn, p_item->guid, p_item->portnum)) {
			OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BB12: "
				"failed to delete entry for GUID:0x%016" PRIx64 " port:%u\n",
				p_item->guid, p_item->portnum);
		}
		free(p_item);
	}
Exit:
	OSM_LOG_EXIT(p_log);
}

static int nvl_prtn_compare_sw_ports(IN const void * p_item1, IN const void * p_item2)
{
	osm_sw_port_t *p_sw_port1 = (osm_sw_port_t *)p_item1;
	osm_sw_port_t *p_sw_port2 = (osm_sw_port_t *)p_item2;

	if (p_sw_port1->guid < p_sw_port2->guid)
		return -1;
	if (p_sw_port1->guid > p_sw_port2->guid)
		return 1;

	if (p_sw_port1->port_num < p_sw_port2->port_num)
		return -1;
	if (p_sw_port1->port_num > p_sw_port2->port_num)
		return 1;

	return 0;
}

/*
 * Call after partition sync (before laas rerouting process)
 *
 * File is in format:
 * 'guid:port_num pkey' (use 15bits pkey, avoid the membership bit)
 *
 * NVL DB port2nvl_prtn holds only switch trunk ports,
 * as the purpose of it is to allow consistency on chosen LAAS ports,
 * while members list of partition (that holds the GPUs and access ports) is recreated on each pkey sweep.
 */
ib_api_status_t osm_nvl_prtn_init(IN osm_nvlink_mgr_t * p_mgr)
{
	osm_opensm_t *p_osm = p_mgr->p_subn->p_osm;
	osm_log_t *p_log = &p_osm->log;
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(p_log);

	cl_qmap_init(&p_mgr->nvl_prtn_tbl);
	cl_fmap_init(&p_mgr->nvl_admin_states, nvl_prtn_compare_sw_ports);

	/* we initialize and restore the db domain of switch trunk port to nvlink partition  map */
	p_mgr->p_port2nvl_prtn = osm_db_domain_init(&p_osm->db, "port2nvl_prtn");
	if (!p_mgr->p_port2nvl_prtn) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BB08: "
			"Error initializing Guid-to-Lid persistent database\n");
		status = IB_ERROR;
		goto Exit;
	}

	if (osm_db_restore(p_mgr->p_port2nvl_prtn)) {
#ifndef __WIN__
		/*
		 * When Windows is BSODing, it might corrupt files that
		 * were previously opened for writing, even if the files
		 * are closed, so we might see corrupted port2nvl_prtn file.
		 */
		if (p_mgr->p_subn->opt.exit_on_fatal) {
			osm_log_v2(p_log, OSM_LOG_SYS, FILE_ID,
				   "FATAL: Error restoring port2nvl_prtn "
				   "persistent database\n");
			status = IB_ERROR;
			goto Exit;
		} else
#endif
			OSM_LOG(p_log, OSM_LOG_ERROR,
				"ERR BB09: Error restoring port2nvl_prtn "
				"persistent database\n");
	}

Exit:
	OSM_LOG_EXIT(p_log);
	return status;
}

static int nvl_prtn_compare_link(const void *lhs, const void *rhs) {
	const osm_physp_t * l_physp = ((osm_nvlink_link_t *)lhs)->p_physp;
	const osm_physp_t * r_physp = ((osm_nvlink_link_t *)rhs)->p_physp;
	uint64_t l_guid = osm_node_get_node_guid_ho(l_physp->p_node);
	uint64_t r_guid = osm_node_get_node_guid_ho(r_physp->p_node);
	uint8_t r_port = osm_physp_get_port_num(r_physp);
	uint8_t l_port = osm_physp_get_port_num(l_physp);

	if (l_guid < r_guid)
		return -1;
	if (l_guid > r_guid)
		return 1;
	if (l_port < r_port)
		return -1;
	if (l_port > r_port)
		return 1;
	return 0;
}


static int nvlink_get_nvl_partition_links_type(IN struct osm_subn * p_subn, IN osm_physp_t * p_physp)
{
	osm_physp_t * p_remote_physp;
	boolean_t is_trunk;
	boolean_t is_fnm;

	/* Notes:
	 * 1. This function implementation support only switches.
	 * 2. This function should be called only after SM configured which port is
	 *    'access' and which port is 'trunk'.
	 * 3. This function assumes SM did not set nvl_laas_pkey on FNM ports.
	 * */
	OSM_ASSERT(osm_node_get_type(p_physp->p_node) == IB_NODE_TYPE_SWITCH);

	is_trunk = osm_physp_is_trunk(p_subn, p_physp);
	is_fnm = osm_physp_is_fnm(p_subn, p_physp);

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (p_remote_physp) {
		if (osm_node_get_type(p_remote_physp->p_node) == IB_NODE_TYPE_SWITCH) {
			OSM_ASSERT(is_trunk);
			return IB_NVLINK_LINK_TRUNK;
		}

		if (osm_node_is_gpu(p_remote_physp->p_node)) {
			OSM_ASSERT(!is_fnm);
			return IB_NVLINK_LINK_ACCESS_SW;
		}

		OSM_ASSERT(FALSE);
		return IB_NVLINK_LINK_UNDEFINED;
	}

	/* We do not have remote port, check according to local physp.. */
	if (is_trunk) {
		return IB_NVLINK_LINK_TRUNK;
	} else if (!is_fnm) {
		return IB_NVLINK_LINK_ACCESS_SW;
	}

	OSM_ASSERT(FALSE);
	return IB_NVLINK_LINK_UNDEFINED;
}

osm_nvlink_link_t * osm_nvl_prtn_get_links(osm_subn_t * p_subn, uint16_t pkey)
{
	osm_log_t *p_log = &p_subn->p_osm->log;
	cl_map_item_t * p_item;
	osm_switch_t * p_sw;
	osm_physp_t * p_physp;
	osm_physp_t * p_remote_physp;
	osm_nvlink_link_t * links = NULL;
	ib_net16_t pkey_no;
	uint8_t port;
	uint8_t type;
	int n;

	OSM_LOG_ENTER(p_log);

	OSM_ASSERT(pkey);
	pkey_no = cl_hton16(pkey & 0x7fff);

	/* Count number of links to report */
	n = 0;
	for (p_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *) p_item;

		for (port = 1; port < p_sw->num_ports; port++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port);
			p_remote_physp = osm_physp_get_remote(p_physp);
			if (p_physp->nvl_laas_pkey != pkey_no)
				continue;
			type = nvlink_get_nvl_partition_links_type(p_subn, p_physp);
			/* Add one for the GPU */
			if (type == IB_NVLINK_LINK_ACCESS_SW && p_remote_physp)
				n++;
			if (type == IB_NVLINK_LINK_ACCESS_SW || type == IB_NVLINK_LINK_TRUNK)
				n++;
		}
	}

	/* Allocate additional element to indicates end of the vector */
	links = (osm_nvlink_link_t *) calloc(n + 1, sizeof(osm_nvlink_link_t));
	if (!links) {
		OSM_LOG(p_log, OSM_LOG_ERROR,
			"ERR BB51: Failed to allocate memory\n");
		goto exit;
	}

	/* Fill the vector */
	n = 0;
	for (p_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *) p_item;

		for (port = 1; port < p_sw->num_ports; port++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port);
			p_remote_physp = osm_physp_get_remote(p_physp);
			if (p_physp->nvl_laas_pkey != pkey_no)
				continue;
			type = nvlink_get_nvl_partition_links_type(p_subn, p_physp);

			if (type == IB_NVLINK_LINK_ACCESS_SW && p_remote_physp) {
				links[n].p_physp = p_remote_physp;
				links[n].type = IB_NVLINK_LINK_ACCESS_GPU;
				n++;
			}

			if (type == IB_NVLINK_LINK_TRUNK || type == IB_NVLINK_LINK_ACCESS_SW) {
				links[n].p_physp = p_physp;
				links[n].type = type;
				n++;
			}
		}
	}

	qsort(links, n, sizeof(osm_nvlink_link_t), nvl_prtn_compare_link);

exit:
	OSM_LOG_EXIT(p_log);

	return links;
}

void osm_nvl_prtn_release_links(osm_nvlink_link_t * links)
{
	free(links);
}

boolean_t osm_nvl_prtn_is_trunk_in_cnd(osm_nvlink_mgr_t * p_mgr, osm_physp_t * p_physp)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	ib_contain_and_drain_port_state_t *p_subn_cnd_port_state;
	osm_switch_t *p_sw;
	ib_cnd_port_state_t in_cnd_state, e_cnd_state;
	uint8_t port_num;

	if (!p_physp)
		return FALSE;

	p_sw = p_physp->p_node->sw;
	if (!p_sw)
		return FALSE;

	p_subn_cnd_port_state = &p_sw->nvlink_sw_data.subn_cnd_port_state;
	if (!osm_physp_is_trunk(p_subn, p_physp))
		return FALSE;

	port_num = p_physp->port_num;
	in_cnd_state = ib_contain_and_drain_port_state_get_ingress_port_state(p_subn_cnd_port_state, port_num);
	e_cnd_state = ib_contain_and_drain_port_state_get_egress_port_state(p_subn_cnd_port_state, port_num);
	if (in_cnd_state == IB_CND_PORT_STATE_SET || e_cnd_state == IB_CND_PORT_STATE_SET)
		return TRUE;

	return FALSE;
}

static ib_rail_filter_config_t * nvl_prtn_get_sw_calc_rail_filter_mad(osm_nvlink_prtn_t * p_prtn,
								      osm_switch_t * p_sw,
								      uint8_t rail,
								      uint8_t ingress_block,
								      uint8_t egress_block)
{
	ib_net64_t sw_guid;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;

	/* On first phase, support here a single RailFilterConfig block (switch num ports < 128) */
	OSM_ASSERT(!ingress_block && !egress_block);
	OSM_ASSERT(rail < OSM_NVLINK_SN_NUM_RAILS);

	sw_guid = osm_node_get_node_guid(p_sw->p_node);
	p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_get(&p_prtn->sw_tbl, sw_guid);
	if (!p_prtn_sw_data)
		return NULL;

	return &p_prtn_sw_data->calc_rail_filter[rail];
}

void osm_nvl_prtn_init_sw_subn_rail_filter(osm_nvlink_mgr_t * p_mgr, osm_switch_t * p_sw)

{
	cl_map_item_t *p_next;
	osm_nvlink_prtn_t *p_prtn;
	ib_net64_t sw_guid;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;
	uint8_t rail;

	sw_guid = osm_node_get_node_guid(p_sw->p_node);

	p_next = cl_qmap_head(&p_mgr->nvl_prtn_tbl);
	while (p_next != cl_qmap_end(&p_mgr->nvl_prtn_tbl)) {
		p_prtn = (osm_nvlink_prtn_t *) p_next;
		p_next = cl_qmap_next(&p_prtn->map_item);

		p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_get(&p_prtn->sw_tbl, sw_guid);
		if (!p_prtn_sw_data)
			continue;

		for (rail = 0; rail < OSM_NVLINK_SN_NUM_RAILS; rail++)
			memset(&p_prtn_sw_data->subn_rail_filter[rail], 0, sizeof(ib_rail_filter_config_t));
	}
}

static void nvl_prtn_rail_filter_remove_port_num_from_rail(osm_nvlink_mgr_t * p_mgr,
							   osm_nvlink_prtn_t * p_prtn,
							   osm_switch_t * p_sw,
							   uint8_t port_num,
							   uint8_t rail)
{
	ib_rail_filter_config_t *p_rail_filter_cfg;
	osm_physp_t * p_physp;
	uint8_t max_block, port_in_block, port_e_block;
	uint8_t ingress_block, egress_block;
	osm_log_t *p_log = &p_mgr->p_subn->p_osm->log;
	boolean_t is_add_req = OSM_NVL_REMOVE_PORT;

	if (!p_sw->nvlink_sw_data.rail_used_mask) {
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"No rails assigned for ports of switch 0x%"PRIx64". "
			"Skip updating port %u to RailFilterConfig\n",
			osm_node_get_node_guid_ho(p_sw->p_node), port_num);
		return;
	}

	/* On first phase, support here a single RailFilterConfig block (switch num ports < 128) */
	max_block = 1;

	port_in_block = port_num / IB_RAIL_FILTER_CFG_INGRESS_MASK_LEN;
	port_e_block = port_num / IB_RAIL_FILTER_CFG_EGRESS_MASK_LEN;

	/*
	 * Determine which rail id to update
	 * If the port does not have assigned rail, update rails used by the switch
	 */
	p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
	OSM_ASSERT(!ib_mlnx_epi_get_is_fnm_port(&p_physp->ext_port_info));

	if (!osm_nvlink_switch_is_rail_used(&p_sw->nvlink_sw_data, rail))
		return;

	/* Set rail egress port masks across all MADs with relevant egress mask block */
	for (ingress_block = 0; ingress_block < max_block; ingress_block++) {
		p_rail_filter_cfg = nvl_prtn_get_sw_calc_rail_filter_mad(
					p_prtn, p_sw, rail, ingress_block, port_e_block);
		if (!p_rail_filter_cfg)
			continue;

		osm_nvlink_rail_filter_config_mark_port_mask(
		    p_rail_filter_cfg, IB_PORT_TYPE_EGRESS, port_num, port_e_block, is_add_req);
	}

	/* Set rail ingress port masks across all MADs with relevant ingress mask block */
	for (egress_block = 0; egress_block < max_block; egress_block++) {
		p_rail_filter_cfg = nvl_prtn_get_sw_calc_rail_filter_mad(
					p_prtn, p_sw, rail, port_in_block, egress_block);
		if (!p_rail_filter_cfg)
			continue;

		osm_nvlink_rail_filter_config_mark_port_mask(
		    p_rail_filter_cfg, IB_PORT_TYPE_INGRESS, port_num, port_in_block, is_add_req);
	}
}

static void nvl_prtn_rail_filter_update_port_num(osm_nvlink_mgr_t * p_mgr,
						 osm_nvlink_prtn_t * p_prtn,
						 osm_switch_t * p_sw,
						 uint8_t port_num,
						 boolean_t is_add_req)
{
	ib_rail_filter_config_t *p_rail_filter_cfg;
	osm_physp_t * p_physp;
	uint8_t max_block, port_in_block, port_e_block;
	uint8_t ingress_block, egress_block;
	uint8_t rail, min_rail, max_rail;
	osm_log_t *p_log = &p_mgr->p_subn->p_osm->log;
	uint16_t pkey;

	if (!p_sw->nvlink_sw_data.rail_used_mask) {
		OSM_LOG(p_log, OSM_LOG_DEBUG,
			"No rails assigned for ports of switch 0x%"PRIx64". "
			"Skip updating port %u to RailFilterConfig\n",
			osm_node_get_node_guid_ho(p_sw->p_node), port_num);
		return;
	}

	/* On first phase, support here a single RailFilterConfig block (switch num ports < 128) */
	max_block = 1;

	port_in_block = port_num / IB_RAIL_FILTER_CFG_INGRESS_MASK_LEN;
	port_e_block = port_num / IB_RAIL_FILTER_CFG_EGRESS_MASK_LEN;

	/*
	 * Determine which rail id to update
	 * If the port does not have assigned rail, update rails used by the switch
	 */
	p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
	OSM_ASSERT(!ib_mlnx_epi_get_is_fnm_port(&p_physp->ext_port_info));

	if (p_physp->rail == OSM_NVLINK_NO_RAIL) {
		/*
		 * If port is in port2nvl_prtn cache, it is trunk that is used for LAAS of active
		 * partition, thus should be kept keep in rail filter
		 */
		if (!ib_mlnx_epi_get_is_trunk_port(&p_physp->ext_port_info) &&
		    osm_db_port2pkey_get(p_mgr->p_port2nvl_prtn, osm_node_get_node_guid_ho(p_sw->p_node),
					 port_num, &pkey)) {
			OSM_LOG(p_log, OSM_LOG_DEBUG,
				"No rail assigned for access port %u of switch 0x%"PRIx64". "
				"Skip updating port to RailFilterConfig\n",
				port_num, osm_node_get_node_guid_ho(p_sw->p_node));
			return;
		}

		min_rail = 0;
		max_rail = OSM_NVLINK_SN_NUM_RAILS - 1;
	} else {
		OSM_ASSERT(osm_nvlink_switch_is_rail_used(&p_sw->nvlink_sw_data, p_physp->rail));
		min_rail = max_rail = p_physp->rail;
	}
	
	for (rail = min_rail; rail <= max_rail; rail++) {
		if (!osm_nvlink_switch_is_rail_used(&p_sw->nvlink_sw_data, rail))
			continue;

		/* Set rail egress port masks across all MADs with relevant egress mask block */
		for (ingress_block = 0; ingress_block < max_block; ingress_block++) {
			p_rail_filter_cfg = nvl_prtn_get_sw_calc_rail_filter_mad(
						p_prtn, p_sw, rail, ingress_block, port_e_block);
			if (!p_rail_filter_cfg)
				continue;

			osm_nvlink_rail_filter_config_mark_port_mask(
			    p_rail_filter_cfg, IB_PORT_TYPE_EGRESS, port_num, port_e_block, is_add_req);
		}

		/* Set rail ingress port masks across all MADs with relevant ingress mask block */
		for (egress_block = 0; egress_block < max_block; egress_block++) {
			p_rail_filter_cfg = nvl_prtn_get_sw_calc_rail_filter_mad(
						p_prtn, p_sw, rail, port_in_block, egress_block);
			if (!p_rail_filter_cfg)
				continue;

			osm_nvlink_rail_filter_config_mark_port_mask(
			    p_rail_filter_cfg, IB_PORT_TYPE_INGRESS, port_num, port_in_block, is_add_req);
		}
	}
}

static void inline nvl_prtn_delete_port_laas_pkey(osm_nvlink_mgr_t * p_mgr,
						  osm_physp_t * p_physp)
{
	uint64_t node_guid;

	OSM_ASSERT(p_physp);

	p_physp->nvl_laas_pkey = 0x0;
	node_guid = osm_node_get_node_guid_ho(p_physp->p_node);

	osm_db_port2pkey_delete(p_mgr->p_port2nvl_prtn, node_guid, p_physp->port_num);
}

static void nvl_prtn_remove_port_from_prtn(osm_subn_t * p_subn, osm_nvlink_prtn_t * p_prtn,
					   osm_physp_t * p_physp)
{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	OSM_ASSERT(p_physp);

	nvl_prtn_delete_port_laas_pkey(p_mgr, p_physp);

	if (p_physp->p_remote_physp)
		nvl_prtn_delete_port_laas_pkey(p_mgr, p_physp->p_remote_physp);

}

static void inline nvl_prtn_add_port_laas_pkey(osm_nvlink_mgr_t * p_mgr,
					       osm_physp_t * p_physp,
					       ib_net16_t pkey)
{
	uint64_t node_guid;

	OSM_ASSERT(p_physp);

	p_physp->nvl_laas_pkey = pkey;

	node_guid = osm_node_get_node_guid_ho(p_physp->p_node);

	if (osm_node_get_type(p_physp->p_node) == IB_NODE_TYPE_SWITCH && p_physp->p_remote_physp &&
	    osm_node_get_type(p_physp->p_remote_physp->p_node) == IB_NODE_TYPE_SWITCH)
		osm_db_port2pkey_set(p_mgr->p_port2nvl_prtn, node_guid, p_physp->port_num, cl_ntoh16(pkey));
}

static void nvl_prtn_add_port_to_prtn(osm_subn_t *p_subn, osm_nvlink_prtn_t * p_prtn,
				      osm_switch_t * p_sw, uint8_t port_num)
{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	osm_physp_t *p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
	if (!p_physp)
		return;

	nvl_prtn_add_port_laas_pkey(p_mgr, p_physp, p_prtn->pkey);

	if (p_physp->p_remote_physp)
		nvl_prtn_add_port_laas_pkey(p_mgr, p_physp->p_remote_physp, p_prtn->pkey);
}


static void nvl_prtn_sw_set_trunk_ports_cnd_state(osm_subn_t * p_subn,
						  osm_nvlink_prtn_t * p_prtn,
						  osm_switch_t * p_sw,
						  ib_cnd_port_state_t cnd_port_state)
{
	static const uint8_t ZEROARRAY[IB_CND_NUM_PORT_BLOCKS_PER_MAD_BLOCK / 2] = {0};
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	ib_contain_and_drain_port_state_t req_cnd_port_state;
	osm_nvlink_mad_context_t mad_context;
	osm_physp_t *p_physp;
	ib_api_status_t status;
	uint8_t attr_mod = 0, num_ports, port_num;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_CONTAIN_AND_DRAIN_SUPPORTED_BIT))
		goto Exit;

	/* memset 0 as "no change" for all ports */
	memset(&req_cnd_port_state, 0, sizeof(ib_contain_and_drain_port_state_t));

	/* Iterate over all ports of the switch and build ContainAndDrainPortState */
	num_ports = osm_node_get_num_physp(p_sw->p_node);
	for (port_num = 1; port_num < num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
		if (!p_physp || !osm_physp_is_trunk(p_subn, p_physp) ||
		    p_physp->nvl_laas_pkey != p_prtn->pkey)
			continue;

		ib_contain_and_drain_port_state_set_ingress_port_state(&req_cnd_port_state,
								       port_num, cnd_port_state);
		ib_contain_and_drain_port_state_set_egress_port_state(&req_cnd_port_state,
								      port_num, cnd_port_state);
	}

	if (!memcmp(&req_cnd_port_state, ZEROARRAY, IB_CND_NUM_PORT_BLOCKS_PER_MAD_BLOCK / 2)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Skip setting ContainAndDrainPortState for partition 0x%x to switch "
			"0x%"PRIx64", as all port states hold no change value\n",
			cl_ntoh16(p_prtn->pkey), cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}


	/* Send ContainAndDrainPortState */
	if (!memcmp(&req_cnd_port_state, &p_sw->nvlink_sw_data.subn_cnd_port_state,
		    sizeof(ib_contain_and_drain_port_state_t)) &&
	    !p_sw->need_update && !p_subn->need_update)
		goto Exit;

	/*
	 * At this point, supports only a single MAD block,
	 * as it holds all switch ports (num ports < 128)
	 */
	memset(&mad_context, 0, sizeof(mad_context));
	mad_context.node_guid = p_sw->p_node->node_info.node_guid;
	mad_context.mad_method = OSM_AR_REQ_TYPE_SET;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB03: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			osm_node_get_node_guid_ho(p_sw->p_node));
		goto Exit;
	}

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp),
			     p_physp, (const uint8_t *)&req_cnd_port_state,
			     sizeof(ib_contain_and_drain_port_state_t),
			     IB_MAD_ATTR_CONTAIN_AND_DRAIN_PORT_STATE, attr_mod, FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB04: "
			"%s ContainAndDrainPortState for partition 0x%x on switch"
			" 0x%" PRIx64 " failed (%s)\n",
			cnd_port_state == IB_CND_PORT_STATE_CLEAR ? "Clearing": "Setting",
			cl_ntoh16(p_prtn->pkey),
			osm_node_get_node_guid_ho(p_sw->p_node),
			ib_get_err_str(status));

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static uint8_t nvl_prtn_laas_get_max_trunk_links(osm_nvlink_prtn_t * p_prtn)
{
	return p_prtn->max_trunk_links;
}

static void nvl_prtn_laas_remove_trunk_ports(osm_subn_t *p_subn,
					     osm_nvlink_prtn_t * p_prtn,
					     osm_switch_t * p_sw,
					     uint8_t start_port)
{
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num;

	num_ports = osm_node_get_num_physp(p_sw->p_node);
	for (port_num = start_port; port_num < num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
		if (!p_physp)
			continue;

		if (!osm_physp_is_trunk(p_subn, p_physp))
			continue;

		/*
		 * Port is not of this partition, shouldn't clear it
		 * In case of p_physp->nvl_laas_pkey == 0, make sure clear the remote
		 * to avoid inconsistency on miswiring.
		 */
		if (p_physp->nvl_laas_pkey && p_physp->nvl_laas_pkey != p_prtn->pkey)
			continue;

		/* Delete laas_pkey from physp and its remote */
		nvl_prtn_remove_port_from_prtn(p_subn, p_prtn, p_physp);
	}
}

int cage_sort(const void *p1, const void *p2, void * context)
{
	int * cages = (int *)context;

	osm_physp_t *p_physp1 = (osm_physp_t *)p1;
	osm_physp_t *p_physp2 = (osm_physp_t *)p2;

	int p1_cage_avail = cages[osm_hierarchy_get_cage(p_physp1)];
	int p2_cage_avail = cages[osm_hierarchy_get_cage(p_physp2)];

	if (p1_cage_avail > p2_cage_avail)
		return -1;

	if (p2_cage_avail > p1_cage_avail)
		return 1;

	if (p_physp1->port_num < p_physp2->port_num)
		return -1;

	if (p_physp1->port_num > p_physp2->port_num)
		return 1;

	return 0;
}

/*
 * In case requested admin state could not be configured to subnet, port may still be in
 * admin state.
 * Thus, port is considered in admin state when it is either configured to admin state or
 * requested to be configured to admin state.
 */
static boolean_t nvl_prtn_is_port_in_admin_state(osm_nvlink_mgr_t * p_mgr, osm_physp_t * p_physp)
{
	uint8_t req_admin_state = nvl_prtn_get_port_req_admin_state(p_mgr, p_physp);
	uint8_t subn_admin_state = ib_mlnx_epi_get_nmx_admin_state(&p_physp->ext_port_info);

	return req_admin_state == IB_NMX_ADMINSTATE_DOWN || req_admin_state == IB_NMX_ADMINSTATE_DIAG ||
	    subn_admin_state == IB_NMX_ADMINSTATE_DOWN || subn_admin_state == IB_NMX_ADMINSTATE_DIAG;
}

static uint8_t nvl_prtn_laas_add_trunk_ports_to_prtn(osm_subn_t *p_subn,
						     osm_nvlink_prtn_t * p_prtn,
						     osm_switch_t * p_sw,
						     uint8_t num_ports_to_add)
{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num;
	uint8_t num_added_trunks = 0;
	int port_index = 0;
	int max_port_index = 0;
	int cage;
	int * cages = NULL;
	osm_physp_t **ports = NULL;

	num_ports = osm_node_get_num_physp(p_sw->p_node);

	/*
	 * Allocate memory to sort ports according to cages
	 * Note: we assume that cage number <= number of ports
	 */
	cages = (int *)calloc(num_ports, sizeof(int));
	ports = (osm_physp_t **)calloc(num_ports - 1, sizeof(osm_physp_t*));
	if (!cages || !ports) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB30: "
			"Failed to allocate memory\n");
		goto exit;
	}

	/* Update number of available ports in each cage */
	for (port_num = 1; port_num < num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
		cage = osm_hierarchy_get_cage(p_physp);

		OSM_ASSERT(cage < (int)num_ports);

		if (ib_port_info_get_port_state(&p_physp->port_info) != IB_LINK_ACTIVE)
			continue;

		if (!osm_physp_is_trunk(p_subn, p_physp))
			continue;

		if (!p_subn->opt.nvl_cage_allocation_mode ) {
			if (!p_physp->nvl_laas_pkey)
				ports[max_port_index++] = p_physp;
			continue;
		}

		/* Gets here only when allocating partition by cage */

		/*
		 * Skip ports with invalid cage.
		 * In case cage was not received from HierarchyInfo, its value is -1.
		 */
		if (cage < 0 || cage >= (int)num_ports) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB31: "
				"Found illegal cage number %d for switch 0x%"PRIx64" port %u. "
				"Skips this port on partition trunk allocation\n",
				 cage, osm_node_get_node_guid_ho(p_sw->p_node), p_physp->port_num);
			continue;
		}

		if (p_physp->nvl_laas_pkey && p_physp->nvl_laas_pkey != p_prtn->pkey) {
			cages[cage] = -1;
			continue;
		}

		if (cages[cage] < 0)
			continue;

		cages[cage]++;
		ports[max_port_index++] = p_physp;
	}

	/*
	 * In case single partition per cage is enforced,
	 * sort ports array according to number of available ports in each cage.
	 */
	if (p_subn->opt.nvl_cage_allocation_mode)
		qsort_r(ports, max_port_index, sizeof(osm_physp_t*), cage_sort, cages);

	for (port_index = 0; port_index < max_port_index; port_index++) {
		p_physp = ports[port_index];
		port_num = p_physp->port_num;
		cage = osm_hierarchy_get_cage(p_physp);

		/* Check that the port is not in a cage that assigned to another partition */
		if (p_subn->opt.nvl_cage_allocation_mode && cages[cage] < 1)
			continue;

		if (nvl_prtn_is_port_in_admin_state(p_mgr, p_physp) ||
		    nvl_prtn_is_port_in_admin_state(p_mgr, p_physp->p_remote_physp))
			continue;

		/* port is added to partition */
		num_added_trunks++;

		nvl_prtn_add_port_to_prtn(p_subn, p_prtn, p_sw, port_num);

		if (num_added_trunks >= num_ports_to_add)
			break;
	}

exit:
	free(cages);
	free(ports);

	return num_added_trunks;
}

static void nvl_prtn_init_rail_filter_vl_mc_uc(ib_rail_filter_config_t * p_rail_filter_cfg)
{
	p_rail_filter_cfg->vl_mask |= IB_RAIL_FILTER_CFG_NVLINK_VL_MASK;
	p_rail_filter_cfg->mc_uc_enable |= (IB_RAIL_FILTER_CFG_MC_ENABLE_BIT |
					    IB_RAIL_FILTER_CFG_UC_ENABLE_BIT);
}

static void nvl_prtn_init_rail_filter_mad(ib_rail_filter_config_t * p_rail_filter_cfg)
{
	memset(p_rail_filter_cfg, 0, sizeof(ib_rail_filter_config_t));

	nvl_prtn_init_rail_filter_vl_mc_uc(p_rail_filter_cfg);
}

static void nvl_prtn_laas_init_calc_rail_filter(osm_switch_t * p_sw,
						osm_nvlink_prtn_sw_data_t * p_prtn_sw_data)
{
	/* On first phase, support here a single RailFilterConfig block (switch num ports < 128) */
	
	uint8_t rail;
	for (rail = 0; rail < OSM_NVLINK_SN_NUM_RAILS; rail++)
		nvl_prtn_init_rail_filter_mad(&p_prtn_sw_data->calc_rail_filter[rail]);
}

static void nvl_prtn_laas_pkey_configuration(osm_subn_t *p_subn, osm_nvlink_prtn_t * p_prtn,
					     osm_switch_t * p_sw,
					     osm_nvlink_prtn_sw_data_t * p_prtn_sw_data)
{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num, start_port_num;
	uint8_t max_num_trunks, cur_num_trunks;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/* NOTE: Currently supports two switches per plane */

	/*
	 * Count the active trunk links that are used for this partition
	 * (SM is allowed to reuse ports that were marked in C&D. Hence no need to check C&D state.
	 *  Also, a this point all partition is at C&D in order to allow reroute)
	 */
	start_port_num = 1;
	cur_num_trunks = 0;
	max_num_trunks = nvl_prtn_laas_get_max_trunk_links(p_prtn);
	num_ports = osm_node_get_num_physp(p_sw->p_node);
	for (port_num = 1; port_num < num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
		if (!p_physp)
			continue;

		if (p_physp->nvl_laas_pkey != p_prtn->pkey)
			continue;

		/*
		 * In case single partition per cage is enforced, remove all previously
		 * configured ports so SM re-choose, in order to minimize cages usage
		 */
		if (p_subn->opt.nvl_cage_allocation_mode) {
			nvl_prtn_remove_port_from_prtn(p_subn, p_prtn, p_physp);
			continue;
		}

		if (!osm_physp_get_remote(p_physp)) {
			/* Delete laas_pkey from physp and its remote */
			nvl_prtn_remove_port_from_prtn(p_subn, p_prtn, p_physp);
			continue;
		}

		/*
		 * For access ports, pkey is updated at the point of add_gpu.
		 * Hence, only trunks laas_pkey is updated here.
		*/
		if (osm_node_is_gpu(p_physp->p_remote_physp->p_node) ||
		    !osm_physp_is_trunk(p_subn, p_physp))
			continue;

		/* If only one side of the link was set to admin state, cannot use it for LAAS */
		if (nvl_prtn_is_port_in_admin_state(p_mgr, p_physp) ||
		    nvl_prtn_is_port_in_admin_state(p_mgr, p_physp->p_remote_physp)) {
			/* Deletes laas_pkey from physp and its remote */
			nvl_prtn_remove_port_from_prtn(p_subn, p_prtn, p_physp);
			continue;
		}

		/* Trunk port is part of partition, add to partition cache and to rail filter */
		if (cur_num_trunks < max_num_trunks) {
			nvl_prtn_add_port_to_prtn(p_subn, p_prtn, p_sw, port_num);
			start_port_num = port_num;
		}

		cur_num_trunks++;
	}

	/* number of valid used trunks is as required. No need to update partition trunk ports */
	if (cur_num_trunks == max_num_trunks) {
		p_prtn->num_trunk_used = max_num_trunks;
		goto Exit;
	}

	/* Remove redundant trunks from the partition, so can be free to use for other partitions */
	if (cur_num_trunks > max_num_trunks) {
		nvl_prtn_laas_remove_trunk_ports(p_subn, p_prtn, p_sw, start_port_num + 1);
		p_prtn->num_trunk_used = max_num_trunks;
		goto Exit;
	}

	/*
	 * If got here, it means cur_num_trunks < max_num_trunks.
	 * Search for free trunk ports and add required number of trunks to the partition.
	 * Note: there are not necessarily enough free trunks to get to max_num_trunks.
	 */
	cur_num_trunks += nvl_prtn_laas_add_trunk_ports_to_prtn(p_subn, p_prtn, p_sw,
								max_num_trunks - cur_num_trunks);
	p_prtn->num_trunk_used = cur_num_trunks;

Exit:
	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
		"Switch 0x%"PRIx64" is set with %u trunk links for partition 0x%x\n",
		osm_node_get_node_guid_ho(p_sw->p_node),
		p_prtn->num_trunk_used, cl_ntoh16(p_prtn->pkey));

	if (p_prtn->num_trunk_used < nvl_prtn_laas_get_max_trunk_links(p_prtn) &&
	    p_subn->opt.nvl_cage_allocation_mode)
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
			"Switch 0x%"PRIx64" could not be configured with requested max trunk "
			"links %u for partition 0x%x. Note nvl_cage_allocation_mode is %u, meaning "
			"each cage trunks must not be allocated to different partitions\n",
			osm_node_get_node_guid_ho(p_sw->p_node),
			nvl_prtn_laas_get_max_trunk_links(p_prtn),
			cl_ntoh16(p_prtn->pkey), p_subn->opt.nvl_cage_allocation_mode);


	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static ib_api_status_t nvl_prtn_add_switch_to_laas(osm_subn_t * p_subn,
						   osm_nvlink_prtn_t * p,
						   osm_switch_t * p_sw)
{
	osm_nvlink_prtn_sw_data_t *p_prtn_sw;
	ib_net64_t sw_guid;

	p_prtn_sw = (osm_nvlink_prtn_sw_data_t *)calloc(1, sizeof(osm_nvlink_prtn_sw_data_t));
	if (!p_prtn_sw)
		return IB_INSUFFICIENT_MEMORY;

	sw_guid = osm_node_get_node_guid(p_sw->p_node);
	if (cl_map_insert(&p->sw_tbl, sw_guid, p_prtn_sw) == NULL) {
		free(p_prtn_sw);
		return IB_INSUFFICIENT_MEMORY;
	}

	return IB_SUCCESS;
}

static ib_api_status_t nvl_prtn_add_remote_switch_to_laas(osm_subn_t * p_subn,
							  osm_nvlink_prtn_t * p,
							  osm_physp_t * p_physp)
{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	osm_physp_t *p_remote_physp;
	osm_switch_t *p_remote_sw;
	ib_net64_t sw_guid;
	ib_api_status_t status;

	/*
	 * Only end ports are supported as members.
	 * Switch ports are marked for LAAS routing, but are not actual members.
	 */
	p_remote_physp = osm_physp_get_remote(p_physp);
	if (!p_remote_physp)
		return IB_SUCCESS;
	p_remote_sw = p_remote_physp->p_node->sw;
	if (osm_node_get_type(p_remote_physp->p_node) != IB_NODE_TYPE_SWITCH || !p_remote_sw)
		return IB_SUCCESS;

	sw_guid = osm_node_get_node_guid(p_remote_physp->p_node);

	/* If switch is not part of partition yet, create item and add it to switch table */
	if (!cl_map_get(&p->sw_tbl, sw_guid)) {
		status = nvl_prtn_add_switch_to_laas(p_subn, p, p_remote_sw);
		if (status)
			return status;
	}

	/*
	 * Add switch remote port to partition only if it is an access port, as such ports are
	 * always part of the partition, while trunk ports of partition are chosen on reroute step.
	 */
	if (!osm_physp_is_trunk(p_subn, p_remote_physp)) {
		nvl_prtn_add_port_to_prtn(p_subn, p, p_remote_sw, p_remote_physp->port_num);
		nvl_prtn_rail_filter_update_port_num(p_mgr, p, p_remote_sw,
						     p_remote_physp->port_num, OSM_NVL_ADD_PORT);
	}

	return IB_SUCCESS;
}

boolean_t osm_nvl_prtn_is_trunk_of_partition(osm_subn_t * p_subn, osm_nvlink_prtn_t * p_prtn,
					     ib_net64_t sw_guid, uint8_t port_num)
{
	osm_switch_t *p_sw;
	osm_physp_t *p_physp;

	p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
	if (!p_sw)
		return FALSE;

	p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
	if (!p_physp || !osm_physp_is_trunk(p_subn, p_physp))
		return FALSE;

	if (p_physp->nvl_laas_pkey != p_prtn->pkey)
		return FALSE;

	return TRUE;
}

static ib_api_status_t nvl_prtn_validate_cnd_port_state(osm_subn_t * p_subn, osm_nvlink_prtn_t * p_prtn,
							ib_cnd_port_state_t req_cnd_state)
{
	cl_map_iterator_t p_next;
	osm_switch_t *p_sw;
	osm_physp_t *p_physp;
	uint8_t num_ports, port_num;
	ib_net64_t sw_guid;
	ib_cnd_port_state_t cnd_state;
	ib_contain_and_drain_port_state_t *p_subn_cnd_port_state;
	boolean_t is_valid = TRUE;

	/*
	 * Iterate over partition switches' trunk ports and validate that all
	 * Contain and Drain port states are set to req_cnd_state.
	 * If not, return error.
	 */

	p_next = cl_map_head(&p_prtn->sw_tbl);
	while (p_next != cl_map_end(&p_prtn->sw_tbl)) {
		sw_guid = cl_map_key(p_next);
		p_next = cl_map_next(p_next);

		p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
		if (!p_sw)
			continue;

		p_subn_cnd_port_state = &p_sw->nvlink_sw_data.subn_cnd_port_state;
		num_ports = osm_node_get_num_physp(p_sw->p_node);
		for (port_num = 1; port_num < num_ports; port_num++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
			if (!p_physp || !osm_physp_is_trunk(p_subn, p_physp))
				continue;

			if (!osm_nvl_prtn_is_trunk_of_partition(p_subn, p_prtn, sw_guid, port_num))
				continue;

			cnd_state = ib_contain_and_drain_port_state_get_egress_port_state(
					p_subn_cnd_port_state, port_num);
			if (cnd_state != req_cnd_state) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
					"Switch 0x%"PRIx64" trunk port num %u of partition 0x%x "
					"is set to egress ContainAndDrain state %u instead of %u\n",
					osm_node_get_node_guid_ho(p_sw->p_node),
					port_num, cl_ntoh16(p_prtn->pkey),
					cnd_state, req_cnd_state);
				is_valid = FALSE;
			}

			cnd_state = ib_contain_and_drain_port_state_get_ingress_port_state(
					p_subn_cnd_port_state, port_num);
			if (cnd_state != req_cnd_state) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
					"Switch 0x%"PRIx64" trunk port num %u of partition 0x%x "
					"is set to ingress ContainAndDrain state %u instead of %u\n",
					osm_node_get_node_guid_ho(p_sw->p_node),
					port_num, cl_ntoh16(p_prtn->pkey),
					cnd_state, req_cnd_state);
				is_valid = FALSE;
			}
		}
	}

	if (!is_valid)
		return IB_ERROR;

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
		"All partition 0x%x ports are set to Contain And Drain %s\n",
		cl_ntoh16(p_prtn->pkey),
		req_cnd_state == IB_CND_PORT_STATE_CLEAR ? "Cleared": "set");

	return IB_SUCCESS;
}

static ib_api_status_t nvl_prtn_reroute_partition_clear_cnd(osm_subn_t * p_subn, osm_nvlink_prtn_t * p_prtn)
{
	ib_net64_t sw_guid;
	cl_map_iterator_t p_next;
	osm_switch_t *p_sw;

	/*
	 * Clear contain and drain port state from all partition trunk ports
	 * by ContainAndDrainPortState SET.
	 *
	 * NOTE!
	 * There might be other ports on ContainAndDrainPortState MAD with state CnD that
	 * are not part of this partition reroute request.
	 * Hence, must clear only the ports of this partition!
	 */

	p_next = cl_map_head(&p_prtn->sw_tbl);
	while (p_next != cl_map_end(&p_prtn->sw_tbl)) {
		sw_guid = cl_map_key(p_next);
		p_next = cl_map_next(p_next);

		p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
		if (!p_sw)
			continue;

		/*
		 * Shouldn't get here. In case CnDInfo failed, partition should've already been
		 * marked in error state by nvl_prtn_sw_rerouting_preprocess
		 */
		OSM_ASSERT(p_sw->nvlink_sw_data.cnd_info_updated);
		if (!p_sw->nvlink_sw_data.cnd_info_updated) {
			OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB24: "
				"ContainAndDrainInfo was not set successfully for switch"
				" 0x%"PRIx64". Set partition 0x%x to error state\n",
				cl_ntoh64(p_sw->p_node->node_info.node_guid),
				cl_ntoh16(p_prtn->pkey));

			p_prtn->state = NVL_PRTN_STATE_ERROR;
			return IB_ERROR;
		}

		nvl_prtn_sw_set_trunk_ports_cnd_state(p_subn, p_prtn, p_sw, IB_CND_PORT_STATE_CLEAR);
	}

	/* On ContainAndDrainPortState receive, set partition status to maintained */
	if (wait_for_pending_transactions(p_subn->p_osm))
		return IB_ERROR;

	return IB_SUCCESS;
}

static ib_api_status_t nvl_prtn_validate_prtn_subn_laas_ports(osm_subn_t * p_subn, osm_nvlink_prtn_t * p_prtn)
{
	static const uint8_t ZEROARRAY[IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN] = {0};
	cl_map_iterator_t p_next;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;
	osm_switch_t *p_sw;
	uint64_t sw_guid;
	uint8_t rail;

	/*
	 * Iterate over partition switches, validating routing was set successfully by
	 * comparing calculated RailFilterConfig to latest one received from subnet.
	 */

	p_next = cl_map_head(&p_prtn->sw_tbl);
	while (p_next != cl_map_end(&p_prtn->sw_tbl)) {
		p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_obj(p_next);
		sw_guid = cl_map_key(p_next);
		p_next = cl_map_next(p_next);

		p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
		if (!p_sw)
			continue;

		for (rail = 0; rail < OSM_NVLINK_SN_NUM_RAILS; rail++) {
			if (!osm_nvlink_switch_is_rail_used(&p_sw->nvlink_sw_data, rail))
				continue;

			/*
			 * In case all rail ports were removed from the partition (for example
			 * due to disconnections), it is used but empty.
			 */
			if (!memcmp(&p_prtn_sw_data->calc_rail_filter[rail].ingress_port_mask, ZEROARRAY,
			    IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN))
				continue;

			if (memcmp(&p_prtn_sw_data->subn_rail_filter[rail],
				   &p_prtn_sw_data->calc_rail_filter[rail],
				   sizeof(ib_rail_filter_config_t)))
				return IB_ERROR;

			p_prtn_sw_data->is_laas_set = TRUE;
		}
	}

	return IB_SUCCESS;
}

static void nvl_prtn_reroute_finish_prtn(osm_subn_t * p_subn, osm_nvlink_prtn_t * p_prtn)
{
	uint32_t sleep_time_millisec = p_subn->opt.nvl_partition_reroute_wait * OSM_STO_MILLISEC;

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
		"Start sleeping for: %u milliseconds before reactivating partition 0x%x\n",
		sleep_time_millisec, cl_ntoh16(p_prtn->pkey));

	usleep(sleep_time_millisec * 1000); /* usleep expects input in microsec */

	/* Mark all partition ports to CnD state ‘clear’ */
	if (nvl_prtn_reroute_partition_clear_cnd(p_subn, p_prtn) ||
	    nvl_prtn_validate_cnd_port_state(p_subn, p_prtn, IB_CND_PORT_STATE_CLEAR)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB05: "
			"Could not clear ContainAndDrain of all the switches "
			"of partition 0x%x. Aborting partition reroute\n",
			cl_ntoh16(p_prtn->pkey));

		p_prtn->state = NVL_PRTN_STATE_ERROR;
		return;
	}

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
		"Done processing partition reroute request of partition 0x%x\n",
		cl_ntoh16(p_prtn->pkey));

	p_prtn->state = NVL_PRTN_STATE_MAINTAIN;
}

/*
 * Call once ALID Routing and LAAS setting is done.
 * This function clears Contain and Drain state from partitions that were rerouted
 * and set them to maintain state.
 */
void osm_nvl_prtn_reroute_finish(osm_subn_t * p_subn)
{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	cl_map_item_t *p_next;
	osm_nvlink_prtn_t *p_prtn;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	p_next = cl_qmap_head(&p_mgr->nvl_prtn_tbl);
	while (p_next != cl_qmap_end(&p_mgr->nvl_prtn_tbl)) {
		p_prtn = (osm_nvlink_prtn_t *) p_next;
		p_next = cl_qmap_next(&p_prtn->map_item);

		if (p_prtn->state != NVL_PRTN_STATE_REROUTE)
			continue;

		nvl_prtn_reroute_finish_prtn(p_subn, p_prtn);
	}

	osm_db_store(p_mgr->p_port2nvl_prtn, p_subn->opt.fsync_high_avail_files);

	/* Updates all GPUs that are new to partition as members */
	cl_qmap_apply_func(&p_subn->gpu_guid_tbl, nvl_prtn_update_new_gpu_to_member_status, NULL);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static void nvl_prtn_sw_rerouting_preprocess(osm_subn_t * p_subn, osm_nvlink_prtn_t * p_prtn, osm_switch_t * p_sw)
{
	/* Set all partition trunk ports to ContainAndDrain port state before rerouting */
	if (p_prtn->state == NVL_PRTN_STATE_REROUTE_REQUESTED)
		nvl_prtn_sw_set_trunk_ports_cnd_state(p_subn, p_prtn, p_sw, IB_CND_PORT_STATE_SET);
}

void osm_nvl_prtn_reroute_preprocess(osm_subn_t * p_subn, osm_nvlink_prtn_t * p_prtn)
{
	cl_map_iterator_t p_next;
	ib_net64_t sw_guid;
	osm_switch_t *p_sw;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
		"Processing partition reroute request for partition 0x%x\n", cl_ntoh16(p_prtn->pkey));

	/* At this point should configure according to latest trunks (from cache) */

	p_next = cl_map_head(&p_prtn->sw_tbl);
	while (p_next != cl_map_end(&p_prtn->sw_tbl)) {
		sw_guid = cl_map_key(p_next);
		p_next = cl_map_next(p_next);

		p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
		if (!p_sw)
			continue;

		/* Mark switch is allowed for rerouting and set partition ports to Contain and Drain */
		nvl_prtn_sw_rerouting_preprocess(p_subn, p_prtn, p_sw);
	}

	/*
	 * If partition is already marked to reroute (new partition),
	 * skips setting Contain and Drain port states,
	 */
	if (p_prtn->state == NVL_PRTN_STATE_REROUTE)
		goto Exit;

	/*
	 * Must not move forward to rerouting step if not all partition ports are set
	 * to ContainAndDrain state.
	 */
	if (wait_for_pending_transactions(p_subn->p_osm) ||
	    nvl_prtn_validate_cnd_port_state(p_subn, p_prtn, IB_CND_PORT_STATE_SET)) {
		p_prtn->state = NVL_PRTN_STATE_ERROR;
		goto Exit;
	}

	p_prtn->state = NVL_PRTN_STATE_REROUTE;

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static void nvl_prtn_build_used_pkeys_list(IN osm_subn_t * p_subn,
					   OUT uint16_t * pkeys,
					   IN uint16_t num_pkeys)
{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	cl_map_item_t *p_next;
	osm_nvlink_prtn_t *p_prtn;
	int i = 0;

	OSM_ASSERT(cl_qmap_count(&p_mgr->nvl_prtn_tbl) <= num_pkeys);

	p_next = cl_qmap_head(&p_mgr->nvl_prtn_tbl);
	while (p_next != cl_qmap_end(&p_mgr->nvl_prtn_tbl)) {
	       p_prtn = (osm_nvlink_prtn_t *) p_next;
	       p_next = cl_qmap_next(p_next);
		pkeys[i++] = cl_ntoh16(p_prtn->pkey);
	}
}

static void nvl_prtn_preprocess_validate_cache(osm_subn_t * p_subn)
{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	uint16_t *pkeys;
	uint16_t num_pkeys;

	num_pkeys = cl_qmap_count(&p_mgr->nvl_prtn_tbl);
	pkeys = (uint16_t *)calloc(num_pkeys, sizeof(uint16_t));
	if (!pkeys)
		return;

	nvl_prtn_build_used_pkeys_list(p_subn, pkeys, num_pkeys);
	nvl_prtn_validate_port2nvl_prtn_db(p_mgr, pkeys, num_pkeys);

	free(pkeys);
}

void osm_nvl_prtn_preprocess(osm_subn_t * p_subn)
{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	cl_map_item_t *p_next;
	osm_nvlink_prtn_t *p_prtn;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/* Validate cache on each sweep, in case links are reconnected and found to be non-trunks */
	nvl_prtn_preprocess_validate_cache(p_subn);

	cl_qmap_apply_func(&p_subn->gpu_guid_tbl, nvl_prtn_update_deleted_gpu_member_status, p_mgr);

	/* Update trunk ports admin state */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, nvl_prtn_process_admin_state_ports, p_mgr);
	(void) wait_for_pending_transactions(p_subn->p_osm);

	p_next = cl_qmap_head(&p_mgr->nvl_prtn_tbl);
	while (p_next != cl_qmap_end(&p_mgr->nvl_prtn_tbl)) {
		p_prtn = (osm_nvlink_prtn_t *) p_next;
		p_next = cl_qmap_next(p_next);

		if (p_prtn->state != NVL_PRTN_STATE_REROUTE_REQUESTED &&
		    p_prtn->state != NVL_PRTN_STATE_REROUTE)
			continue;

		osm_nvl_prtn_reroute_preprocess(p_subn, p_prtn);
	}

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static osm_nvlink_prtn_sw_data_t * nvl_prtn_get_sw_prtn_data(IN osm_subn_t * p_subn,
							     IN osm_switch_t * p_sw,
							     IN ib_net16_t pkey,
							     OUT osm_nvlink_prtn_t **pp_prtn)
{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	cl_map_item_t *p_item;
	ib_net64_t sw_guid;
	osm_nvlink_prtn_t *p_prtn;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;

	p_item = cl_qmap_get(&p_mgr->nvl_prtn_tbl, pkey);
	if (p_item == cl_qmap_end(&p_mgr->nvl_prtn_tbl))
		return NULL;

	p_prtn = (osm_nvlink_prtn_t *)p_item;
	if (pp_prtn)
		*pp_prtn = p_prtn;

	sw_guid = osm_node_get_node_guid(p_sw->p_node);

	p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_get(&p_prtn->sw_tbl, sw_guid);
	if (!p_prtn_sw_data)
		return NULL;


	return p_prtn_sw_data;
}


void osm_nvl_prtn_cpy_subn_rail_filter_to_prtn(osm_subn_t * p_subn, osm_switch_t * p_sw,
					       ib_net16_t pkey, uint8_t rail, uint8_t ingress_block,
					       uint8_t egress_block,
					       ib_rail_filter_config_t * p_rail_filter)
{
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;

	/* On first phase, support here a single RailFilterConfig block (switch num ports < 128) */
	OSM_ASSERT(!ingress_block && !egress_block);
	OSM_ASSERT(rail < OSM_NVLINK_SN_NUM_RAILS);

	p_prtn_sw_data = nvl_prtn_get_sw_prtn_data(p_subn, p_sw, pkey, NULL);
	if (!p_prtn_sw_data)
		return;

	memcpy(&p_prtn_sw_data->subn_rail_filter[rail], p_rail_filter, sizeof(ib_rail_filter_config_t));
}

void osm_nvl_prtn_rail_filter_config_send_block(osm_subn_t * p_subn,
						osm_switch_t * p_sw,
						ib_rail_filter_config_t rail_filter,
						ib_net16_t pkey,
						uint8_t rail,
						uint8_t ingress_block,
						uint8_t egress_block,
						boolean_t zero_prev_ports)
{
	static const uint8_t ZEROARRAY[IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN] = {0};
	osm_sm_t *p_sm = &p_subn->p_osm->sm;
	osm_rail_filter_cfg_context_t context;
	osm_physp_t *p_physp0;
	uint32_t attr_mod;
	ib_api_status_t status;

	p_physp0 = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp0) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB01: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	if (!memcmp(rail_filter.ingress_port_mask, ZEROARRAY,
		    IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"RailFilterConfig with ingress port mask 0 is not allowed. "
			"Skip setting RailFilterConfig of partition 0x%x for switch 0x%"PRIx64"\n",
			cl_ntoh16(pkey), cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	/* Using ingress port mask, no need for port number on attr_mod */
	attr_mod = (ingress_block << IB_RAIL_FILTER_ATTR_MOD_INGRESS_BLOCK_SHIFT) |
	    (egress_block << IB_RAIL_FILTER_ATTR_MOD_EGRESS_BLOCK_SHIFT);

	memset(&context, 0, sizeof(context));
	context.node_guid = p_sw->p_node->node_info.node_guid;
	context.mad_method = IB_MAD_METHOD_SET;
	context.ingress_block = ingress_block;
	context.egress_block = egress_block;
	context.pkey = pkey;
	context.rail = rail;
	context.is_for_zero_prev_ports = zero_prev_ports;

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp0),
			     p_physp0, (const uint8_t *)&rail_filter,
			     sizeof(ib_rail_filter_config_t), IB_MAD_ATTR_RAIL_FILTER_CONFIG,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp0->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB02: "
			"Setting RailFilterConfig of partition 0x%x (ingress block %u, egress block %u) "
			"to 0x%" PRIx64 " failed (%s)\n",
			cl_ntoh16(pkey), ingress_block, egress_block,
			cl_ntoh64(p_physp0->port_guid), ib_get_err_str(status));
		return;
	}

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
		"Setting RailFilterConfig of partition 0x%x (ingress block %u, egress block %u) "
		"to 0x%" PRIx64 "\n",
		cl_ntoh16(pkey), ingress_block, egress_block, cl_ntoh64(p_physp0->port_guid));
}

void osm_nvl_prtn_send_rail_filter_of_prtn(osm_subn_t * p_subn, osm_switch_t * p_sw,
					   ib_net16_t pkey, uint8_t rail, uint8_t ingress_block,
					   uint8_t egress_block)
{
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;
	ib_rail_filter_config_t * p_req_rail_filter;
	osm_nvlink_prtn_t *p_prtn;

	/* On first phase, support here a single RailFilterConfig block (switch num ports < 128) */
	OSM_ASSERT(!ingress_block && !egress_block);

	p_prtn_sw_data = nvl_prtn_get_sw_prtn_data(p_subn, p_sw, pkey, &p_prtn);
	if (!p_prtn_sw_data)
		return;

	p_req_rail_filter =
	    nvl_prtn_get_sw_calc_rail_filter_mad(p_prtn, p_sw, rail, ingress_block, egress_block);
	if (!p_req_rail_filter)
		return;

	osm_nvl_prtn_rail_filter_config_send_block(p_subn, p_sw,
						   *p_req_rail_filter,
						   pkey, rail, ingress_block, egress_block,
						   OSM_NVL_PRTN_SET_NEW_RAIL_FILTER);
}

static void nvl_prtn_laas_update_rail_filter(osm_subn_t * p_subn, osm_nvlink_prtn_t * p_prtn,
					     osm_switch_t * p_sw, uint8_t rail,
					     osm_nvlink_prtn_sw_data_t * p_prtn_sw_data)
{
	ib_rail_filter_config_t req_rail_filter;
	uint8_t i;
	static const uint8_t ZEROARRAY[IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN] = {0};

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_RAIL_FIILTER_SUPPORTED_BIT))
		return;

	/* On first phase, support here a single RailFilterConfig block (switch num ports < 128) */

	/* Skip if RailFilter was not changed */
	if (!memcmp(&p_prtn_sw_data->subn_rail_filter[rail],
		    &p_prtn_sw_data->calc_rail_filter[rail],
		    sizeof(ib_rail_filter_config_t)))
		return;

	/* Build ingress ports mask that hold only the ports that were removed from partition */
	nvl_prtn_init_rail_filter_mad(&req_rail_filter);
	for (i = 0; i < IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN; i++) {
		req_rail_filter.ingress_port_mask[i] =
		    p_prtn_sw_data->subn_rail_filter[rail].ingress_port_mask[i] &
		    ~(p_prtn_sw_data->calc_rail_filter[rail].ingress_port_mask[i]);
	}

	if (memcmp(p_prtn_sw_data->subn_rail_filter[rail].ingress_port_mask,
		   p_prtn_sw_data->calc_rail_filter[rail].ingress_port_mask,
		   sizeof(p_prtn_sw_data->calc_rail_filter[rail].ingress_port_mask)) &&
	    memcmp(req_rail_filter.ingress_port_mask, ZEROARRAY, IB_RAIL_FILTER_CFG_INGRESS_MASK_BYTES_LEN)) {
		/* Ingress ports masks are different, zero rail filter for removed ports first */

		/* Setting zero egress ports to previous ingress port that were removed from partition */
		osm_nvl_prtn_rail_filter_config_send_block(p_subn, p_sw, req_rail_filter,
							   p_prtn->pkey, rail, 0, 0,
							   OSM_NVL_PRTN_ZERO_PREV_RAIL_FILTER);

		/*
		 * On receiving response to the zeroing MAD,
		 * send calculated RailFilterConfig with the new required configuration
		 */
	} else { /* Ingress ports were not changed, no need to zero */
		osm_nvl_prtn_send_rail_filter_of_prtn(p_subn, p_sw, p_prtn->pkey, rail, 0, 0);
	}
}

static uint8_t nvl_prtn_get_prtn_state_per_plane(osm_gpu_t * p_gpu, uint8_t port_num)
{
	osm_physp_t *p_physp;
	uint8_t plane;

	p_physp = osm_node_get_physp_ptr(p_gpu->p_node, port_num);
	if (!p_physp)
		return NVL_PRTN_GPU_STATUS_NON_MEMBER;

	plane = osm_gpu_get_port_plane(p_gpu, port_num);
	return p_gpu->prtn_member_status[plane];
}

static void nvl_prtn_update_sw_tbl(osm_subn_t * p_subn, osm_nvlink_prtn_t * p_prtn)
{
	cl_map_iterator_t p_next, p_item_sw_tbl;
	cl_map_item_t * p_item;
	osm_switch_t * p_sw;
	osm_physp_t * p_physp;
	uint8_t port, gpu_prtn_state;
	ib_net64_t sw_guid, remote_sw_guid;
	osm_node_t *p_remote_node;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw;

	/*
	 * Update partition switch table to hold switches that correspond to one of the following:
	 * 	a. Has a neighbor ALID that is part of the partition
	 * 	b. Partition max_trunk_links != 0 and switch has a neighbor switch (of the same plane)
	 *	   that is part of the partition
	 */

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/* Add to sw_tbl all switches that holds a neighbor ALID that is part of the partition */
	for (p_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *) p_item;

		for (port = 1; port < p_sw->num_ports; port++) {
			gpu_prtn_state = NVL_PRTN_GPU_STATUS_UNKNOWN;

			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port);

			if (!p_physp || p_physp->nvl_laas_pkey != p_prtn->pkey ||
			    osm_physp_is_fnm(p_subn, p_physp))
				continue;

			if (!p_physp->p_remote_physp)
				continue;

			p_remote_node = p_physp->p_remote_physp->p_node;
			if (osm_node_is_gpu(p_remote_node))
				gpu_prtn_state = nvl_prtn_get_prtn_state_per_plane(p_remote_node->p_gpu,
										   p_physp->p_remote_physp->port_num);

			sw_guid = osm_node_get_node_guid(p_physp->p_node);
			p_prtn_sw = (osm_nvlink_prtn_sw_data_t *)cl_map_get(&p_prtn->sw_tbl, sw_guid);
			if (!p_prtn_sw) {
				nvl_prtn_add_switch_to_laas(p_subn, p_prtn, p_sw);
				p_prtn_sw =
				    (osm_nvlink_prtn_sw_data_t *)cl_map_get(&p_prtn->sw_tbl, sw_guid);
				if (!p_prtn_sw)
					continue;

				/*
				 * Mark the switch as new to partition, if remote GPU is new to
				 * partition, but keep iterating on ports. In case finds a GPU
				 * that is already a member, it is an indication that the switch
				 * is already configured for this partition and no need to reroute.
				 */
				if (osm_node_is_gpu(p_remote_node) && gpu_prtn_state == NVL_PRTN_GPU_STATUS_NEW)
					p_prtn_sw->is_laas_set = FALSE;
			}

			if (osm_node_is_gpu(p_remote_node) && gpu_prtn_state == NVL_PRTN_GPU_STATUS_MEMBER)
				p_prtn_sw->is_laas_set = TRUE;
		}
	}

	/*
	 * In case number of trunk links > 0, iterate over sw_tbl and make sure remote
	 * switches are also part of the table
	 */
	if (!nvl_prtn_laas_get_max_trunk_links(p_prtn))
		goto Exit;

	p_next = cl_map_head(&p_prtn->sw_tbl);
	while (p_next != cl_map_end(&p_prtn->sw_tbl)) {
		p_item_sw_tbl = p_next;
		sw_guid = cl_map_key(p_item_sw_tbl);
		p_next = cl_map_next(p_item_sw_tbl);

		p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
		if (!p_sw)
			continue;

		for (port = 1; port < p_sw->num_ports; port++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, port);

			if (!p_physp || !p_physp->p_remote_physp ||
			    osm_node_get_type(p_physp->p_remote_physp->p_node) != IB_NODE_TYPE_SWITCH ||
			    osm_physp_is_fnm(p_subn, p_physp))
				continue;

			remote_sw_guid = osm_node_get_node_guid(p_physp->p_remote_physp->p_node);
			if (!cl_map_get(&p_prtn->sw_tbl, remote_sw_guid)) {
				nvl_prtn_add_remote_switch_to_laas(p_subn, p_prtn, p_physp);
			}
		}
	}

	// TODO: Remove from prtn->sw_tbl all switches that are not part of the partition.

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static void nvl_prtn_laas_process_prtn_update_rail_filter(osm_subn_t * p_subn,
							  osm_nvlink_prtn_t * p_prtn,
							  osm_switch_t * p_sw)

{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	ib_net64_t sw_guid;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;
	uint8_t rail, port_num;
	osm_physp_t *p_physp;
	boolean_t is_sw_reroute;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	OSM_ASSERT(p_sw);

	sw_guid = osm_node_get_node_guid(p_sw->p_node);
	p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_get(&p_prtn->sw_tbl, sw_guid);
	if (!p_prtn_sw_data) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB08: "
			"Could not update RailFilterConfig of switch 0x%"PRIx64" for pkey 0x%x\n",
			cl_ntoh64(sw_guid), cl_ntoh16(p_prtn->pkey));
		goto Exit;
	}

	is_sw_reroute = (p_prtn->state == NVL_PRTN_STATE_REROUTE) || !p_prtn_sw_data->is_laas_set;
	if (is_sw_reroute) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Reroute partition 0x%x - updating RailFilterConfig of switch 0x%"PRIx64"%s\n",
			cl_ntoh16(p_prtn->pkey), cl_ntoh64(sw_guid),
			p_prtn_sw_data->is_laas_set ? "" : " (new to partition)");

	} else if (p_prtn->state == NVL_PRTN_STATE_MAINTAIN) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Partition 0x%x must not be rerouted - checking for new access ports for"
			"switch 0x%"PRIx64"\n",
			cl_ntoh16(p_prtn->pkey), cl_ntoh64(sw_guid));
	} else {
		/* shouldn't get here */
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB0A: "
			"Partition 0x%x must not be rerouted - skips switch 0x%"PRIx64"\n",
			cl_ntoh16(p_prtn->pkey), cl_ntoh64(sw_guid));
		goto Exit;
	}


	/* If on reroute, set UC, MC, VL and zero port masks in order to rebuild */
	if (is_sw_reroute)
		nvl_prtn_laas_init_calc_rail_filter(p_sw, p_prtn_sw_data);

	/* Update access ports of RailFilterConfig */
	for (port_num = 0; port_num < p_sw->num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_sw->p_node, port_num);
		if (!p_physp || osm_physp_is_fnm(p_subn, p_physp))
			continue;

		/* Initialize the RailFilterConfig relevant fields according to switch reroute status */
		if ((p_physp->rail < OSM_NVLINK_SN_NUM_RAILS) && !is_sw_reroute)
			nvl_prtn_init_rail_filter_vl_mc_uc(&p_prtn_sw_data->
							   calc_rail_filter[p_physp->rail]);

		/* At this point, pkeys of GPUs that were deleted by request are already cleared */
		if (p_physp->nvl_laas_pkey != p_prtn->pkey) {
			if (!is_sw_reroute)
				nvl_prtn_rail_filter_update_port_num(p_mgr, p_prtn, p_sw, port_num,
								     OSM_NVL_REMOVE_PORT);
			continue;
		}

		/* Remove access port from any rail that it is not belong to */
		if (!is_sw_reroute) {
			for (rail = 0; rail < OSM_NVLINK_SN_NUM_RAILS; rail++) {
				/* Keep the port if remote does not exist - can be either access or trunk */
				if (!osm_nvlink_switch_is_rail_used(&p_sw->nvlink_sw_data, rail) ||
				    p_physp->rail == rail || !p_physp->p_remote_physp ||
				    osm_physp_is_trunk(p_subn, p_physp))
					continue;

				nvl_prtn_rail_filter_remove_port_num_from_rail(p_mgr, p_prtn, p_sw,
									       port_num, rail);
			}
		}
		/*
		 * Updates both access and trunk -
		 * on this flow trunk link nvl_laas_pkey is not changed. Meaning LAAS image
		 * of the subnet is kept using this field value, that is updated only
		 * on reroute.
		 */
		nvl_prtn_rail_filter_update_port_num(p_mgr, p_prtn, p_sw, port_num, OSM_NVL_ADD_PORT);
	}

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static void nvl_prtn_laas_process_prtn(osm_subn_t * p_subn, osm_nvlink_prtn_t * p_prtn)
{
	cl_map_iterator_t p_next, p_item;
	osm_switch_t *p_sw;
	ib_net64_t sw_guid;
	osm_nvlink_prtn_sw_data_t *p_prtn_sw_data;
	uint8_t rail;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/* Update switch table to hold all switches that should be configured for this partition */
	nvl_prtn_update_sw_tbl(p_subn, p_prtn);

	if (nvl_prtn_validate_sw_tbl_cnd_info(p_subn, p_prtn))
		return;

	/* Configure LAAS pkeys to ports */
	p_next = cl_map_head(&p_prtn->sw_tbl);
	while (p_next != cl_map_end(&p_prtn->sw_tbl)) {
		p_item = p_next;
		sw_guid = cl_map_key(p_item);
		p_next = cl_map_next(p_item);

		p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
		if (!p_sw)
			continue;

		p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_obj(p_item);

		if (p_prtn->state != NVL_PRTN_STATE_REROUTE && p_prtn_sw_data->is_laas_set)
			continue;

		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_DEBUG,
			"Reroute partition 0x%x - updating LAAS of switch 0x%"PRIx64"%s\n",
			cl_ntoh16(p_prtn->pkey), cl_ntoh64(sw_guid),
			p_prtn_sw_data->is_laas_set ? "" : " (new to partition)");

		/* Chooses LAAS trunks */
		nvl_prtn_laas_pkey_configuration(p_subn, p_prtn, p_sw, p_prtn_sw_data);
	}

	if (p_prtn->state == NVL_PRTN_STATE_ERROR)
		goto Exit;

	/* Build rail filter according to chosen pkeys */
	p_next = cl_map_head(&p_prtn->sw_tbl);
	while (p_next != cl_map_end(&p_prtn->sw_tbl)) {
		p_item = p_next;
		sw_guid = cl_map_key(p_item);
		p_next = cl_map_next(p_item);

		p_sw = osm_get_switch_by_guid(p_subn, sw_guid);
		if (!p_sw)
			continue;

		p_prtn_sw_data = (osm_nvlink_prtn_sw_data_t *)cl_map_obj(p_item);

		nvl_prtn_laas_process_prtn_update_rail_filter(p_subn, p_prtn, p_sw);

		/*
		 * If partition ports were changes, set RailFilterConfig.
		 * Zeros rail filter for removed ports, and update rail filter for current ports
		 */
		for (rail = 0; rail < OSM_NVLINK_SN_NUM_RAILS; rail++) {
			if (!osm_nvlink_switch_is_rail_used(&p_sw->nvlink_sw_data, rail))
				continue;
			nvl_prtn_laas_update_rail_filter(p_subn, p_prtn, p_sw, rail, p_prtn_sw_data);
		}
	}

	if (!wait_for_pending_transactions(p_subn->p_osm) &&
	    p_prtn->state != NVL_PRTN_STATE_ERROR &&
	    nvl_prtn_validate_prtn_subn_laas_ports(p_subn, p_prtn)) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BB07: "
			"RailFilterConfig was not set successfully to all switches "
			"of partition 0x%x. Failed updating partition access ports\n",
			cl_ntoh16(p_prtn->pkey));

		p_prtn->state = NVL_PRTN_STATE_ERROR;
	}

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

void osm_nvl_prtn_laas_process(osm_subn_t * p_subn)
{
	osm_nvlink_mgr_t *p_mgr = &p_subn->p_osm->sm.nvl_mgr;
	osm_nvlink_prtn_t *p_prtn;
	cl_map_item_t *p_next;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	osm_nvlink_rail_assignment(p_subn);

	p_next = cl_qmap_head(&p_mgr->nvl_prtn_tbl);
	while (p_next != cl_qmap_end(&p_mgr->nvl_prtn_tbl)) {
		p_prtn = (osm_nvlink_prtn_t *) p_next;
		p_next = cl_qmap_next(p_next);

		nvl_prtn_laas_process_prtn(p_subn, p_prtn);
	}

	(void) wait_for_pending_transactions(p_subn->p_osm);

	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

uint8_t osm_nvl_prtn_get_prtn_state(osm_nvlink_mgr_t * p_mgr, ib_net16_t pkey)
{
	cl_map_item_t *p_prtn_item;
	osm_nvlink_prtn_t *p_nvl_prtn;
	osm_log_t *p_log = &p_mgr->p_subn->p_osm->log;

	pkey &= cl_hton16((uint16_t) ~ 0x8000);

	p_prtn_item = cl_qmap_get(&p_mgr->nvl_prtn_tbl, pkey);
	if (p_prtn_item == cl_qmap_end(&p_mgr->nvl_prtn_tbl)) {
		OSM_LOG(p_log, OSM_LOG_ERROR, "ERR BB13: "
			"Could not find NVL partition for pkey 0x%x\n",
			cl_ntoh16(pkey));

		return NVL_PRTN_STATE_ERROR;
	}

	p_nvl_prtn = (osm_nvlink_prtn_t *)p_prtn_item;

	return p_nvl_prtn->state;
}

char * osm_nvl_prtn_get_prtn_state_str(osm_nvlink_mgr_t * p_mgr, ib_net16_t pkey)
{
        cl_map_item_t *p_prtn_item;
        osm_nvlink_prtn_t *p_nvl_prtn;

        pkey &= cl_hton16((uint16_t) ~ 0x8000);

        p_prtn_item = cl_qmap_get(&p_mgr->nvl_prtn_tbl, pkey);
        if (p_prtn_item == cl_qmap_end(&p_mgr->nvl_prtn_tbl))
                return "N/A";

        p_nvl_prtn = (osm_nvlink_prtn_t *)p_prtn_item;

	switch (p_nvl_prtn->state) {
	case NVL_PRTN_STATE_MAINTAIN:
		return "MAINTAIN";
	case NVL_PRTN_STATE_REROUTE_REQUESTED:
		return "REROUTE REQUESTED";
	case NVL_PRTN_STATE_REROUTE:
		return "REROUTE";
	case NVL_PRTN_STATE_ERROR:
		return "ERROR";
	default:
		return "N/A";
	}
}

