/*
 * Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <iba/ib_types.h>
#include <complib/cl_debug.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_PORT_RECOVERY_POLICY_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_helper.h>

extern int wait_for_pending_transactions(osm_opensm_t * p_osm);

static inline uint8_t is_port_recovery_policy_updated_per_profile(uint8_t profile_updated_mask, uint8_t profile)
{
	return (profile_updated_mask & (1 << profile));
}

static inline boolean_t port_recovery_policy_config_memcmp_subn_calc(ib_port_recovery_policy_config_t *p_calc_port_recovery,
								     ib_port_recovery_policy_config_t *p_subn_port_recovery)
{
	return ((p_calc_port_recovery->recovery_type_en != p_subn_port_recovery->recovery_type_en) ||
		(p_calc_port_recovery->draining_timeout != p_subn_port_recovery->draining_timeout) ||
		(p_calc_port_recovery->link_down_timeout != p_subn_port_recovery->link_down_timeout));
}

boolean_t osm_port_recovery_policy_can_activate_link(osm_physp_t *p_physp)
{
	osm_physp_t *p_remote_physp;
	boolean_t active_link = TRUE;

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (!p_physp->allow_link_activation ||
	    (p_remote_physp && !p_remote_physp->allow_link_activation))
		 active_link = FALSE;

	return active_link;
}

boolean_t port_recovery_policy_can_send_sw_profile_configuration(osm_switch_t *p_sw, uint8_t profile)
{
	uint8_t port;
	uint8_t num_physp;
	boolean_t send_conf = TRUE;
	osm_node_t *p_node;
	osm_physp_t *p_physp;

	num_physp = osm_node_get_num_physp(p_sw->p_node);
	p_node = p_sw->p_node;

	for (port = 0; port < num_physp; port++) {
		p_physp = osm_node_get_physp_ptr(p_node, port);
		if (p_physp->port_recovery_policy_profile == profile &&
		    osm_physp_get_port_state(p_physp) > IB_LINK_INIT) {
			send_conf = FALSE;
			break;
		}
	}

	return send_conf;
}

boolean_t osm_port_recovery_policy_can_send_profiles_config(osm_switch_t *p_sw)
{
	uint8_t port;
	uint8_t num_physp;
	boolean_t send_conf = TRUE;
	osm_node_t *p_node;
	osm_physp_t *p_physp;

	num_physp = osm_node_get_num_physp(p_sw->p_node);
	p_node = p_sw->p_node;

	for (port = 0; port < num_physp; port++) {
		p_physp = osm_node_get_physp_ptr(p_node, port);
		if (p_physp->port_reset_required) {
			send_conf = FALSE;
			break;
		}
	}

	return send_conf;
}

static void send_port_recovery_policy_get_per_profile(osm_sm_t * p_sm, osm_switch_t * p_sw, uint8_t profile)
{
	ib_port_recovery_policy_config_t *p_port_recovery_policy_subn;
	osm_physp_t *p_physp0;
	osm_port_recovery_policy_context_t context;
	ib_api_status_t status;
	uint32_t attr_mod;
	uint8_t profile_mask = 1 << profile;

	/* Profile zero is reserved on Set for this feature */
	OSM_ASSERT(profile);

	p_port_recovery_policy_subn =
	    &p_sw->port_recovery_policy_data.subn_port_recovery_policy[profile];

	if (!p_sm->p_subn->need_update && !p_sw->need_update &&
	    (p_sw->port_recovery_policy_data.port_recovery_policy_updated & profile_mask))
		return;

	p_sw->port_recovery_policy_data.port_recovery_policy_updated &= ~profile_mask;
	memset(p_port_recovery_policy_subn, 0, sizeof(ib_port_recovery_policy_config_t));

	p_physp0 = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp0) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR BE05: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	attr_mod = profile;

	memset(&context, 0, sizeof(context));
	context.node_guid = osm_node_get_node_guid(p_sw->p_node);
	context.port_guid = osm_physp_get_port_guid(p_physp0);
	context.mad_method = IB_MAD_METHOD_GET;
	context.profile = profile;
	context.port = p_physp0->port_num;

	status = osm_req_get(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp0), p_physp0,
			     IB_MAD_ATTR_PORT_RECOVERY_POLICY_CONFIG,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp0->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR BE06: "
			"Getting PortRecoveryPolicyConfig (profile %u) to "
			"0x%" PRIx64 " failed (%s)\n",
			profile, cl_ntoh64(p_physp0->port_guid),
			ib_get_err_str(status));
	}
}

static void send_port_recovery_policy_set_per_profile(osm_sm_t * p_sm, osm_switch_t * p_sw, uint8_t profile)
{
	ib_port_recovery_policy_config_t *p_port_recovery_policy_calc;
	ib_port_recovery_policy_config_t *p_port_recovery_policy_subn;
	osm_physp_t *p_physp0;
	osm_port_recovery_policy_context_t context;
	ib_api_status_t status;
	uint32_t attr_mod;
	uint8_t port_recovery_policy_updated;

	/* Profile zero is reserved on Set for this feature */
	OSM_ASSERT(profile);

	p_port_recovery_policy_calc =
	    &p_sw->port_recovery_policy_data.calc_port_recovery_policy[profile];
	p_port_recovery_policy_subn =
	    &p_sw->port_recovery_policy_data.subn_port_recovery_policy[profile];

	port_recovery_policy_updated =
	    is_port_recovery_policy_updated_per_profile(p_sw->port_recovery_policy_data.port_recovery_policy_updated, profile);

	if (port_recovery_policy_updated &&
	    !port_recovery_policy_config_memcmp_subn_calc(p_port_recovery_policy_calc, p_port_recovery_policy_subn))
		return;

	/* Send the MAD only if the profile is set for this switch */
	if (!osm_switch_is_profile_set_on_switch(p_sw, profile, IB_PROFILES_FEATURE_PORT_RECOVERY_POLICY))
		return;

	/* Check if all the ports of the profile configuration are in INIT/DOWN state */
	if (!port_recovery_policy_can_send_sw_profile_configuration(p_sw, profile))
		return;

	memset(p_port_recovery_policy_subn, 0, sizeof(ib_port_recovery_policy_config_t));

	p_physp0 = osm_node_get_physp_ptr(p_sw->p_node, 0);
	if (!p_physp0) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR BE02: "
			"Could not find port0 for switch 0x%"PRIx64"\n",
			cl_ntoh64(p_sw->p_node->node_info.node_guid));
		return;
	}

	attr_mod = profile;

	memset(&context, 0, sizeof(context));
	context.node_guid = osm_node_get_node_guid(p_sw->p_node);
	context.port_guid = osm_physp_get_port_guid(p_physp0);
	context.mad_method = IB_MAD_METHOD_SET;
	context.profile = profile;
	context.port = p_physp0->port_num;

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp0), p_physp0,
			     (const uint8_t *) p_port_recovery_policy_calc,
			     sizeof(ib_port_recovery_policy_config_t),
			     IB_MAD_ATTR_PORT_RECOVERY_POLICY_CONFIG,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp0->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR BE03: "
			"Setting PortRecoveryPolicyConfig (profile %u) to "
			"0x%" PRIx64 " failed (%s)\n",
			profile, cl_ntoh64(p_physp0->port_guid),
			ib_get_err_str(status));
	}
}

static void send_sw_port_recovery_policy_get(IN cl_map_item_t * p_map_item, IN void * cxt)
{
	osm_sm_t *p_sm = (osm_sm_t *)cxt;
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	uint8_t profile;

	/* Check if switch supports PortRecoveryPolicyConfig MAD */
	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_PORT_RECOVERY_POLICY_CONFIG_SUPPORTED_BIT))
		return;

	/*
	 * Send PortRecoveryPolicyConfig get MAD per profile.
	 */
	for (profile = IB_PORT_RECOVERY_POLICY_START_PROFILE; profile <= IB_PROFILE_CONFIG_MAX_PROFILE; profile++) {
		send_port_recovery_policy_get_per_profile(p_sm, p_sw, profile);
	}
}

static void send_sw_port_recovery_policy_set(IN cl_map_item_t * p_map_item, IN void * cxt)
{
	osm_sm_t *p_sm = (osm_sm_t *)cxt;
	osm_switch_t *p_sw = (osm_switch_t *)p_map_item;
	uint8_t profile;

	/* Check if switch supports PortRecoveryPolicyConfig MAD */
	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_PORT_RECOVERY_POLICY_CONFIG_SUPPORTED_BIT))
		return;

	/* Skip switches of which profiles weren't configured yet */
	if (!osm_switch_are_profiles_configured(p_sw, IB_PROFILES_FEATURE_PORT_RECOVERY_POLICY))
		return;

	/*
	 * Send PortRecoveryPolicyConfig MAD per profile.
	 */
	for (profile = IB_PORT_RECOVERY_POLICY_START_PROFILE; profile <= p_sw->port_recovery_policy_data.max_profile; profile++) {
		send_port_recovery_policy_set_per_profile(p_sm, p_sw, profile);
	}
}

boolean_t osm_port_recovery_policy_is_config_changed(osm_physp_t *p_physp)
{
	osm_port_recovery_policy_sw_data_t *p_port_recovery_policy_data;
	osm_node_t *p_node = p_physp->p_node;
	int subn_profile = 0;
	int calc_profile = 0;
	boolean_t is_conf_changed = FALSE;

	if (!p_node->sw) {
		if (port_recovery_policy_config_memcmp_subn_calc(&p_physp->calc_port_recovery_policy,
								 &p_physp->subn_port_recovery_policy)) {
			is_conf_changed = TRUE;
		}

		goto Exit;
	}

	p_port_recovery_policy_data = &p_node->sw->port_recovery_policy_data;
	calc_profile = p_physp->port_recovery_policy_profile;
	subn_profile = osm_profiles_config_get_profile(p_port_recovery_policy_data->ports_profiles, p_physp->port_num);

	if (calc_profile != subn_profile) {
		is_conf_changed = TRUE;
		goto Exit;
	}

	if (port_recovery_policy_config_memcmp_subn_calc(&p_port_recovery_policy_data->calc_port_recovery_policy[calc_profile],
							 &p_port_recovery_policy_data->subn_port_recovery_policy[subn_profile])) {
		is_conf_changed = TRUE;
	}

Exit:
	return is_conf_changed;
}

static void send_port_recovery_policy_get_per_port(IN cl_map_item_t * p_map_item, IN void * cxt)
{
	osm_port_recovery_policy_context_t context;
	ib_api_status_t status;
	uint8_t attr_mod;
	osm_sm_t *p_sm = (osm_sm_t *)cxt;
	osm_port_t *p_port = (osm_port_t *)p_map_item;
	osm_node_t *p_node = p_port->p_node;
	osm_physp_t *p_physp = p_port->p_physp;

	if (osm_node_get_type(p_port->p_node) != IB_NODE_TYPE_CA)
		return;

	/* Check if device supports PortRecoveryPolicyConfig MAD */
	if (!ib_general_info_is_cap_enabled(&p_node->general_info,
					    IB_GENERAL_INFO_IS_PORT_RECOVERY_POLICY_CONFIG_SUPPORTED_BIT))
		return;

	if (!p_sm->p_subn->need_update && !p_physp->need_update &&
	    !p_sm->p_subn->coming_out_of_standby &&
	    p_physp->port_recovery_policy_updated)
		return;

	memset(&p_physp->subn_port_recovery_policy, 0, sizeof(ib_port_recovery_policy_config_t));
	p_physp->port_recovery_policy_updated = FALSE;

	attr_mod = 0;
	memset(&context, 0, sizeof(context));
	context.node_guid = osm_node_get_node_guid(p_node);
	context.port_guid = osm_physp_get_port_guid(p_physp);
	context.mad_method = IB_MAD_METHOD_GET;
	context.profile = 0;
	context.port = p_physp->port_num;

	status = osm_req_get(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp), p_physp,
			     IB_MAD_ATTR_PORT_RECOVERY_POLICY_CONFIG,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR BE07: "
			"Getting PortRecoveryPolicyConfig from port"
			"0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(p_physp->port_guid),
			ib_get_err_str(status));
	}
}

static void send_port_recovery_policy_set_per_port(IN cl_map_item_t * p_map_item, IN void * cxt)
{
	osm_port_recovery_policy_context_t context;
	ib_api_status_t status;
	uint8_t attr_mod;
	osm_sm_t *p_sm = (osm_sm_t *)cxt;
	osm_port_t *p_port = (osm_port_t *)p_map_item;
	osm_node_t *p_node = p_port->p_node;
	osm_physp_t *p_physp = p_port->p_physp;

	if (osm_node_get_type(p_port->p_node) != IB_NODE_TYPE_CA)
		return;

	/* Check if device supports PortRecoveryPolicyConfig MAD */
	if (!ib_general_info_is_cap_enabled(&p_node->general_info,
					    IB_GENERAL_INFO_IS_PORT_RECOVERY_POLICY_CONFIG_SUPPORTED_BIT))
		return;

	if (p_physp->port_recovery_policy_updated &&
	    !port_recovery_policy_config_memcmp_subn_calc(&p_physp->calc_port_recovery_policy, &p_physp->subn_port_recovery_policy))
		return;

	/* Don't send configuration in case the port is not in INIT/DOWN state */
	if (osm_physp_get_port_state(p_physp) > IB_LINK_INIT)
		return;

	memset(&p_physp->subn_port_recovery_policy, 0, sizeof(ib_port_recovery_policy_config_t));

	attr_mod = 0;
	memset(&context, 0, sizeof(context));
	context.node_guid = osm_node_get_node_guid(p_node);
	context.port_guid = osm_physp_get_port_guid(p_physp);
	context.mad_method = IB_MAD_METHOD_SET;
	context.profile = 0;
	context.port = p_physp->port_num;

	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_physp), p_physp,
			     (const uint8_t *) &p_physp->calc_port_recovery_policy,
			     sizeof(ib_port_recovery_policy_config_t),
			     IB_MAD_ATTR_PORT_RECOVERY_POLICY_CONFIG,
			     cl_hton32(attr_mod), FALSE,
			     ib_port_info_get_m_key(&p_physp->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&context);

	if (status != IB_SUCCESS) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR BE04: "
			"Setting PortRecoveryPolicyConfig to port"
			"0x%" PRIx64 " failed (%s)\n",
			cl_ntoh64(p_physp->port_guid),
			ib_get_err_str(status));
	}
}

static void send_get_port_recovery_profiles_config(IN cl_map_item_t * p_map_item, IN void * cxt)
{
	osm_sm_t *p_sm = (osm_sm_t *) cxt;
	osm_switch_t *p_sw = (osm_switch_t *) p_map_item;

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_PORT_RECOVERY_POLICY_CONFIG_SUPPORTED_BIT))
		return;

	osm_subn_send_profiles_config_blocks(p_sm, p_sw,
					     IB_PROFILES_FEATURE_PORT_RECOVERY_POLICY,
					     IB_MAD_METHOD_GET);
}

static void send_set_port_recovery_profiles_config(IN cl_map_item_t * p_map_item, IN void * cxt)
{
	osm_sm_t *p_sm = (osm_sm_t *) cxt;
	osm_switch_t *p_sw = (osm_switch_t *) p_map_item;

	if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					    IB_GENERAL_INFO_IS_PORT_RECOVERY_POLICY_CONFIG_SUPPORTED_BIT))
		return;

	osm_subn_send_profiles_config_blocks(p_sm, p_sw,
					     IB_PROFILES_FEATURE_PORT_RECOVERY_POLICY,
					     IB_MAD_METHOD_SET);
}

void osm_port_recovery_policy_get_subnet_config(osm_subn_t *p_subn)
{
	osm_sm_t *sm = &p_subn->p_osm->sm;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	if (!p_subn->opt.port_recovery_policy_enabled)
		goto Exit;

	/* Get port profiles of PortRecoveryPolicy feature by ProfileConfig MAD */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, send_get_port_recovery_profiles_config, sm);
	if (wait_for_pending_transactions(p_subn->p_osm))
		goto Exit;

	/* Send get PortRecoveryPolicyConfig to switches */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, send_sw_port_recovery_policy_get, sm);
	if (wait_for_pending_transactions(p_subn->p_osm))
		goto Exit;

	/* Send PortRecoveryPolicyConfig get to CAs */
	cl_qmap_apply_func(&p_subn->port_guid_tbl, send_port_recovery_policy_get_per_port, sm);
	if (wait_for_pending_transactions(p_subn->p_osm))
		goto Exit;
Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

void osm_port_recovery_policy_configure(osm_subn_t *p_subn)
{
	osm_sm_t *sm = &p_subn->p_osm->sm;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	if (!p_subn->opt.port_recovery_policy_enabled)
		goto Exit;

	/* Set port profiles of PortRecoveryPolicy feature by ProfileConfig MAD */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, send_set_port_recovery_profiles_config, sm);
	if (wait_for_pending_transactions(p_subn->p_osm))
		goto Exit;

	/* Send PortRecoveryPolicyConfig to switches according to parsed cfg parameters */
	cl_qmap_apply_func(&p_subn->sw_guid_tbl, send_sw_port_recovery_policy_set, sm);
	if (wait_for_pending_transactions(p_subn->p_osm))
		goto Exit;

	/* Send PortRecoveryPolicyConfig to CAs according to parsed cfg parameters */
	cl_qmap_apply_func(&p_subn->port_guid_tbl, send_port_recovery_policy_set_per_port, sm);
	if (wait_for_pending_transactions(p_subn->p_osm))
		goto Exit;

Exit:
	OSM_LOG_EXIT(&p_subn->p_osm->log);
}

static ib_port_recovery_policy_config_t *get_subn_port_recovery_policy(osm_physp_t *p_physp)
{
	ib_port_recovery_policy_config_t *p_port_recovery_policy;
	osm_port_recovery_policy_sw_data_t *p_port_recovery_policy_data;
	uint8_t subn_profile = 0;

	if (p_physp->p_node->sw) {
		p_port_recovery_policy_data = &p_physp->p_node->sw->port_recovery_policy_data;
		subn_profile = osm_profiles_config_get_profile(p_port_recovery_policy_data->ports_profiles, p_physp->port_num);
		p_port_recovery_policy = &p_port_recovery_policy_data->subn_port_recovery_policy[subn_profile];
	} else {
		p_port_recovery_policy = &p_physp->subn_port_recovery_policy;
	}

	return p_port_recovery_policy;
}

boolean_t osm_port_recovery_policy_is_link_conf_equal(osm_sm_t *sm, osm_physp_t *p_physp)
{
	ib_port_recovery_policy_config_t *p_port_recovery_policy;
	ib_port_recovery_policy_config_t *p_remote_port_recovery_policy;
	osm_physp_t *p_remote_physp;
	boolean_t equal = TRUE;

	if (!sm->p_subn->opt.port_recovery_policy_enabled)
		goto Exit;

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (!p_remote_physp)
		goto Exit;

	if (!ib_general_info_is_cap_enabled(&p_physp->p_node->general_info,
					    IB_GENERAL_INFO_IS_PORT_RECOVERY_POLICY_CONFIG_SUPPORTED_BIT) ||
	    !ib_general_info_is_cap_enabled(&p_remote_physp->p_node->general_info,
					    IB_GENERAL_INFO_IS_PORT_RECOVERY_POLICY_CONFIG_SUPPORTED_BIT)) {
		goto Exit;
	}

	p_port_recovery_policy = get_subn_port_recovery_policy(p_physp);
	p_remote_port_recovery_policy = get_subn_port_recovery_policy(p_remote_physp);

	if (p_port_recovery_policy->recovery_type_en != p_remote_port_recovery_policy->recovery_type_en) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Link 0x%" PRIx64 ":%u - 0x%" PRIx64
			"%u has mismatch in recovery_type_en configuration: %u != %u\n",
			osm_node_get_node_guid_ho(p_physp->p_node), p_physp->port_num,
			osm_node_get_node_guid_ho(p_remote_physp->p_node), p_remote_physp->port_num,
			cl_ntoh16(p_port_recovery_policy->recovery_type_en),
			cl_ntoh16(p_remote_port_recovery_policy->recovery_type_en));
		equal = FALSE;
	}

	if (p_port_recovery_policy->link_down_timeout != p_remote_port_recovery_policy->link_down_timeout) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Link 0x%" PRIx64 ":%u - 0x%" PRIx64
			"%u has mismatch in link_down_timeout configuration: %u != %u\n",
			osm_node_get_node_guid_ho(p_physp->p_node), p_physp->port_num,
			osm_node_get_node_guid_ho(p_remote_physp->p_node), p_remote_physp->port_num,
			cl_ntoh16(p_port_recovery_policy->link_down_timeout),
			cl_ntoh16(p_remote_port_recovery_policy->link_down_timeout));
		equal = FALSE;
	}

	if (p_physp->p_node->sw && p_remote_physp->p_node->sw &&
	    (p_port_recovery_policy->draining_timeout != p_remote_port_recovery_policy->draining_timeout)) {
		OSM_LOG(sm->p_log, OSM_LOG_VERBOSE,
			"Link 0x%" PRIx64 ":%u - 0x%" PRIx64
			"%u has mismatch in draining_timeout configuration: %u != %u\n",
			osm_node_get_node_guid_ho(p_physp->p_node), p_physp->port_num,
			osm_node_get_node_guid_ho(p_remote_physp->p_node), p_remote_physp->port_num,
			p_port_recovery_policy->draining_timeout,
			p_remote_port_recovery_policy->draining_timeout);
		equal = FALSE;
	}

Exit:
	return equal;
}

boolean_t is_port_recovery_policy_link_conf_valid(osm_sm_t *sm, osm_physp_t *p_physp) 
{
	boolean_t valid = TRUE;
	osm_physp_t *p_remote_physp;
	ib_port_recovery_policy_config_t *p_port_recovery_policy, *p_remote_port_recovery_policy;
	ib_port_recovery_policy_config_t zero_port_recovery_policy;
	boolean_t is_port_recovery_policy_parsed, is_remote_port_recovery_policy_parsed;

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (!p_remote_physp)
		goto Exit;

	is_port_recovery_policy_parsed = p_physp->p_node->sw ?
	    p_physp->p_node->conf_params.is_port_recovery_policy_parsed :
	    p_physp->conf_params.is_port_recovery_policy_parsed;
	is_remote_port_recovery_policy_parsed = p_remote_physp->p_node->sw ?
	    p_remote_physp->p_node->conf_params.is_port_recovery_policy_parsed :
	    p_remote_physp->conf_params.is_port_recovery_policy_parsed;

	if (!is_port_recovery_policy_parsed && !is_remote_port_recovery_policy_parsed)
		goto Exit;

	memset(&zero_port_recovery_policy, 0, sizeof(ib_port_recovery_policy_config_t));
	if (is_port_recovery_policy_parsed) {
		p_port_recovery_policy = p_physp->p_node->sw ?
		    &p_physp->p_node->sw->conf_params.port_recovery_policy_config[p_physp->conf_params.port_recovery_policy_profile] :
		    &p_physp->conf_params.port_recovery_policy_config;
	} else {
		p_port_recovery_policy = &zero_port_recovery_policy;
	}

	if (is_remote_port_recovery_policy_parsed) {
		p_remote_port_recovery_policy = p_remote_physp->p_node->sw ?
		    &p_remote_physp->p_node->sw->conf_params.port_recovery_policy_config[p_remote_physp->conf_params.port_recovery_policy_profile] :
		    &p_remote_physp->conf_params.port_recovery_policy_config;
	} else {
		p_remote_port_recovery_policy = &zero_port_recovery_policy;
	}


	if (p_port_recovery_policy->recovery_type_en != p_remote_port_recovery_policy->recovery_type_en) {
		valid = FALSE;
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR BE08: Invalid Port Recovery Policy configuration on link 0x%" PRIx64 ":%u - 0x%" PRIx64
			"%u, mismatch in recovery_type_en: %u != %u\n",
			osm_node_get_node_guid_ho(p_physp->p_node), p_physp->port_num,
			osm_node_get_node_guid_ho(p_remote_physp->p_node), p_remote_physp->port_num,
			p_port_recovery_policy->recovery_type_en,
			p_remote_port_recovery_policy->recovery_type_en);
	}
	
	if (p_port_recovery_policy->link_down_timeout != p_remote_port_recovery_policy->link_down_timeout) {
		valid = FALSE;
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR BE09: Invalid Port Recovery Policy configuration "
			"on link 0x%" PRIx64 ":%u - 0x%" PRIx64
			"%u, mismatch in link_down_timeout: %u != %u\n",
			osm_node_get_node_guid_ho(p_physp->p_node), p_physp->port_num,
			osm_node_get_node_guid_ho(p_remote_physp->p_node), p_remote_physp->port_num,
			p_port_recovery_policy->link_down_timeout,
			p_remote_port_recovery_policy->link_down_timeout);
	}

	if (p_physp->p_node->sw && p_remote_physp->p_node->sw &&
	    (p_port_recovery_policy->draining_timeout != p_remote_port_recovery_policy->draining_timeout)) {
		valid = FALSE;
		OSM_LOG(sm->p_log, OSM_LOG_ERROR,
			"ERR BE0A: Invalid Port Recovery Policy configuration "
			"on link 0x%" PRIx64 ":%u - 0x%" PRIx64
			"%u, mismatch in draining_timeout: %u != %u\n",
			osm_node_get_node_guid_ho(p_physp->p_node), p_physp->port_num,
			osm_node_get_node_guid_ho(p_remote_physp->p_node), p_remote_physp->port_num,
			p_port_recovery_policy->draining_timeout,
			p_remote_port_recovery_policy->draining_timeout);
	}

Exit:
	return valid;
}

static void port_recovery_policy_build_config(osm_sm_t *sm)
{
	ib_port_recovery_policy_config_t port_recovery_policy[IB_PROFILE_CONFIG_NUM_PROFILES];
	ib_port_recovery_policy_config_t *p_port_recovery_policy;
	osm_switch_t *p_sw;
	osm_physp_t *p_physp;
	cl_map_item_t *p_item;
	boolean_t is_one_non_down_port_exists = FALSE;
	boolean_t is_port_recovery_policy_parsed;
	int i;

	memset(port_recovery_policy, 0, sizeof(port_recovery_policy));

	for (p_item = cl_qmap_head(&sm->p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&sm->p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *)p_item;
		is_one_non_down_port_exists = FALSE;
		is_port_recovery_policy_parsed = p_sw->p_node->conf_params.is_port_recovery_policy_parsed;

		if (!ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
						    IB_GENERAL_INFO_IS_PORT_RECOVERY_POLICY_CONFIG_SUPPORTED_BIT))
			continue;

		for (i = 1; i < p_sw->p_node->physp_tbl_size; i++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, i);
			if (!p_physp)
				continue;

			/* Skip special ports */
			if (p_physp->p_remote_physp && osm_physp_get_special_port_type(p_physp->p_remote_physp))
				continue;

			/* TODO: Need to make sure that no mads are send if there is a bad configuration */
			p_physp->port_recovery_policy_prev_profile = p_physp->port_recovery_policy_profile;
			if (ib_port_info_get_port_state(&p_physp->port_info) != IB_LINK_DOWN &&
			    is_port_recovery_policy_link_conf_valid(sm, p_physp) &&
			    is_port_recovery_policy_parsed) {
				p_physp->port_recovery_policy_profile = p_physp->conf_params.port_recovery_policy_profile;
				is_one_non_down_port_exists = TRUE;
			} else {
				p_physp->port_recovery_policy_profile = IB_PORT_RECOVERY_POLICY_DEFAULT_PROFILE;
			}
		}

		if (is_port_recovery_policy_parsed && is_one_non_down_port_exists) {
			memcpy(&p_sw->port_recovery_policy_data.calc_port_recovery_policy, &p_sw->conf_params.port_recovery_policy_config,
			       sizeof(p_sw->port_recovery_policy_data.calc_port_recovery_policy));

			p_sw->port_recovery_policy_data.max_profile = p_sw->conf_params.port_recovery_policy_max_profile;
		} else {
			memcpy(&p_sw->port_recovery_policy_data.calc_port_recovery_policy, &port_recovery_policy,
		       	       sizeof(p_sw->port_recovery_policy_data.calc_port_recovery_policy));
			p_sw->port_recovery_policy_data.max_profile = IB_PORT_RECOVERY_POLICY_DEFAULT_PROFILE;
		}
	}

	for (p_item = cl_qmap_head(&sm->p_subn->port_guid_tbl);
	     p_item != cl_qmap_end(&sm->p_subn->port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_port_t *p_port = (osm_port_t *)p_item;

		if (p_port->p_node->sw)
			continue;

		if (!ib_general_info_is_cap_enabled(&p_port->p_node->general_info,
						    IB_GENERAL_INFO_IS_PORT_RECOVERY_POLICY_CONFIG_SUPPORTED_BIT))
			continue;

		/* Skip special ports */
		if (p_port->p_physp->p_remote_physp &&
		    (osm_node_get_type(p_port->p_physp->p_remote_physp->p_node) == IB_NODE_TYPE_SWITCH) &&
		    osm_physp_get_special_port_type(p_port->p_physp))
			continue;

		is_port_recovery_policy_parsed = p_port->p_physp->conf_params.is_port_recovery_policy_parsed;
		p_port_recovery_policy =
		    is_port_recovery_policy_parsed && is_port_recovery_policy_link_conf_valid(sm, p_port->p_physp) ?
		    &p_port->p_physp->conf_params.port_recovery_policy_config :
		    &port_recovery_policy[IB_PORT_RECOVERY_POLICY_DEFAULT_PROFILE];

		memcpy(&p_port->p_physp->calc_port_recovery_policy, p_port_recovery_policy,
		       sizeof(p_port->p_physp->calc_port_recovery_policy));
	}
}

static void port_recovery_policy_build_disable_config(osm_subn_t *p_subn)
{
	osm_switch_t *p_sw;
	osm_physp_t *p_physp;
	cl_map_item_t *p_item;
	int i;

	for (p_item = cl_qmap_head(&p_subn->sw_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->sw_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw = (osm_switch_t *)p_item;

		for (i = 1; i < p_sw->p_node->physp_tbl_size; i++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, i);
			if (!p_physp)
				continue;

			/* Skip special ports and down ports */
			if ((ib_port_info_get_port_state(&p_physp->port_info) == IB_LINK_DOWN) ||
			    ib_mlnx_epi_get_special_port_type(&p_physp->ext_port_info))	{
				p_physp->port_recovery_policy_profile = IB_PORT_RECOVERY_POLICY_DEFAULT_PROFILE;
				continue;
			}

			p_physp->port_recovery_policy_profile = IB_PORT_RECOVERY_POLICY_DISABLE_PROFILE;
		}
		memset(&p_sw->port_recovery_policy_data.calc_port_recovery_policy, 0,
		       sizeof(p_sw->port_recovery_policy_data.calc_port_recovery_policy));
		p_sw->port_recovery_policy_data.max_profile = IB_PORT_RECOVERY_POLICY_DISABLE_PROFILE;
	}

	for (p_item = cl_qmap_head(&p_subn->port_guid_tbl);
	     p_item != cl_qmap_end(&p_subn->port_guid_tbl);
	     p_item = cl_qmap_next(p_item)) {
		osm_port_t *p_port = (osm_port_t *)p_item;

		if (p_port->p_node->sw)
			continue;

		memset(&p_port->p_physp->calc_port_recovery_policy, 0,
		       sizeof(p_port->p_physp->calc_port_recovery_policy));
	}
}

void osm_port_recovery_policy_calculate(osm_subn_t *p_subn)
{
	if (p_subn->opt.port_recovery_policy_enabled == OSM_CAPABILITY_ENABLED)
		port_recovery_policy_build_config(&p_subn->p_osm->sm);
	else if (p_subn->opt.port_recovery_policy_enabled == OSM_CAPABILITY_DISABLED)
		port_recovery_policy_build_disable_config(p_subn);
}

void osm_port_recovery_policy_preprocess_links(osm_sm_t * sm)
{
	osm_switch_t *p_sw;
	osm_node_t *p_node;
	osm_physp_t *p_physp;
	uint8_t i, num_physp;
	boolean_t link_reset_required;
	int disallowed_activation_links_count = 0;

	OSM_LOG_ENTER(sm->p_log);

	foreach_node(sm->p_subn, p_node) {
		num_physp = osm_node_get_num_physp(p_node);

		for (i = 1; i < num_physp; i++) {
			p_physp = osm_node_get_physp_ptr(p_node, i);
			if (!p_physp)
				continue;

			p_physp->allow_link_activation = TRUE;
		}
	}

	if (!sm->p_subn->opt.port_recovery_policy_enabled)
		goto Exit;

	foreach_switch(sm->p_subn, p_sw) {
		num_physp = osm_node_get_num_physp(p_sw->p_node);

		link_reset_required = !osm_port_recovery_policy_can_send_profiles_config(p_sw);
		for (i = 1; i < num_physp; i++) {
			p_physp = osm_node_get_physp_ptr(p_sw->p_node, i);
			if (!p_physp)
				continue;

			if (link_reset_required) {
				p_physp->allow_link_activation = FALSE;
				disallowed_activation_links_count++;
			}
		}
	}

	/*
	 * Log number of ports that will not be activated due to port recovery policy change.
	 * Needed for debugging cases where port might not be activated by SM.
	 */
	if (disallowed_activation_links_count) {
		OSM_LOG(sm->p_log, OSM_LOG_INFO,
			"Number of ports that will not be activated "
			"due to port recovery policy change: %d\n",
			disallowed_activation_links_count);
	}

Exit:
	OSM_LOG_EXIT(sm->p_log);
}

void osm_port_recovery_policy_config_rcv_process(IN void *context, IN void *data)
{
	osm_sm_t *sm = context;
	osm_madw_t *p_madw = data;
	ib_port_recovery_policy_config_t *p_port_recovery_policy;
	ib_smp_t *p_smp;
	osm_port_recovery_policy_context_t *p_context;
	osm_node_t *p_node;
	osm_port_t *p_port;

	OSM_ASSERT(sm);
	OSM_ASSERT(p_madw);

	OSM_LOG_ENTER(sm->p_log);

	p_smp = osm_madw_get_smp_ptr(p_madw);
	p_context = (osm_port_recovery_policy_context_t *)&p_madw->context;
	p_port_recovery_policy = ib_smp_get_payload_ptr(p_smp);

	OSM_ASSERT(p_smp->attr_id == IB_MAD_ATTR_PORT_RECOVERY_POLICY_CONFIG);

	CL_PLOCK_EXCL_ACQUIRE(sm->p_lock);

	if (ib_smp_get_status(p_smp)) {
		OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
			"MAD status 0x%x received\n",
			cl_ntoh16(ib_smp_get_status(p_smp)));
		goto Exit;
	}

	p_port = osm_get_port_by_guid(sm->p_subn, p_context->port_guid);
	if (PF(!p_port)) {
		OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR BE01: "
			"No port object for port with GUID 0x%" PRIx64
			" for parent node GUID 0x%" PRIx64
			", TID 0x%" PRIx64 "\n",
			cl_ntoh64(p_context->port_guid),
			cl_ntoh64(p_context->node_guid), cl_ntoh64(p_smp->trans_id));
		goto Exit;
	}

	p_node = p_port->p_node;
	OSM_ASSERT(p_node);

	if (p_node->sw) {
		memcpy(&p_node->sw->port_recovery_policy_data.subn_port_recovery_policy[p_context->profile],
		       p_port_recovery_policy, sizeof(ib_port_recovery_policy_config_t));
		if (p_context->mad_method == IB_MAD_METHOD_GET)
			p_node->sw->port_recovery_policy_data.port_recovery_policy_updated |= (1 << p_context->profile);
	} else {
		memcpy(&p_port->p_physp->subn_port_recovery_policy, p_port_recovery_policy,
		       sizeof(ib_port_recovery_policy_config_t));
		if (p_context->mad_method == IB_MAD_METHOD_GET)
			p_port->p_physp->port_recovery_policy_updated = TRUE;
	}

Exit:
	CL_PLOCK_RELEASE(sm->p_lock);
	OSM_LOG_EXIT(sm->p_log);
}
