/*
 * Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Main entry point into version 2 routing subsystem.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <math.h>
#include <regex.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/time.h>

#include <complib/cl_types_osd.h>
#include <complib/cl_types.h>
#include <complib/cl_byteswap.h>
#include <complib/cl_heap.h>
#include <complib/cl_bitmap.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTING_V2_DIJKSTRA_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_routing.h>
#include <opensm/osm_advanced_routing.h>
#include <opensm/osm_graph.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_routing_v2.h>

/* ---------------------------------------------------------------------------------------------- */

static int routing_ar_dijkstra_switch(IN osm_route_context_t *p_context, IN osm_vertex_t *p_base_vertex)
{
	int			src;
	int			dst;
	int			base;
	uint8_t			*visited;
	uint8_t			*distance;
	uint32_t		dist;
	osm_vdq_t		*p_vdq;
	osm_route_t		*p_dst_route;
	osm_route_t		*p_prev_route;
	osm_vertex_t		*p_src_vertex;
	osm_vertex_t		*p_dst_vertex;
	osm_edge_t		*p_edge;
	osm_ucast_mgr_t		*p_mgr;
	osm_graph_t		*p_graph;
	cl_heap_t		heap;
	cl_heap_t		*p_heap;

	p_mgr = p_context->p_mgr;
	p_vdq = p_context->task_context;

	p_graph = osm_ucast_mgr_plane_to_graph(p_mgr, p_context->plane);

	base = p_base_vertex->index;

	osm_routing_vdq_reset(p_vdq);

	visited  = p_vdq->visited;
	distance = p_vdq->distance;

	/*
	 * Set the base vertex as already visited.
	 */
	visited[base] = TRUE;

	/*
	 * Create the heap.
	 */
	p_heap = &heap;
	cl_heap_construct(p_heap);
	cl_heap_init(p_heap, p_graph->num_vertices + 1, 4, &osm_routing_index_update, &osm_routing_key_cmp);

	/*
	 * Seed the heap with the src switch's neighbors.
	 */
	foreach_graph_vertex_edge(p_base_vertex, p_edge) {
		dist = p_edge->weight;

		dst = p_edge->remote_vertex_index;
		p_dst_vertex = osm_graph_index_to_vertex(p_graph, dst);
		p_dst_route = osm_vertex_to_vertex_route(p_context, p_base_vertex, p_dst_vertex);

		if (!visited[dst]) {
			visited[dst] = TRUE;

			distance[dst] = dist;
			osm_route_ports_set(p_dst_route, p_edge->port_num);
			cl_heap_insert(p_heap, routing_key(p_dst_route, dist), p_dst_route);
		} else if (dist < distance[dst]) {
			distance[dst] = dist;
			osm_route_ports_clear(p_dst_route);
			osm_route_ports_set(p_dst_route, p_edge->port_num);

			cl_heap_modify_key(p_heap, routing_key(p_dst_route, dist), p_dst_route->heap_index);
		} else if (dist == distance[dst]) {
			osm_route_ports_set(p_dst_route, p_edge->port_num);
		}
	}

	/*
	 * Process the heap.
	 */
	while (!cl_heap_is_empty(p_heap)) {
		p_prev_route = (osm_route_t *)cl_heap_extract_root(p_heap);
		p_prev_route->heap_index = 0xffffffff;

		src = p_prev_route->dst;
		p_src_vertex = osm_graph_index_to_vertex(p_graph, src);

		foreach_graph_vertex_edge(p_src_vertex, p_edge) {
			dist = distance[src] + p_edge->weight;

			dst = p_edge->remote_vertex_index;
			if (dst == base) {
				continue;
			}

			p_dst_vertex = osm_graph_index_to_vertex(p_graph, dst);
			p_dst_route = osm_vertex_to_vertex_route(p_context, p_base_vertex, p_dst_vertex);

			if (!visited[dst]) {
				visited[dst] = TRUE;

				distance[dst] = dist;
				osm_route_ports_copy(p_dst_route, p_prev_route);
				cl_heap_insert(p_heap, routing_key(p_dst_route, dist), p_dst_route);
			} else if (dist < distance[dst]) {
				distance[dst] = dist;
				osm_route_ports_copy(p_dst_route, p_prev_route);

				cl_heap_modify_key(p_heap, routing_key(p_dst_route, dist), p_dst_route->heap_index);
			} else if (dist == distance[dst]) {
				osm_route_ports_merge(p_dst_route, p_prev_route);
			}
		}
	}

	cl_heap_destroy(p_heap);
	return 0;
}

/* ---------------------------------------------------------------------------------------------- */

static void *dijkstra_task_init(IN osm_route_context_t *p_context)
{
	return osm_routing_vdq_init(p_context->p_mgr, p_context->p_graph);
}

static void dijkstra_task_fini(IN void *p_void)
{
	osm_routing_vdq_free(p_void);
}

/* ---------------------------------------------------------------------------------------------- */

int osm_routing_v2_ar_dijkstra_route(IN osm_route_context_t *p_context)
{
	int			status;

	p_context->sw_function = routing_ar_dijkstra_switch;
	p_context->sw_init_function = dijkstra_task_init;
	p_context->sw_fini_function = dijkstra_task_fini;
	status = osm_routing_per_vertex(p_context);

	return status;
}
