/*
 * Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTING_V2_MAXFLOW_C
#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_graph.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_routing_v2.h>

#define	MAXFLOW_MAX_SUPPORTED_PORTS	255

typedef struct {
	int				flow;
	int				capacity;
	int				residual_edge_index;
} maxflow_edge_t;

typedef struct {
	uint32_t			rank;
	boolean_t			calculate_flow;
} maxflow_vertex_t;

typedef struct {
	uint8_t				*ports;
	uint8_t				num_ports;
} maxflow_port_group_t;

typedef struct {
	uint32_t			vertex_index_of_representative;
	uint32_t			vertex_index_of_second_representative;
	uint8_t				representative_port_map[MAXFLOW_MAX_SUPPORTED_PORTS];
	uint8_t				second_representative_port_map[MAXFLOW_MAX_SUPPORTED_PORTS];
	maxflow_port_group_t		*port_groups;
} maxflow_port_groups_t;

typedef struct {
	uint32_t			visited_token;
	uint32_t			*visited;
	boolean_t			*dirty;

	uint32_t			num_vertices;
	osm_vertex_t			*p_dst_vertex;
	osm_graph_t			*p_graph;
	osm_subn_t			*p_subn;
	osm_log_t			*p_log;
	maxflow_edge_t			*edges;
	maxflow_vertex_t		*vertices;

	uint16_t			dst_lid_ho;
	int				count_sort_array_size;
	cl_list_t			*count_sort;
	maxflow_port_groups_t		*src_port_groups;
	osm_ucast_mgr_t			*p_mgr;
} maxflow_t;

typedef struct vertex_index_to_port_ {
	int				remote_index;
	int				port_num;
} vertex_index_to_port_t;

typedef struct key_to_vertex_ {
	char key[2048];
	int vertex_index;
	uint8_t ports[MAXFLOW_MAX_SUPPORTED_PORTS];
} key_to_vertex_t;

#define	edge_to_maxflow_edge(P_MAXFLOW, P_EDGE)		&P_MAXFLOW->edges[P_EDGE->index]
#define	vertex_to_maxflow_vertex(P_MAXFLOW, P_VERTEX)	&P_MAXFLOW->vertices[P_VERTEX->index]

/* ================================================================================================
 * Flow functions.
 */

static inline boolean_t flow_is_residual(IN maxflow_t *p_maxflow, IN osm_edge_t *p_edge)
{
	maxflow_edge_t		*p_maxflow_edge = edge_to_maxflow_edge(p_maxflow, p_edge);

	return (p_maxflow_edge->capacity == 0);
}

static inline int flow_get_remaining_capacity(IN maxflow_t *p_maxflow, IN osm_edge_t *p_edge)
{
	maxflow_edge_t		*p_maxflow_edge = edge_to_maxflow_edge(p_maxflow, p_edge);

	return (p_maxflow_edge->capacity - p_maxflow_edge->flow);
}

static inline void flow_augment(IN maxflow_t *p_maxflow, IN osm_edge_t *p_edge, IN int bottle_neck)
{
	maxflow_edge_t		*p_maxflow_edge = edge_to_maxflow_edge(p_maxflow, p_edge);
	maxflow_edge_t		*p_maxflow_residual_edge = &p_maxflow->edges[p_maxflow_edge->residual_edge_index];

	p_maxflow_edge->flow += bottle_neck;
	p_maxflow_residual_edge->flow -= bottle_neck;
}

__attribute__((unused))
static void destroy_count_sort_array(IN maxflow_t *p_maxflow)
{
	int i;
	int array_size = p_maxflow->count_sort_array_size;
	cl_list_t *count_sort = p_maxflow->count_sort;

	if (!count_sort)
		return;

	for (i = 0; i < array_size; i++) {
		cl_list_t *p_list = &count_sort[i];

		cl_list_remove_all(p_list);
		cl_list_destroy(p_list);
	}

	free(p_maxflow->count_sort);
	p_maxflow->count_sort = NULL;
}

__attribute__((unused))
static void flow_destroy_port_groups_array(IN maxflow_t *p_maxflow)
{
	int			i;
	int			j;
	maxflow_port_groups_t	*src_port_groups = p_maxflow->src_port_groups;
	maxflow_port_group_t	*port_groups;
	maxflow_port_group_t	*port_group;

	if (!src_port_groups)
		return;

	for (i = 0; i < p_maxflow->num_vertices; i++) {
		port_groups = src_port_groups[i].port_groups;
		if (!port_groups)
			continue;

		for (j = 0; j < p_maxflow->num_vertices; j++) {
			port_group = &port_groups[j];

			if (port_group->ports) {
				free(port_group->ports);
				port_group->num_ports = 0;
			}
		}
		free(port_groups);
	}

	destroy_count_sort_array(p_maxflow);

	free(src_port_groups);
	p_maxflow->src_port_groups = NULL;
}

static cl_status_t flow_init_count_sort_array(IN maxflow_t *p_maxflow)
{
	int			i;
	cl_status_t		status = CL_SUCCESS;

	p_maxflow->count_sort = (cl_list_t *)calloc(p_maxflow->count_sort_array_size, sizeof(cl_list_t));
	if (!p_maxflow->count_sort) {
		OSM_LOG(p_maxflow->p_log, OSM_LOG_ERROR, "ERR C901: allocation of count_sort array failed\n");
		status = CL_INSUFFICIENT_RESOURCES;
		goto Exit;
	}

	for (i = 0; i < p_maxflow->count_sort_array_size; i++) {
		cl_list_init(&p_maxflow->count_sort[i], 10);
	}

Exit:
	return status;
}

static int _flow_dfs_solve(IN osm_route_context_t *p_context,
			   IN osm_vertex_t *p_vertex,
			   IN int flow,
			   IN boolean_t is_src_vertex,
			   IN int *start_idx)
{
	maxflow_t		*p_maxflow = p_context->re_context;
	osm_graph_t		*p_graph = p_maxflow->p_graph;
	osm_vertex_t		*p_dst_vertex = p_maxflow->p_dst_vertex;
	uint32_t		*visited = p_maxflow->visited;
	boolean_t		*dirty = p_maxflow->dirty;
	int			min_flow, bottle_neck, remote_index;
	int			remaining_capacity;
	osm_edge_t		*p_edge = NULL;
	osm_vertex_t		*p_remote_vertex;
	uint16_t		min_hops_vertex;
	uint16_t		edge_hops;
	cl_list_iterator_t	next_list_iter = NULL;
	cl_list_iterator_t	cur_list_iter = NULL;
	cl_list_t		*p_list = NULL;
	cl_list_t		*p_tmp_list = NULL;
	int			j = *start_idx;
	int			count_sort_array_size = p_maxflow->count_sort_array_size;

	if (p_vertex == p_maxflow->p_dst_vertex) {
		return flow;
	}

	visited[p_vertex->index] = p_maxflow->visited_token;
	min_hops_vertex = osm_routing_get_least_hops(p_context, p_vertex, p_dst_vertex);

	if (is_src_vertex) {
		p_list = &p_maxflow->count_sort[j];
		next_list_iter = cl_list_head(p_list);
		cur_list_iter = next_list_iter;
	}

	for (int i = 0; i < p_vertex->num_edges; i++) {
		if (is_src_vertex) {
			p_edge = NULL;
			while (j < count_sort_array_size) {
				if (next_list_iter != cl_list_end(p_list)) {
					p_edge = (osm_edge_t *) cl_list_obj(next_list_iter);
					cur_list_iter = next_list_iter;
					next_list_iter = cl_list_next(cur_list_iter);
					break;
				} else {
					if ((j == *start_idx) && !cl_list_count(p_list))
						*start_idx = j + 1;

					j++;
					if (j < count_sort_array_size) {
						p_list = &p_maxflow->count_sort[j];
						next_list_iter = cl_list_head(p_list);
					}
				}
			}

			if (!p_edge) {
				OSM_LOG(p_maxflow->p_log, OSM_LOG_ERROR, "ERR C902: edge missing in count sort array\n");
				return 0;
			}
		} else {
			p_edge = &p_vertex->edges[i];
		}

		p_remote_vertex = osm_graph_index_to_vertex(p_graph, p_edge->remote_vertex_index);
		remote_index = p_remote_vertex->index;

		remaining_capacity = flow_get_remaining_capacity(p_maxflow, p_edge);
		edge_hops = osm_routing_get_hops(p_context, p_vertex, p_dst_vertex, p_edge);

		if (remaining_capacity && (visited[remote_index] != p_maxflow->visited_token) &&
		    (flow_is_residual(p_maxflow, p_edge) || (min_hops_vertex == edge_hops))) {
			min_flow = remaining_capacity > flow ? flow : remaining_capacity;
			bottle_neck = _flow_dfs_solve(p_context, p_remote_vertex, min_flow, FALSE, start_idx);

			if (bottle_neck > 0) {
				flow_augment(p_maxflow, p_edge, bottle_neck);
				dirty[remote_index] = TRUE;
				if (is_src_vertex) {
					cl_list_remove_item(p_list, cur_list_iter);
					p_tmp_list = &p_maxflow->count_sort[j + bottle_neck];
					cl_list_insert_tail(p_tmp_list, p_edge);
				}

				return bottle_neck;
			}
		}
	}

	return 0;
}


static int flow_dfs_solve(IN osm_route_context_t *p_context, IN osm_vertex_t *p_src_vertex, IN int *start_idx)
{
	maxflow_t		*p_maxflow = p_context->re_context;

	if (p_src_vertex == p_maxflow->p_dst_vertex)
		return 0;

	return _flow_dfs_solve(p_context, p_src_vertex, INT32_MAX, TRUE, start_idx);
}

static int maxflow_flow_max_flow_solve(IN osm_route_context_t *p_context,
				       IN osm_vertex_t *p_src_vertex,
				       IN osm_vertex_t *p_dst_vertex,
				       IN int *start_idx)
{
	int			f;
	int			max_flow = 0;
	uint16_t		hops_vertex;
	maxflow_t		*p_maxflow = p_context->re_context;

	hops_vertex = osm_routing_get_least_hops(p_context, p_src_vertex, p_dst_vertex);
	if (hops_vertex == OSM_NO_PATH) {
		return 0;
	}

	p_maxflow->visited_token = 2;
	p_maxflow->p_dst_vertex = p_dst_vertex;

	for (f = flow_dfs_solve(p_context, p_src_vertex, start_idx);
	     f != 0; f = flow_dfs_solve(p_context, p_src_vertex, start_idx)) {
		p_maxflow->visited_token++;
		max_flow += f;
	}

	return max_flow;
}



static int maxflow_flow_update_edge_directions(IN maxflow_t *p_maxflow, IN osm_vertex_t *p_src_vertex)
{
	int			result = 0;
	uint32_t		queue_in = 0;
	uint32_t		queue_out = 0;
	osm_graph_t		*p_graph = p_maxflow->p_graph;
	osm_vertex_t		**queue = NULL;
	uint32_t		*visited = p_maxflow->visited;
	osm_vertex_t		*p_vertex, *p_remote_vertex;
	osm_edge_t		*p_edge;
	int			visited_token;
	int			remote_index;
	maxflow_edge_t		*p_maxflow_edge;
	maxflow_vertex_t	*p_maxflow_vertex;
	maxflow_vertex_t	*p_maxflow_remote_vertex;

	queue = calloc(p_graph->num_vertices, sizeof(osm_vertex_t *));
	if (queue == NULL) {
		OSM_LOG(p_maxflow->p_log, OSM_LOG_ERROR, "ERR C903: allocation of queue failed\n");
		return -1;
	}

	p_maxflow->visited_token = 1;
	visited_token = p_maxflow->visited_token;
	memset(visited, 0, sizeof(uint32_t) * p_graph->num_vertices);
	foreach_graph_vertex(p_graph, p_vertex) {
		p_maxflow_vertex = &p_maxflow->vertices[p_vertex->index];
		p_maxflow_vertex->rank = 0;
	}

	queue[queue_in++] = p_src_vertex;
	visited[p_src_vertex->index] = p_maxflow->visited_token;

	while (queue_out < queue_in) {
		p_vertex = queue[queue_out++];
		p_maxflow_vertex = vertex_to_maxflow_vertex(p_maxflow, p_vertex);

		/* Add adjacent vertices to queue. */
		foreach_graph_vertex_edge(p_vertex, p_edge) {
			remote_index = p_edge->remote_vertex_index;
			p_remote_vertex = osm_graph_index_to_vertex(p_graph, remote_index);
			p_maxflow_remote_vertex = vertex_to_maxflow_vertex(p_maxflow, p_remote_vertex);

			if (visited[remote_index] != visited_token) {
				visited[remote_index] = visited_token;

				p_maxflow_remote_vertex = vertex_to_maxflow_vertex(p_maxflow, p_remote_vertex);
				p_maxflow_remote_vertex->rank = p_maxflow_vertex->rank + 1;
				queue[queue_in++] = p_remote_vertex;
			}

			p_maxflow_edge = edge_to_maxflow_edge(p_maxflow, p_edge);
			p_maxflow_edge->capacity = p_maxflow_remote_vertex->rank > p_maxflow_vertex->rank ? 1 : 0;
		}
	}

	free(queue);
	return result;
}

static void maxflow_flow_init_flow_and_visited(IN maxflow_t *p_maxflow)
{
	osm_graph_t		*p_graph = p_maxflow->p_graph;
	boolean_t		*dirty = p_maxflow->dirty;
	osm_vertex_t		*p_vertex;
	osm_edge_t		*p_edge;
	maxflow_edge_t		*p_maxflow_edge;
	maxflow_edge_t		*p_maxflow_residual_edge;
	int			index;

	memset(p_maxflow->visited, 0, sizeof(int) * p_maxflow->num_vertices);
	for (index = 0; index < p_maxflow->num_vertices; index++) {
		if (!dirty[index])
			continue;

		p_vertex = osm_graph_index_to_vertex(p_graph, index);
		dirty[index] = FALSE;
		foreach_graph_vertex_edge(p_vertex, p_edge) {
			p_maxflow_edge = &p_maxflow->edges[p_edge->index];
			p_maxflow_residual_edge = &p_maxflow->edges[p_maxflow_edge->residual_edge_index];

			p_maxflow_edge->flow = 0;
			p_maxflow_residual_edge->flow = 0;
		}
	}
}

static int cmpfunc (IN const void * lhs, IN const void * rhs) {
	vertex_index_to_port_t *p_left = (vertex_index_to_port_t *) lhs;
	vertex_index_to_port_t *p_right = (vertex_index_to_port_t *) rhs;

	if (p_left->remote_index == p_right->remote_index)
		return (p_left->port_num - p_right->port_num);

	return (p_left->remote_index - p_right->remote_index);
}

static int key_to_vertex_cmp_func(IN const void *lhs, IN const void* rhs)
{
	key_to_vertex_t *p_left = (key_to_vertex_t *)lhs;
	key_to_vertex_t *p_right = (key_to_vertex_t *)rhs;

	return strcmp(p_left->key, p_right->key);
}

static cl_status_t flow_init_port_groups_array(IN maxflow_t *p_maxflow)
{
	int			i;
	osm_graph_t		*p_graph;
	cl_status_t		status = CL_SUCCESS;

	p_graph = p_maxflow->p_graph;
	p_maxflow->src_port_groups = (maxflow_port_groups_t *)calloc(p_graph->num_vertices, sizeof(maxflow_port_groups_t));
	if (!p_maxflow->src_port_groups) {
		OSM_LOG(p_maxflow->p_log, OSM_LOG_ERROR, "ERR C904: allocation of src_port_groups failed\n");
		status = CL_INSUFFICIENT_RESOURCES;
		goto Exit;
	}

	for (i = 0; i < p_graph->num_vertices; i++) {
		p_maxflow->src_port_groups[i].port_groups = calloc(p_graph->num_vertices, sizeof(maxflow_port_group_t));
		if (!p_maxflow->src_port_groups[i].port_groups) {
			OSM_LOG(p_maxflow->p_log, OSM_LOG_ERROR, "ERR C906: allocation of port_groups failed\n");
			status = CL_INSUFFICIENT_RESOURCES;
			goto Exit;
		}
	}

Exit:
	return status;
}

static void init_representative_vertex_index(IN maxflow_t *p_maxflow, IN key_to_vertex_t *current_key,
					     IN key_to_vertex_t *current_representative_key,
					     IN key_to_vertex_t *current_second_representative_key)
{
	int			i;
	uint8_t			port_of_first_representative, port_of_second_representative, port_of_current_key;
	maxflow_port_groups_t	*src_port_groups = &p_maxflow->src_port_groups[current_key->vertex_index];

	src_port_groups->vertex_index_of_representative = current_representative_key->vertex_index;
	src_port_groups->vertex_index_of_second_representative = current_second_representative_key->vertex_index;

	for (i = 0; i < MAXFLOW_MAX_SUPPORTED_PORTS; i++) {
		port_of_first_representative = current_representative_key->ports[i];
		port_of_second_representative = current_second_representative_key->ports[i];
		port_of_current_key = current_key->ports[i];

		src_port_groups->representative_port_map[port_of_first_representative] = port_of_current_key;
		src_port_groups->second_representative_port_map[port_of_second_representative] = port_of_current_key;
	}
}

static void flow_choose_groups(IN maxflow_t *p_maxflow)
{
	osm_edge_t		*p_edge;
	int			i, n, j, num_edges, num_sources, num_in_current_group;
	int			key_index = 0;
	osm_vertex_t		*p_vertex;
	osm_graph_t		*p_graph;
	vertex_index_to_port_t	*remote_indexes;
	key_to_vertex_t		*current_key = NULL;
	key_to_vertex_t		*current_second_key = NULL;
	key_to_vertex_t		*keys_indexes = (key_to_vertex_t *) calloc(p_maxflow->count_sort_array_size, sizeof(key_to_vertex_t));

	if (!keys_indexes) {
		OSM_LOG(p_maxflow->p_log, OSM_LOG_ERROR, "ERR C907: allocation of key indexes failed\n");
		return;
	}

	p_graph = p_maxflow->p_graph;

	if (flow_init_port_groups_array(p_maxflow))
		goto Exit;

	foreach_graph_vertex(p_graph, p_vertex) {
		char *key = keys_indexes[key_index].key;
		maxflow_vertex_t *p_maxflow_vertex = &p_maxflow->vertices[p_vertex->index];

		if (!p_maxflow_vertex->calculate_flow)
			continue;

		keys_indexes[key_index].vertex_index = p_vertex->index;
		num_edges = p_vertex->num_edges;
		remote_indexes = (vertex_index_to_port_t *) malloc(sizeof(vertex_index_to_port_t) * num_edges);
		if (!remote_indexes) {
			OSM_LOG(p_maxflow->p_log, OSM_LOG_ERROR, "ERR C905: allocation of key indexes failed\n");
			goto Exit;
		}

		j = 0;
		foreach_graph_vertex_edge(p_vertex, p_edge) {
			remote_indexes[j].remote_index = p_edge->remote_vertex_index;
			remote_indexes[j].port_num = p_edge->port_num;
			j++;
		}
		qsort(remote_indexes, num_edges, sizeof(vertex_index_to_port_t), cmpfunc);

		n = 0;
		for (j = 0; j < num_edges; j++) {
			keys_indexes[key_index].ports[j] = remote_indexes[j].port_num;
			if (j == 0)
				n = sprintf(key, "%d", remote_indexes[j].remote_index);
			else
				n += snprintf(key + n, sizeof(keys_indexes[key_index].key) - n, ",%d", remote_indexes[j].remote_index);
		}

		key_index++;
		free(remote_indexes);
	}

	qsort(keys_indexes, p_maxflow->count_sort_array_size, sizeof(key_to_vertex_t), key_to_vertex_cmp_func);

	if (p_maxflow->count_sort_array_size < 2)
		goto Exit;

	current_key = &keys_indexes[0];
	current_second_key = &keys_indexes[0];
	init_representative_vertex_index(p_maxflow, current_key, current_key, current_second_key);

	num_sources = 1;
	num_in_current_group = 1;
	for (i = 1; i < p_maxflow->count_sort_array_size; i++) {
		if (strcmp(current_key->key, keys_indexes[i].key)) {
			current_key = &keys_indexes[i];
			current_second_key = &keys_indexes[i];
			num_sources++;

			num_in_current_group = 1;
		} else {
			num_in_current_group++;
			if (num_in_current_group == 2)
				current_second_key = &keys_indexes[i];
		}

		init_representative_vertex_index(p_maxflow, &keys_indexes[i], current_key, current_second_key);
	}

Exit:
	free(keys_indexes);
	return;
}

static void flow_run_src_to_dst_flow(IN osm_route_context_t *p_context,
				     IN osm_vertex_t *p_src_vertex,
				     IN osm_vertex_t *p_dst_vertex,
				     IN maxflow_port_groups_t *src_port_groups,
				     IN int *p_start_index)
{
	int			max_flow;
	int			num_ports;
	uint8_t			*ports;
	uint16_t		dst_index;
	osm_edge_t		*p_edge;
	maxflow_t		*p_maxflow;
	maxflow_edge_t		*p_maxflow_edge;

	p_maxflow = p_context->re_context;
	dst_index = p_dst_vertex->index;

	maxflow_vertex_t *p_maxflow_vertex = &p_maxflow->vertices[p_dst_vertex->index];
	if ((p_src_vertex == p_dst_vertex) || !p_maxflow_vertex->calculate_flow) {
		return;
	}

	maxflow_flow_init_flow_and_visited(p_maxflow);
	max_flow = maxflow_flow_max_flow_solve(p_context, p_src_vertex, p_dst_vertex, p_start_index);
	OSM_LOG(p_maxflow->p_log, OSM_LOG_DEBUG, "max_flow = %d\n", max_flow);

	ports = (uint8_t *)calloc(1, p_src_vertex->num_edges);
	if (ports != NULL) {
		num_ports = 0;
		foreach_graph_vertex_edge(p_src_vertex, p_edge) {
			p_maxflow_edge = edge_to_maxflow_edge(p_maxflow, p_edge);
			if (p_maxflow_edge->flow == 1) {
				ports[num_ports] = p_edge->port_num;
				num_ports++;
			}
		}

		src_port_groups->port_groups[dst_index].ports = ports;
		src_port_groups->port_groups[dst_index].num_ports = num_ports;
	} else {
		OSM_LOG(p_maxflow->p_log, OSM_LOG_ERROR, "ERR C908: allocation of ports failed\n");
	}
}

static void maxflow_free(IN maxflow_t *p_maxflow)
{
	if (p_maxflow != NULL) {
		flow_destroy_port_groups_array(p_maxflow);

		free(p_maxflow->visited);
		free(p_maxflow->dirty);
		free(p_maxflow->vertices);
		free(p_maxflow->edges);
		free(p_maxflow);
	}
}

/* ================================================================================================
 * Routing functions.
 */

static int maxflow_route_minhop(IN osm_route_context_t *p_context, IN osm_vertex_t *p_src_vertex, IN osm_vertex_t *p_dst_vertex)
{
	int			i;
	uint8_t			count;
	uint8_t			ports[IB_ENDP_MAX_NUMBER_PORTS];
	osm_route_t		*p_route;

	p_route = osm_vertex_to_vertex_route(p_context, p_src_vertex, p_dst_vertex);

	memset(ports, 0, sizeof(ports));
	count = osm_routing_get_minhop_port_nums(p_context, p_src_vertex, p_dst_vertex, ports);

	for (i = 0; i < count; i++) {
		osm_route_ports_set(p_route, ports[i]);
	}

	return 0;
}

static int maxflow_route_leaf_to_leaf(IN osm_route_context_t *p_context, IN osm_vertex_t *p_src_vertex, IN osm_vertex_t *p_dst_vertex)
{
	int			i;
	maxflow_t		*p_maxflow;
	osm_route_t		*p_route;

	int			vertex_index_of_representative;
	uint8_t			*port_map_of_representative;
	uint8_t			*ports;
	maxflow_port_group_t	*port_group_representative;
	maxflow_port_groups_t	*src_port_groups;
	maxflow_port_groups_t	*src_port_group;

	p_maxflow = p_context->re_context;
	src_port_groups = p_maxflow->src_port_groups;
	src_port_group = &src_port_groups[p_src_vertex->index];

	if (src_port_group->vertex_index_of_representative != p_dst_vertex->index) {
		vertex_index_of_representative = src_port_group->vertex_index_of_representative;
		port_map_of_representative = src_port_group->representative_port_map;
	} else {
		vertex_index_of_representative = src_port_group->vertex_index_of_second_representative;
		port_map_of_representative = src_port_group->second_representative_port_map;
	}

	port_group_representative = &src_port_groups[vertex_index_of_representative].port_groups[p_dst_vertex->index];
	ports = port_group_representative->ports;

	p_route = osm_vertex_to_vertex_route(p_context, p_src_vertex, p_dst_vertex);
	for (i = 0; i < port_group_representative->num_ports; i++) {
		osm_route_ports_set(p_route, port_map_of_representative[ports[i]]);
	}

	return 0;
}

static int maxflow_route(IN osm_route_context_t *p_context, IN osm_vertex_t *p_src_vertex)
{
	boolean_t		is_src_leaf;
	boolean_t		is_dst_leaf;
	osm_graph_t		*p_graph;
	osm_vertex_t		*p_dst_vertex;
	maxflow_vertex_t	*p_maxflow_src_vertex;
	maxflow_vertex_t	*p_maxflow_dst_vertex;
	maxflow_t		*p_maxflow;

	p_maxflow = p_context->re_context;
	p_graph = p_maxflow->p_graph;

	p_maxflow_src_vertex = vertex_to_maxflow_vertex(p_maxflow, p_src_vertex);
	is_src_leaf = p_maxflow_src_vertex->calculate_flow;

	foreach_graph_vertex(p_graph, p_dst_vertex) {
		p_maxflow_dst_vertex = vertex_to_maxflow_vertex(p_maxflow, p_dst_vertex);
		is_dst_leaf = p_maxflow_dst_vertex->calculate_flow;

		if (p_dst_vertex != p_src_vertex) {
			if (is_src_leaf && is_dst_leaf) {
				maxflow_route_leaf_to_leaf(p_context, p_src_vertex, p_dst_vertex);
			} else {
				maxflow_route_minhop(p_context, p_src_vertex, p_dst_vertex);
			}
		}
	}

	return 0;
}

static int maxflow_route_representative_vertex(IN osm_route_context_t *p_context, IN osm_vertex_t *p_vertex)
{
	int			dst_index;
	int			start_index;
	osm_graph_t		*p_graph;
	osm_edge_t		*p_edge;
	osm_vertex_t		*p_dst_vertex;
	maxflow_vertex_t	*p_maxflow_vertex;
	maxflow_port_groups_t	*src_port_groups;
	maxflow_t		*p_maxflow;

	p_maxflow = p_context->re_context;
	p_graph = p_maxflow->p_graph;
	start_index = 0;

	p_maxflow->p_mgr = p_context->p_mgr;
	p_maxflow->p_subn = p_maxflow->p_mgr->p_subn;
	p_maxflow->p_log = p_maxflow->p_mgr->p_log;

	/*
	 * Only max flow the representative vertices.
	 */
	p_maxflow_vertex = vertex_to_maxflow_vertex(p_maxflow, p_vertex);
	src_port_groups = &p_maxflow->src_port_groups[p_vertex->index];

	if (p_maxflow_vertex->calculate_flow) {
		if ((src_port_groups->vertex_index_of_representative == p_vertex->index) ||
		    (src_port_groups->vertex_index_of_second_representative == p_vertex->index)) {

			for (int i = 0; i < p_maxflow->count_sort_array_size; i++) {
				cl_list_remove_all(&p_maxflow->count_sort[i]);
			}

			maxflow_flow_update_edge_directions(p_maxflow, p_vertex);
			foreach_graph_vertex_edge(p_vertex, p_edge) {
				cl_list_insert_tail(&p_maxflow->count_sort[0], p_edge);
			}

			src_port_groups = &p_maxflow->src_port_groups[p_vertex->index];
			if (src_port_groups->vertex_index_of_representative == p_vertex->index) {
				foreach_graph_vertex(p_graph, p_dst_vertex) {
					flow_run_src_to_dst_flow(p_context, p_vertex, p_dst_vertex, src_port_groups, &start_index);
				}
			} else if (src_port_groups->vertex_index_of_second_representative == p_vertex->index) {
				dst_index = src_port_groups->vertex_index_of_representative;
				p_dst_vertex = osm_graph_index_to_vertex(p_graph, dst_index);
				flow_run_src_to_dst_flow(p_context, p_vertex, p_dst_vertex, src_port_groups, &start_index);
			}
		}
	}

	return 0;
}

/* ================================================================================================
 * Internal function.
 */
static int maxflow_init_ports(IN maxflow_t *p_maxflow)
{
	uint32_t		num_edges;
	boolean_t		hca_exists;
	boolean_t		is_leaf;
	osm_graph_t		*p_graph;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_remote_sw;
	osm_vertex_t		*p_vertex;
	osm_edge_t		*p_edge;
	maxflow_vertex_t	*p_maxflow_vertex;

	p_graph = p_maxflow->p_graph;

	foreach_graph_vertex(p_graph, p_vertex) {
		hca_exists = FALSE;
		is_leaf = TRUE;

		p_sw = osm_graph_vertex_to_switch(p_vertex);
		if (p_vertex->num_hcas) {
			hca_exists = TRUE;
		}
		num_edges = p_vertex->num_edges;

		foreach_graph_vertex_edge(p_vertex, p_edge) {
			p_remote_sw = osm_graph_index_to_switch(p_graph, p_edge->remote_vertex_index);
			if (p_sw->rank < p_remote_sw->rank) {
				is_leaf = FALSE;
				break;
			}
		}

		if (is_leaf && hca_exists && num_edges) {
			p_maxflow_vertex = &p_maxflow->vertices[p_vertex->index];

			if (!p_maxflow_vertex->calculate_flow) {
				p_maxflow_vertex->calculate_flow = TRUE;
				p_maxflow->count_sort_array_size++;
			}
		}
	}

	return 0;
}

/* ---------------------------------------------------------------------------------------------- */

static void maxflow_init_vertices(IN maxflow_t *p_maxflow)
{
}

/* ---------------------------------------------------------------------------------------------- */

static void maxflow_init_edges(IN maxflow_t *p_maxflow)
{
	osm_vertex_t		*p_vertex;
	osm_vertex_t		*p_remote_vertex;
	osm_edge_t		*p_edge;
	osm_edge_t		*p_remote_edge;
	osm_graph_t		*p_graph;
	maxflow_edge_t		*p_maxflow_edge;
	maxflow_edge_t		*p_maxflow_remote_edge;

	p_graph = p_maxflow->p_graph;

	foreach_graph_vertex(p_graph, p_vertex) {
		foreach_graph_vertex_edge(p_vertex, p_edge) {
			p_maxflow_edge = &p_maxflow->edges[p_edge->index];

			p_remote_vertex = osm_graph_index_to_vertex(p_graph, p_edge->remote_vertex_index);
			foreach_graph_vertex_edge(p_remote_vertex, p_remote_edge) {
				if ((p_remote_edge->vertex_index == p_edge->remote_vertex_index) &&
				    (p_remote_edge->port_num == p_edge->remote_port_num)) {

					p_maxflow_remote_edge = &p_maxflow->edges[p_remote_edge->index];

					p_maxflow_edge->residual_edge_index = p_remote_edge->index;
					p_maxflow_remote_edge->residual_edge_index = p_edge->index;
					break;
				}
			}
		}
	}
}

/* ---------------------------------------------------------------------------------------------- */

static int maxflow_init_hop_count(IN osm_graph_t *p_graph,
				IN osm_vertex_t *p_base_vertex,
				IN osm_vddq_t *p_vddq)
{
	int			i;
	int			v;
	int			dst;
	int			dir;
	int			base;
	int			hop_count;
	int			q_max;
	int			neighbor;
	int			neighbor_dir;
	uint8_t			*distance;
	uint8_t			*visited;
	uint8_t			*direction;
	uint16_t		*queue;
	uint16_t		lid_ho;
	uint16_t		base_lid_ho;
	osm_vertex_t		*p_vertex;
	osm_edge_t		*p_base_edge;
	osm_edge_t		*p_edge;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_dst_sw;
	osm_switch_t		*p_base_sw;
	osm_switch_t		*p_neighbor_sw;

	base = p_base_vertex->index;
	p_base_sw = osm_graph_index_to_switch(p_graph, base);
	dir = ROUTING_UP;

	base_lid_ho = osm_node_get_base_lid_ho(p_base_sw->p_node, 0);
	osm_switch_set_hops(p_base_sw, base_lid_ho, 0, 0);

	queue     = p_vddq->queue;
	visited   = p_vddq->visited;
	distance  = p_vddq->distance;
	direction = p_vddq->direction;

	foreach_graph_vertex_edge(p_base_vertex, p_base_edge) {
		osm_routing_vddq_reset(p_vddq);

		q_max = 0;
		queue[q_max++] = base;
		visited[base] = TRUE;
		distance[base] = 0;
		direction[base] = ROUTING_UP;

		neighbor = p_base_edge->remote_vertex_index;
		p_neighbor_sw = osm_graph_index_to_switch(p_graph, neighbor);
		neighbor_dir = (p_neighbor_sw->rank == p_base_sw->rank + 1) ? ROUTING_DOWN : ROUTING_UP;

		lid_ho = osm_node_get_base_lid_ho(p_neighbor_sw->p_node, 0);
		osm_switch_set_hops(p_base_sw, lid_ho, p_base_edge->port_num, p_base_edge->weight);

		queue[q_max++] = neighbor;
		visited[neighbor] = TRUE;
		distance[neighbor] = p_base_edge->weight;
		direction[neighbor] = neighbor_dir;

		dir = neighbor_dir;
		for (i = 1; i < q_max; i++) {
			v = queue[i];
			p_sw = osm_graph_index_to_switch(p_graph, v);
			p_vertex = osm_graph_index_to_vertex(p_graph, v);
			dir = direction[v];
			hop_count = distance[v];

			foreach_graph_vertex_edge(p_vertex, p_edge) {
				dst = p_edge->remote_vertex_index;
				p_dst_sw = osm_graph_index_to_switch(p_graph, dst);
				lid_ho = osm_node_get_base_lid_ho(p_dst_sw->p_node, 0);

				if ((dir == ROUTING_UP) && (p_dst_sw->rank == p_sw->rank - 1)) {
					dir = ROUTING_UP;
				} else if ((dir == ROUTING_UP) && (p_dst_sw->rank == p_sw->rank + 1)) {
					dir = ROUTING_DOWN;
				} else if ((dir == ROUTING_DOWN) && (p_dst_sw->rank == p_sw->rank + 1)) {
					dir = ROUTING_DOWN;
				} else {
					continue;
				}

				if (!visited[dst]) {
					visited[dst] = TRUE;
					distance[dst] = hop_count + p_edge->weight;
					direction[dst] = dir;
					queue[q_max++] = dst;

					lid_ho = osm_node_get_base_lid_ho(p_dst_sw->p_node, 0);
					osm_switch_set_hops(p_base_sw, lid_ho, p_base_edge->port_num, distance[dst]);
				}
			}
		}
	}

	return 0;
}

static int maxflow_init_hops(IN maxflow_t *p_maxflow)
{
	osm_vertex_t		*p_vertex;
	osm_vertex_t		*p_remote_vertex;
	osm_edge_t		*p_edge;
	osm_graph_t		*p_graph;
	osm_vddq_t		*p_vddq;

	p_graph = p_maxflow->p_graph;

	p_vddq = osm_routing_vddq_init(p_maxflow->p_mgr, p_graph);
	if (p_vddq == NULL) {
		return -1;
	}

	foreach_graph_vertex(p_graph, p_vertex) {
		foreach_graph_vertex(p_graph, p_remote_vertex) {
			uint16_t lid_ho = osm_node_get_base_lid_ho(p_remote_vertex->p_sw->p_node, 0);
			foreach_graph_vertex_edge(p_vertex, p_edge) {
				osm_switch_set_hops(p_vertex->p_sw, lid_ho, p_edge->port_num, OSM_NO_PATH);
			}
		}
	}


	foreach_graph_vertex(p_graph, p_vertex) {
		maxflow_init_hop_count(p_graph, p_vertex, p_vddq);
	}

	osm_routing_vddq_free(p_vddq);
	return 0;
}

/* ---------------------------------------------------------------------------------------------- */

static maxflow_t *maxflow_init(IN osm_route_context_t *p_context)
{
	uint32_t		*visited;
	boolean_t		*dirty;
	maxflow_t		*p_maxflow;
	maxflow_edge_t		*maxflow_edges;
	osm_ucast_mgr_t		*p_mgr;
	osm_graph_t		*p_graph;
	maxflow_vertex_t	*maxflow_vertices;
	osm_vertex_t		*p_vertex;

	p_mgr = p_context->p_mgr;
	p_graph = p_context->p_graph;

	p_maxflow         = calloc(1, sizeof(maxflow_t));
	visited           = calloc(p_graph->num_vertices, sizeof(uint32_t));
	dirty             = calloc(p_graph->num_vertices, sizeof(boolean_t));
	maxflow_vertices  = calloc(p_graph->num_vertices, sizeof(maxflow_vertex_t));
	maxflow_edges     = calloc(p_graph->num_edges, sizeof(maxflow_edge_t));

	if (osm_any_null(5, p_maxflow, visited, dirty, maxflow_vertices, maxflow_edges)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR C909: allocation of maxflow structs failed\n");
		free(p_maxflow);
		free(visited);
		free(dirty);
		free(maxflow_vertices);
		free(maxflow_edges);

		return NULL;
	}

	p_maxflow->num_vertices = p_graph->num_vertices;
	p_maxflow->p_mgr = p_mgr;
	p_maxflow->p_subn = p_mgr->p_subn;
	p_maxflow->p_log = p_mgr->p_log;
	p_maxflow->p_graph = p_graph;
	p_maxflow->vertices = maxflow_vertices;
	p_maxflow->edges = maxflow_edges;
	p_maxflow->visited = visited;
	p_maxflow->dirty = dirty;
	p_maxflow->count_sort = NULL;

	maxflow_init_vertices(p_maxflow);
	maxflow_init_edges(p_maxflow);
	maxflow_init_hops(p_maxflow);
	maxflow_init_ports(p_maxflow);
	flow_choose_groups(p_maxflow);

	foreach_graph_vertex(p_graph, p_vertex) {
		visited[p_vertex->index] = 1;
	}

	flow_init_count_sort_array(p_maxflow);

	/*
	 * Cache the MAXFLOW struct.
	 */
	p_context->re_context = p_maxflow;

	/*
	 * The representative vertices are routed so that the other vertices which
	 * are isomorphic can just be duplicates of them.
	 */
	foreach_graph_vertex(p_graph, p_vertex) {
		maxflow_route_representative_vertex(p_context, p_vertex);
	}

	return p_maxflow;
}

/* ================================================================================================
 * API function.
 */

int osm_routing_v2_maxflow_route(IN osm_route_context_t *p_context)
{
	int			status = -1;
	maxflow_t		*p_maxflow;

	if ((p_maxflow = maxflow_init(p_context)) != NULL) {
		p_context->sw_function = maxflow_route;
		p_context->sw_init_function = NULL;
		p_context->sw_fini_function = NULL;
		status = osm_routing_per_vertex(p_context);

		maxflow_free(p_maxflow);
	}

	return status;
}
