/*
 * Copyright (c) 2022-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2009 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_ucast_mgr_t.
 * This file implements the Unicast Manager object.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_debug.h>
#include <complib/cl_qlist.h>
#include <complib/cl_mpthreadpool.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_UCAST_MGR_V2_C
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_msgdef.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_routing_chain.h>
#include <opensm/osm_port_group.h>
#include <opensm/osm_advanced_routing.h>
#include <opensm/osm_db_pack.h>
#include <opensm/osm_router_mgr.h>
#include <opensm/osm_routing_v2.h>
#include <opensm/osm_unhealthy_ports.h>

/* Aport can't contain port 0 */
#define INVALID_APORT_PORT_NUM 0

#define OSM_UPDN_ROOTS_FILE OSM_DUMP_FILES_PREFIX "-updn-roots.dump"

extern int wait_for_pending_transactions(osm_opensm_t *p_osm);

/* ----------------------------------------------------------------------------------------------
 * Routing V2 section.
 */

/* ---------------------------------------------------------------------------------------------- */

static void ucast_mgr_clean_graph(IN osm_ucast_mgr_t *p_mgr, IN int plane)
{
	osm_graph_t *p_graph;

	OSM_LOG_ENTER(p_mgr->p_log);

	p_graph = osm_ucast_mgr_plane_to_graph(p_mgr, plane);
	if (!p_graph) {
		goto Exit;
	}

	osm_graph_destroy(p_graph);
	p_mgr->graphs[plane] = NULL;

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
}

void ucast_mgr_v2_cleanup(osm_ucast_mgr_t *p_mgr)
{
	int			plane;

	OSM_LOG_ENTER(p_mgr->p_log);

	foreach_ucast_plane(p_mgr, plane) {
		ucast_mgr_clean_graph(p_mgr, plane);
		osm_routing_v2_destroy(p_mgr, plane);
	}

	OSM_LOG_EXIT(p_mgr->p_log);
}

/* ---------------------------------------------------------------------------------------------- */

static uint8_t ucast_mgr_get_ar_port_mask(IN osm_route_context_t *v2_context,
					  IN osm_switch_t *p_src_sw,
					  IN osm_switch_t *p_dst_sw,
					  IN osm_ar_subgroup_t *p_ar_group)
{
	int			status = 0;
	uint64_t		ports[4];

	if (osm_routing_get_ports(v2_context, p_src_sw, p_dst_sw, ports)) {
		p_ar_group->mask64[0] = cl_hton64(ports[3]);
		p_ar_group->mask64[1] = cl_hton64(ports[2]);
		p_ar_group->mask64[2] = cl_hton64(ports[1]);
		p_ar_group->mask64[3] = cl_hton64(ports[0]);

		status = 1;
	}

	return status;
}

/* ------------------------------------------------------------------------- */

typedef struct routing_ar_parameters {
	boolean_t ar_enable;
	boolean_t fr_enable;
	boolean_t allow_hbf;
	uint8_t sub_grps_active;
} routing_ar_parameters_t;

/*
 * This function can be called by all routing engines with allow_hbf == TRUE,
 * DFP routing engine should be called with allow_hbf == FALSE
 */
static boolean_t ucast_mgr_set_required_ar_info_config(IN osm_ucast_mgr_t * p_mgr,
						       IN osm_switch_t *p_sw,
						       IN routing_ar_parameters_t *p_config_params)
{
	osm_advanced_routing_data_t *p_ar_data;
	osm_subn_t *p_subn = p_mgr->p_subn;
	uint16_t ar_sl_mask = p_subn->opt.ar_sl_mask;
	uint16_t hbf_sl_mask = p_subn->opt.hbf_sl_mask;
	boolean_t wait_for_responses = FALSE;
	uint64_t sw_guid;
	uint16_t sw_lid_ho;
	uint8_t ar_transport_mask = p_subn->opt.ar_transport_mask;
	boolean_t enable = FALSE;
	unsigned ar_configured = OSM_SW_AR_CONF_CONFIGURED;
	boolean_t fr_supported;

	OSM_LOG_ENTER(p_mgr->p_log);

	p_ar_data = &p_sw->ar_data;
	sw_guid = osm_node_get_node_guid_ho(p_sw->p_node);
	sw_lid_ho = osm_node_get_base_lid_ho(p_sw->p_node, 0);
	fr_supported = p_ar_data->ar_info.enable_features & IB_AR_INFO_FEATURE_MASK_FR_SUPPORTED;
	enable = p_config_params->ar_enable || (p_config_params->fr_enable && fr_supported);

	/*
	 * In case both AR and FR are disabled by configuration, mark switch as not configured.
	 * Switch that failed to get ar info can't be marked for ar_configured.
	 */
	if (!enable || !p_sw->ar_data.ar_info_updated) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Switch GUID 0x%" PRIx64 ", LID %u, neither AR nor FR is enabled, "
			"disable AR configuration\n",
			sw_guid, sw_lid_ho);
		ar_configured = OSM_SW_AR_CONF_DISABLED;
		goto Exit;
	}

	/* reset the calculated data */
	memset(&p_ar_data->req_ar_info, 0, sizeof(p_ar_data->req_ar_info));

	if (ar_sl_mask != OSM_ALL_SL_VL_AR_ENABLED) {
		if (!(p_ar_data->ar_info.enabled_and_cap_fields & IB_AR_INFO_CAP_MASK_BY_SL_CAP)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR C806: Switch GUID 0x%" PRIx64 ", LID %u, " \
				"do not support by_sl_cap.\n",
				sw_guid, sw_lid_ho);
			ar_configured = OSM_SW_AR_CONF_DISABLED;
			goto Exit;
		} else {
			p_ar_data->req_ar_info.enabled_and_cap_fields |= IB_AR_INFO_CAP_MASK_BY_SL_EN;
			p_ar_data->req_ar_info.enable_by_sl_mask = cl_hton16(ar_sl_mask);
		}
	}

	if (ar_transport_mask != OSM_AR_CFG_TRANSPORT_MASK_EN_ALL) {
		if (!(p_ar_data->ar_info.enabled_and_cap_fields & IB_AR_INFO_CAP_MASK_BY_TRANS_CAP)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR C805: Switch GUID 0x%" PRIx64 ", LID %u, " \
				"does not support by_transport_disable.\n",
				sw_guid, sw_lid_ho);
			ar_configured = OSM_SW_AR_CONF_DISABLED;
			goto Exit;
		} else {
			p_ar_data->req_ar_info.enabled_and_cap_fields |= IB_AR_INFO_CAP_MASK_BY_TRANS_CAP;
			/* the bits 4-7 of 'by transport disable' field are reserved */
			p_ar_data->req_ar_info.by_trns_disable = (~ar_transport_mask) & 0xF;
		}
	}

	/* Check if the switch supports dynamic capability calculation */
	if (!(p_ar_data->ar_info.enabled_and_cap_fields & IB_AR_INFO_CAP_MASK_DYN_CAP_CALC_SUP)) {
		ar_configured = OSM_SW_AR_CONF_DISABLED;
		goto Exit;
	}

	ib_ar_info_set_sub_grps_active(&p_ar_data->req_ar_info, p_config_params->sub_grps_active);

	/* here we can enable 'E' bit, that will be used either for AR or Fault Routing (FR) */
	p_ar_data->req_ar_info.enable_features |= IB_AR_INFO_FEATURE_MASK_AR_ENABLED;

	/* enable FR if needed and supported */
	if (p_config_params->fr_enable &&
	    (p_ar_data->ar_info.enable_features & IB_AR_INFO_FEATURE_MASK_FR_SUPPORTED)) {
		p_ar_data->req_ar_info.enable_features |= IB_AR_INFO_FEATURE_MASK_FR_ENABLED;
	}

	p_ar_data->req_ar_info.ageing_time_value = cl_hton32(0);

	/* no ARN, FRN anymore not for legacy switches nor for planarized switches */

	/*
	 * Don't support HBF for dfp. Can configure HBF data for DFP
	 * when the hbf sl mask is 0.
	 * DFP routing engine should be called with allow_hbf == FALSE
	 */
	if (p_config_params->allow_hbf &&
	    (osm_ar_info_is_hbf_support(&p_ar_data->ar_info) ||
	     (osm_subn_is_nvlink_enabled(p_mgr->p_subn) && osm_ar_info_is_nvl_hbf_supported(&p_ar_data->ar_info)))) {
		/* Set HBF enable, when HBF SL mask is not 0. The bit is 0 after reset of req_ar_info */
		if (hbf_sl_mask)
			p_ar_data->req_ar_info.enable_features1 |= IB_AR_INFO_HBF_EN_BY_SL_HBF_EN;

		p_ar_data->req_ar_info.enable_by_sl_mask_hbf = cl_hton16(hbf_sl_mask);
	}

	/*
	 * Check if capabilities according to required configuration
	 * are already known.
	 * If not, query the switch.
	 */
	if (!osm_ar_capabilities_search(p_ar_data, &p_ar_data->req_ar_info)) {
		if (osm_ar_query_ar_info(p_mgr->sm, p_sw, &p_ar_data->req_ar_info)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR C801: "
				"failed to query ARInfo on switch 0x%" PRIx64 "\n", sw_guid);
		} else
			wait_for_responses = TRUE;
	}

Exit:
	if (ar_configured == OSM_SW_AR_CONF_DISABLED && p_sw->ar_configured) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Clear old lft for Switch GUID 0x%" PRIx64 ", LID %u "
			"(because was configured in previous cycle)\n", sw_guid, sw_lid_ho);
	}

	p_sw->ar_configured = ar_configured;

	OSM_LOG_EXIT(p_mgr->p_log);
	return wait_for_responses;
}

static void ucast_mgr_validate_ar_config_capabilities(IN osm_ucast_mgr_t * p_mgr)
{
	osm_switch_t *p_sw;
	osm_advanced_routing_data_t *p_ar_data;
	const ib_ar_info_t *p_req_ar_info;

	foreach_switch(p_mgr->p_subn, p_sw) {
		/* this switch was already marked as not suitable for AR */
		if (p_sw->ar_configured == OSM_SW_AR_CONF_DISABLED) {
			continue;
		}

		p_ar_data = &p_sw->ar_data;
		p_req_ar_info = osm_ar_capabilities_search(p_ar_data, &p_ar_data->req_ar_info);

		if (!p_req_ar_info) {
			p_sw->ar_configured = OSM_SW_AR_CONF_DISABLED;
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
				"ERR C807: switch 0x%" PRIx64 " does not support"
				" required capabilities\n",
				osm_node_get_node_guid_ho(p_sw->p_node));
		} else {
			osm_ar_set_required_ar_info(p_ar_data, p_req_ar_info);
		}
	}
}

static void ucast_mgr_set_hbf_config(IN osm_ucast_mgr_t * p_mgr)
{
	OSM_LOG_ENTER(p_mgr->p_log);
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_switch_t *p_sw;
	osm_advanced_routing_data_t *p_ar_data;
	uint64_t sw_guid;
	uint16_t sw_lid_ho;
	boolean_t is_hbf_supported, is_hbf_enabled;
	uint32_t seed;
	uint64_t hash_fields;

	foreach_switch(p_subn, p_sw) {
		p_ar_data = &p_sw->ar_data;
		sw_guid = osm_node_get_node_guid_ho(p_sw->p_node);
		sw_lid_ho = osm_node_get_base_lid_ho(p_sw->p_node, 0);
		is_hbf_supported = osm_ar_info_is_hbf_support(&p_ar_data->ar_info);
		is_hbf_enabled = osm_ar_info_is_hbf_enabled(&p_ar_data->req_ar_info);

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Switch GUID 0x%" PRIx64 ", LID %u: ar_configured %u hbf is %s supported, hbf is %s enabled .\n",
			sw_guid, sw_lid_ho,
			p_sw->ar_configured,
			is_hbf_supported ? "" : "not ", is_hbf_enabled ? "" : "not ");

		if (p_sw->ar_configured == OSM_SW_AR_CONF_DISABLED)
			continue;

		/* NVLink HBF */
		if (osm_subn_is_nvlink_enabled(p_subn) &&
		    osm_ar_info_is_nvl_hbf_supported(&p_ar_data->ar_info) &&
		    osm_ar_info_is_hbf_enabled(&p_ar_data->req_ar_info)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Switch GUID 0x%" PRIx64 ", LID %u: Configures NVLink HBF\n",
				sw_guid, sw_lid_ho);

			p_sw->nvlink_sw_data.calc_nvl_hbf_cfg.hash_type =
			    (p_subn->opt.nvl_hbf_hash_type << IB_NVL_HBF_CONFIG_HASH_TYPE_SHIFT) &
			    IB_NVL_HBF_CONFIG_HASH_TYPE_MASK;

			p_sw->nvlink_sw_data.calc_nvl_hbf_cfg.packet_hash_bitmask =
			    p_subn->opt.nvl_hbf_packet_hash_bitmask;

			p_sw->nvlink_sw_data.calc_nvl_hbf_cfg.seed =
			    cl_hton32(p_subn->opt.nvl_hbf_seed);

			p_sw->nvlink_sw_data.calc_nvl_hbf_cfg.fields_enable =
			    cl_hton64(p_subn->opt.nvl_hbf_fields_enable);
		}

		/* set hbf configuration when the bit is enabled */
		if (is_hbf_supported && is_hbf_enabled) {
			/* configurable seed */
			p_sw->ar_data.req_hbf_config.hbf_seed_type =
				(p_subn->opt.hbf_hash_type << 4) |
				(0x3 & p_subn->opt.hbf_seed_type);

			/* take the lower 32 bits of the GUID */
			seed = p_subn->opt.hbf_seed == OSM_HBF_HASH_SEED_GUID ?
				sw_guid & 0xFFFFFFFF : p_subn->opt.hbf_seed;

			p_sw->ar_data.req_hbf_config.seed = cl_ntoh32(seed);

			/* currently only 32 LSB of 64 bits are defined */
			hash_fields = 0xFFFFFFFF & cl_ntoh64(p_subn->opt.hbf_hash_fields);

			/* disable specific hash fields bits if no capability */
			if (!osm_ar_info_is_hbf_bthdqp_hash_support(&p_ar_data->ar_info))
				hash_fields &= ~IB_HBF_CONFIG_HASH_FIELDS_BTH_DESTINATION_QP;
			if (!osm_ar_info_is_hbf_dceth_hash_support(&p_ar_data->ar_info)) {
				hash_fields &= ~IB_HBF_CONFIG_HASH_FIELDS_DCETH_V1_SRC_QP;
				hash_fields &= ~IB_HBF_CONFIG_HASH_FIELDS_DCETH_V1_ISID;
			}

			p_sw->ar_data.req_hbf_config.fields_enable = cl_hton64(hash_fields);
		}
	}

	OSM_LOG_EXIT(p_mgr->p_log);
}

/*
 * TODO: all routing engines should have sub routine set_required_ar_info that call this function.
 * DFP for example will call a different routine, as it has different configuration.
 */
static void ucast_mgr_set_required_config(IN osm_ucast_mgr_t * p_mgr)
{
	boolean_t wait_for_responses = FALSE;
	routing_ar_parameters_t ar_parameters = {0};
	osm_subn_t *p_subn = p_mgr->p_subn;
	ar_configuration_params_t *p_ar_conf = &p_subn->ar_conf_params;
	osm_switch_t *p_sw;

	/* the routing engine AR parameters, represent what routing engine configures */
	/* TODO may change in the future to get parameters by some API from routing engine in use */
	ar_parameters.allow_hbf = TRUE;
	/*
	 * sub_grps_active - the top sub group to use (num sub groups - 1), as set in ar_info.
	 * TODO add different configuration per different routing engine/ar configuration
	 */
	ar_parameters.sub_grps_active = 0;

	if (p_ar_conf->ar_mode == OSM_ADV_ROUTING_MODE_ENABLED ||
	    p_ar_conf->ar_mode == OSM_ADV_ROUTING_MODE_RN)
		ar_parameters.ar_enable = TRUE;

	if (p_ar_conf->shield_mode == OSM_ADV_ROUTING_MODE_ENABLED ||
	    p_ar_conf->shield_mode == OSM_ADV_ROUTING_MODE_RN)
		ar_parameters.fr_enable = TRUE;

	/* set required ar info and get dynamically calculated capabilities accordingly */
	foreach_switch(p_mgr->p_subn, p_sw) {
		if (ucast_mgr_set_required_ar_info_config(p_mgr, p_sw, &ar_parameters)) {
			wait_for_responses = TRUE;
		}
	}

	/* If we sent queries, wait for responses */
	if (wait_for_responses) {
		if (!p_subn->opt.enable_queries_during_routing)
			CL_PLOCK_RELEASE(p_mgr->sm->p_lock);

		CL_PLOCK_RELEASE(&p_subn->p_osm->routing_lock);
		(void)wait_for_pending_transactions(p_subn->p_osm);
		CL_PLOCK_EXCL_ACQUIRE(&p_subn->p_osm->routing_lock);

		if (!p_subn->opt.enable_queries_during_routing)
			CL_PLOCK_EXCL_ACQUIRE(p_mgr->sm->p_lock);
	}

	ucast_mgr_validate_ar_config_capabilities(p_mgr);

	/* set required hbf config (not for dfp) */
	if (ar_parameters.allow_hbf)
		ucast_mgr_set_hbf_config(p_mgr);
}

static int ucast_mgr_routing_preprocess(osm_ucast_mgr_t * p_mgr)
{
	uint8_t				plane;
	int				ret = 0;

	OSM_LOG_ENTER(p_mgr->p_log);

	/*
	 * Create the routing contexts.
	 */
	foreach_ucast_plane(p_mgr, plane) {
		/*
		 * In some scenarios we might get an empty graph in some plane.
		 * This does not mean that the routing engine should fail and
		 * fallover to a differnt one.
		 * Instead, preprocessing should continue, in order to set AR
		 * capabilities and so on.
		 */
		if (osm_graph_is_empty(osm_ucast_mgr_plane_to_graph(p_mgr, plane))) {
			continue;
		}

		p_mgr->v2_context[plane] = osm_routing_v2_create(p_mgr, plane);
		if (p_mgr->v2_context[plane] == NULL) {
			ucast_mgr_v2_cleanup(p_mgr);
			ret = -1;
			goto Exit;
		}
		p_mgr->v2_context[plane]->plane = plane;
	}

	/*
	 * Set the AR capabilities.
	 */
	ucast_mgr_set_required_config(p_mgr);

	/*
	 * Prepare Adaptive Routing (AR) tables
	 */
	osm_ar_prepare_tables(p_mgr->sm->p_subn, OSM_AR_INFO_REQ);

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}

/* ---------------------------------------------------------------------------------------------- */

static void ucast_add_routing_direct_hosts(IN osm_ucast_mgr_t *p_mgr, IN OUT osm_switch_t *p_sw)
{
	uint8_t num_ports, port_num, remote_port;
	osm_node_t *p_node, *p_remote_node;
	osm_physp_t *p_physp;

	p_node = p_sw->p_node;

	/* add static routing to CAs/RTRs connected directly to the switch */
	num_ports = osm_node_get_num_physp(p_node);
	for (port_num = 1 ; port_num < num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_node, port_num);

		if (!p_physp || !osm_physp_is_valid(p_physp))
			continue;

		p_remote_node = osm_node_get_remote_node(p_node, port_num, &remote_port);

		if (!p_remote_node || osm_node_get_type(p_remote_node) == IB_NODE_TYPE_SWITCH)
			continue;

		if (lookup_isolated_port(p_mgr->sm, p_node, port_num))
			continue;

		if (osm_planar_is_link_asymmetric(p_mgr->p_subn, p_physp, NULL))
			continue;

		osm_switch_set_routing(p_sw, p_remote_node, remote_port,
				       port_num, OSM_AR_IB_LID_STATE_STATIC, 0);
	}
}

static void ucast_mgr_clear_set_switch_to_itself(IN osm_ucast_mgr_t *p_mgr, IN OUT osm_switch_t *p_sw)
{
	OSM_LOG_ENTER(p_mgr->p_log);

	uint16_t sw_lid;

	osm_switch_clear_new_lfts(p_sw);
	osm_switch_clear_hops(p_sw);

	/* static routing on switch to itself via port 0 */
	osm_switch_set_routing(p_sw, p_sw->p_node, 0, 0, OSM_AR_IB_LID_STATE_STATIC, 0);
	sw_lid = osm_node_get_base_lid_ho(p_sw->p_node, 0);
	osm_switch_set_hops(p_sw, sw_lid, 0, 0);

	/* add static routing to any hosts directly connected to this switch */
	ucast_add_routing_direct_hosts(p_mgr, p_sw);

	OSM_LOG_EXIT(p_mgr->p_log);
}

static int
ucast_mgr_clear_set_switches_to_itself_par(void ** task, unsigned task_len, void * context,
					   unsigned thread_id)
{
	osm_ucast_mgr_t *p_mgr = (osm_ucast_mgr_t *)context;
	osm_switch_t *p_sw;
	osm_switch_t ** sw_arr = (osm_switch_t **)task;
	unsigned i = 0;

	for (i = 0; i < task_len; i++) {
		p_sw = sw_arr[i];
		ucast_mgr_clear_set_switch_to_itself(p_mgr, p_sw);
	}
	return 0;
}

static void ucast_mgr_clear_set_switches_to_themselves(IN osm_ucast_mgr_t *p_mgr)
{
	OSM_LOG_ENTER(p_mgr->p_log);

	cl_qmap_t *p_sw_guid_tbl;

	p_sw_guid_tbl = &p_mgr->p_subn->sw_guid_tbl;

	cl_mp_thread_pool_qmap_apply(&p_mgr->thread_pool, p_sw_guid_tbl,
				     CL_MP_THREAD_POOL_BATCH_PER_THREAD,
				     ucast_mgr_clear_set_switches_to_itself_par, p_mgr);

	OSM_LOG_EXIT(p_mgr->p_log);
}

/*
 * This function is called to fill in the hops of each switch (lid matrix).
 * The hops are currently used for multicast routing and dump to opensm-lid-matrix.dump.
 * Currently the plane symmetry enforcement is in place, and multicast tree is built
 * for plane #1.
 */
static void ucast_mgr_set_hops(IN osm_ucast_mgr_t *p_mgr)
{
	OSM_LOG_ENTER(p_mgr->p_log);

	uint8_t plane;
	osm_graph_t *p_graph;
	osm_route_context_t *v2_context;
	osm_vertex_t *p_src_vertex, *p_dest_vertex;
	osm_edge_t *p_edge;
	osm_switch_t *p_src_sw, *p_dest_sw;
	uint8_t hops;
	uint16_t dest_sw_lid;

	foreach_ucast_plane(p_mgr, plane) {
		p_graph = osm_ucast_mgr_plane_to_graph(p_mgr, plane);

		if (osm_graph_is_empty(p_graph))
			continue;

		v2_context = osm_ucast_mgr_plane_to_context(p_mgr, plane);

		foreach_graph_vertex(p_graph, p_src_vertex) {

			p_src_sw = osm_graph_vertex_to_switch(p_src_vertex);

			foreach_graph_vertex(p_graph, p_dest_vertex) {
				if (p_src_vertex == p_dest_vertex)
					continue;

				p_dest_sw = osm_graph_vertex_to_switch(p_dest_vertex);
				dest_sw_lid = osm_node_get_base_lid_ho(p_dest_sw->p_node, 0);

				foreach_graph_vertex_edge(p_src_vertex, p_edge) {
					hops = osm_routing_get_hops(v2_context, p_src_vertex, p_dest_vertex, p_edge);
					if (hops != OSM_NO_PATH) {
						osm_switch_set_hops(p_src_sw, dest_sw_lid, p_edge->port_num, hops);
					}
				}
			}
		}
	}

	OSM_LOG_EXIT(p_mgr->p_log);
}

static int ucast_mgr_routing_global_postprocess(osm_ucast_mgr_t *p_mgr)
{
	OSM_LOG_ENTER(p_mgr->p_log);

	/* set hops on all switches */
	ucast_mgr_set_hops(p_mgr);

	/* Complete routing inside non-prism systems (a.k.a. Black Mamba) via FNM links */
	osm_planar_add_fnm_routing(p_mgr->sm);

	/*
	 * Make sure there is routing inside system, before completing routes between
	 * different plane asics between systems.
	 */
	osm_planar_add_systems_routing(p_mgr->sm);

	OSM_LOG_EXIT(p_mgr->p_log);

	return 0;
}

static void ucast_mgr_post_routing_ar_group_assignment(IN osm_ucast_mgr_t *p_mgr)
{
	osm_subn_t *p_subn = p_mgr->p_subn;
	cl_list_iterator_t iter;
	osm_switch_t *p_sw;
	uint16_t sw_lid;
	/* currently support only plft 0 (ar_updn) */
	int plft_id = 0;

	OSM_LOG_ENTER(p_mgr->p_log);

	foreach_switch(p_subn, p_sw) {
		if (!p_sw->p_node)
			continue;

		/* skip switches that don't have CAs (except special ports as AN, RTR) */
		if (!osm_switch_has_cas(p_sw))
			continue;

		sw_lid = osm_node_get_base_lid_ho(p_sw->p_node, 0);

		osm_ucast_mgr_make_groupid(p_mgr, sw_lid, plft_id);
	}

	/*
	 * Make sure that switches that were added by routing engine, and might not have CAs,
	 * still get AR group id.
	 */
	foreach_leaf_iter(p_subn, iter) {
		p_sw = leaf_iter_to_switch(iter);

		if (!p_sw->p_node)
			continue;

		sw_lid = osm_node_get_base_lid_ho(p_sw->p_node, 0);

		osm_ucast_mgr_make_groupid(p_mgr, sw_lid, plft_id);
	}

	OSM_LOG_EXIT(p_mgr->p_log);
}

static int ucast_mgr_routing_plane_postprocess(osm_ucast_mgr_t *p_mgr, int plane)
{
	OSM_LOG_ENTER(p_mgr->p_log);

	osm_graph_t *p_graph;


	p_graph = osm_ucast_mgr_plane_to_graph(p_mgr, plane);
	if (!osm_graph_is_empty(p_graph)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Post-process plane %d routing\n", plane);
		osm_ucast_mgr_post_routing(p_mgr, plane);
	} else {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Post-process plane %d - skip routing, graph is empty\n", plane);
	}

	OSM_LOG_EXIT(p_mgr->p_log);

	return 0;
}

extern int osm_apply_func_on_port_group_nodes(char *pgrp_name, int pfn_func(void *, uint64_t , char *),
	                                      void *cxt, char *pi, void *p_osm, boolean_t goto_node);

static osm_switch_t *ucast_mgr_find_hub_switch_by_plane(osm_ucast_mgr_t * p_mgr, uint8_t plane);

int osm_ucast_mgr_route_v2(struct osm_routing_engine *r, osm_opensm_t * osm)
{
	int ret = 0;
	uint8_t plane;
	osm_ucast_mgr_t * p_mgr = &osm->sm.ucast_mgr;
	osm_route_context_t *p_context;

	OSM_LOG_ENTER(&osm->log);

	/*
	 * If the routing engine is version 2, then use the new sequence.
	 */
	if (ucast_mgr_routing_preprocess(p_mgr) != 0) {
		return -1;
	}

	foreach_ucast_plane(p_mgr, plane) {
		/* clean leafs found in previous plane, for each plane to work only with its own leafs */
		osm_subn_clear_leafs(p_mgr->p_subn);

		if (osm_graph_is_empty(osm_ucast_mgr_plane_to_graph(p_mgr, plane))) {
			OSM_LOG(&osm->log, OSM_LOG_DEBUG, "Plane %u graph is empty\n", plane);
			continue;
		}

		p_context = osm_ucast_mgr_plane_to_context(p_mgr, plane);
		ret = osm_routing_v2(p_context, r);
		if (ret != 0) {
			OSM_LOG(&osm->log, OSM_LOG_ERROR, "ERR C804: osm_routing_v2 failed, ret=%d\n", ret);
			return -1;
		}

		osm_switch_t *p_hub_sw = ucast_mgr_find_hub_switch_by_plane(p_mgr, plane);
		OSM_LOG(&osm->log, OSM_LOG_DEBUG,
			"hub switch %s per plane %u.\n",
			p_hub_sw ? osm_switch_name(p_hub_sw) : "None",
			plane);
	}

	if (osm_routing_engine_is_ar_updn(r) || osm_routing_engine_is_updn(r))
		osm_dump_roots(osm, OSM_UPDN_ROOTS_FILE);

	/*
	 * TODO: for support of running SM with old configurations that affect static port selection
	 * call osm_ucast_mgr_build_lft_tables(p_mgr) instead of just clearing,
	 * that uses osm_switch_recommend_path (implements legacy selection of static port)
	 */
	ucast_mgr_clear_set_switches_to_themselves(p_mgr);

	/* assign AR Group IDs to switches */
	ucast_mgr_post_routing_ar_group_assignment(p_mgr);

	int rev_plane; /* int for '--' not to overflow */
	/*
	 * HACK: Static routing from Crocodile to Crocodile destination should be on plane #1.
	 * This is to make sure that the multicast tree building process, which is done for plane #1,
	 * doesn't add switches that don't belong to plane #1 (Black Mamba asics of different planes).
	 * Crocodile switch appears in every plane.
	 * End routing with plane #1, to make sure that routing from Crocodile to Crocodile is through plane #1.
	 */
	for (rev_plane = ucast_end_plane(p_mgr); rev_plane >= ucast_start_plane(p_mgr); rev_plane--) {
		ucast_mgr_routing_plane_postprocess(p_mgr, rev_plane);
	}

	ucast_mgr_routing_global_postprocess(p_mgr);

	foreach_ucast_plane(p_mgr, plane) {
		osm_routing_v2_destroy(p_mgr, plane);
	}


	OSM_LOG_EXIT(&osm->log);

	return ret;
}

int osm_ucast_mgr_create_graphs(IN osm_ucast_mgr_t * p_mgr)
{
	int			ret = 0;
	uint8_t			plane;

	OSM_LOG_ENTER(p_mgr->p_log);

	p_mgr->num_planes = (p_mgr->p_subn->is_planarized) ? IB_MAX_NUM_OF_PLANES_XDR + 1 : 1;

	/*
	 * Delete the old graphs (if any).
	 */
	foreach_ucast_plane(p_mgr, plane) {
		ucast_mgr_clean_graph(p_mgr, plane);
	}

	/*
	 * Create the graphs.
	 */
	foreach_ucast_plane(p_mgr, plane) {
		if ((p_mgr->graphs[plane] = osm_graph_create(p_mgr, plane)) == NULL) {
			ucast_mgr_v2_cleanup(p_mgr);
			ret = -1;
			goto Exit;
		}
	}

	/*
	 * Setup the default hop counts
	 */
	foreach_ucast_plane(p_mgr, plane) {
		osm_routing_set_default_hopcounts(p_mgr, p_mgr->graphs[plane]);
	}

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}

osm_graph_t *osm_ucast_mgr_plane_to_graph(osm_ucast_mgr_t *p_mgr, uint8_t plane)
{
	CL_ASSERT(plane <= IB_MAX_NUM_OF_PLANES_XDR);
	return p_mgr->graphs[plane];
}

osm_route_context_t *osm_ucast_mgr_plane_to_context(osm_ucast_mgr_t *p_mgr, uint8_t plane)
{
	CL_ASSERT(plane <= IB_MAX_NUM_OF_PLANES_XDR);
	return p_mgr->v2_context[plane];
}

static int ucast_mgr_reachable_switch_count(IN osm_route_context_t *p_context,
					    IN osm_graph_t *p_graph,
					    IN osm_switch_t *p_hub_sw)
{
        int reachable_count;
	osm_vertex_t *p_hub_vertex;
	osm_vertex_t *p_dst_vertex;

	p_hub_vertex = osm_graph_switch_to_vertex(p_graph, p_hub_sw);

	reachable_count = 0;
	foreach_graph_vertex(p_graph, p_dst_vertex) {
		if (osm_routing_get_least_hops(p_context, p_hub_vertex, p_dst_vertex) != OSM_NO_PATH) {
			reachable_count++;
		}
	}

        return reachable_count;
}

static osm_switch_t *ucast_mgr_find_hub_switch_by_plane(osm_ucast_mgr_t * p_mgr, uint8_t plane)
{
	osm_subn_t *p_subn;
	osm_graph_t *p_graph;
	osm_switch_t *p_sw;
	osm_switch_t *p_hub_sw;
	osm_switch_t *p_optimal_sw;
	osm_vertex_t *p_vertex;
	int max_reachable;
	int reachable_count;
	ib_net64_t hub_switch_guid;
	osm_route_context_t *v2_context;

	OSM_LOG_ENTER(p_mgr->p_log);

	OSM_ASSERT(p_mgr);

	p_subn = p_mgr->p_subn;
	v2_context = osm_ucast_mgr_plane_to_context(p_mgr, plane);

	/*
	 * Verify that the topology supports hub switches.
	 */
	if (!cl_list_count(&p_subn->leaf_switches_list)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
			"There is no optional hub switch for routing missing paths\n");
		osm_subn_clear_hub_switch_by_plane(p_subn, plane);
		return NULL;
	}

	p_graph = osm_ucast_mgr_plane_to_graph(p_mgr, plane);
	if (osm_graph_is_empty(p_graph)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE, "No switches in the graph for plane %d\n", plane);
		osm_subn_clear_hub_switch_by_plane(p_subn, plane);
		return NULL;
	}

	/*
	 * Does the previous hub switch still work?
	 */
	p_sw = osm_subn_get_hub_switch_by_plane(p_subn, plane);
	if ((p_sw != NULL) && (p_sw->graph_index[plane] >= 0)) {
		reachable_count = ucast_mgr_reachable_switch_count(v2_context, p_graph, p_sw);
		if (reachable_count == p_graph->num_vertices) {
			p_hub_sw = p_sw;
			OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE, "Using previous hub switch %s (guid 0x%016lx)\n",
				osm_switch_name(p_hub_sw),
				osm_node_get_node_guid_ho(p_hub_sw->p_node));
			goto Exit;
		}
	}

	/*
	 * Find the leaf switch which can reach the most switches in the topology.
	 */
	max_reachable = 0;
	p_optimal_sw = NULL;
	p_hub_sw = NULL;

	foreach_graph_vertex(p_graph, p_vertex) {
		p_sw = osm_graph_vertex_to_switch(p_vertex);

		/* TODO: leaf status should be marked for graph vertex in routing data. */
		if (p_vertex->num_hcas == 0) {
			continue;
		}

		reachable_count = ucast_mgr_reachable_switch_count(v2_context, p_graph, p_sw);
		if (reachable_count == p_graph->num_vertices) {
			p_hub_sw = p_sw;
			break;
		}

		if (reachable_count > max_reachable) {
			max_reachable = reachable_count;
			p_optimal_sw = p_sw;
		}
	}

	if ((p_hub_sw == NULL) && (p_optimal_sw != NULL)) {
		p_hub_sw = p_optimal_sw;
		OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
			"Failed to find a perfect hub switch, using optimal hub switch %s (guid 0x%016lx).\n",
			osm_switch_name(p_hub_sw),
			osm_node_get_node_guid_ho(p_hub_sw->p_node));
	}

Exit:
	if (p_hub_sw != NULL) {
		hub_switch_guid = osm_node_get_node_guid(p_hub_sw->p_node);
		osm_subn_set_hub_switch_guid_by_plane(p_subn, hub_switch_guid, plane);
	} else {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Failed to find hub switch for plane %u\n", plane);
		osm_subn_clear_hub_switch_by_plane(p_subn, plane);
	}

	return p_hub_sw;
}

static void ucast_add_routing(IN osm_ucast_mgr_t *p_mgr,
			      IN osm_switch_t *p_sw,
			      IN osm_switch_t *p_dest_sw,
			      IN uint8_t static_port,
			      IN uint8_t state,
			      IN uint16_t ar_group_id)
{
	osm_node_t *p_dest_node, *p_remote_node;
	uint8_t num_ports, port_num, remote_port;
	osm_physp_t *p_physp;

	p_dest_node = p_dest_sw->p_node;

	/*
	 * Routing to switch lids should be static, to be able to manage broken
	 * Black Mamba destination systems, with asics without FNM links.
	 */
	osm_switch_set_routing(p_sw, p_dest_node, 0, static_port, OSM_AR_IB_LID_STATE_STATIC, 0);

	/* add routing to CAs/RTRs connected to the destination switch */
	num_ports = osm_node_get_num_physp(p_dest_node);

	for (port_num = 1 ; port_num < num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_dest_node, port_num);

		if (!p_physp || !osm_physp_is_valid(p_physp))
			continue;

		p_remote_node = osm_node_get_remote_node(p_dest_node, port_num, &remote_port);

		if (!p_remote_node || osm_node_get_type(p_remote_node) == IB_NODE_TYPE_SWITCH)
			continue;

		if (lookup_isolated_port(p_mgr->sm, p_dest_node, port_num))
			continue;

		if (osm_planar_is_link_asymmetric(p_mgr->p_subn, p_physp, NULL))
			continue;

		osm_switch_set_routing(p_sw, p_remote_node, remote_port,
				       static_port, state, ar_group_id);
	}
}

static uint8_t ucast_mgr_get_switch_missing_routes(IN osm_ucast_mgr_t *p_mgr,
						   IN OUT osm_switch_t *p_sw,
						   IN osm_switch_t *p_dest_sw,
						   uint8_t plane,
						   osm_ar_subgroup_t *p_subgroup)
{
	/*TODO change to use api to get hub switch by plane */
	if (!p_mgr->p_subn->hub_switch_guid[plane]) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"No hub switch for plane %u.\n", plane);
		return 0;
	}

	cl_map_item_t *p_item;
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_switch_t *p_hub_sw;
	osm_route_context_t *v2_context;
	osm_ar_subgroup_t routes_hub_dest;
	uint8_t routes = 0;

	ib_net64_t hub_guid = p_subn->hub_switch_guid[plane];
	p_item = cl_qmap_get(&p_subn->sw_guid_tbl, hub_guid);
	OSM_ASSERT(p_item != cl_qmap_end(&p_subn->sw_guid_tbl));
	p_hub_sw = (osm_switch_t *)p_item;

	v2_context = osm_ucast_mgr_plane_to_context(p_mgr, plane);

	/* make sure there is route from hub switch to destination */
	routes = ucast_mgr_get_ar_port_mask(v2_context, p_hub_sw, p_dest_sw, &routes_hub_dest);
	if (!routes) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"No routes on plane %u from sw %s to dest sw %s "
			"via hub switch %s (no routes from hub to destination).\n",
			plane, osm_switch_name(p_sw), osm_switch_name(p_dest_sw),
			osm_switch_name(p_hub_sw));
		return routes;
	}

	/* get routes from switch to the hub switch */
	routes = ucast_mgr_get_ar_port_mask(v2_context, p_sw, p_hub_sw, p_subgroup);
	if (!routes) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"No routes on plane %u from sw %s to dest sw %s "
			"via hub switch %s (no routes from src to hub).\n",
			plane, osm_switch_name(p_sw), osm_switch_name(p_dest_sw),
			osm_switch_name(p_hub_sw));
	}
	return routes;
}

static void
ucast_mgr_build_routing_on_switch(IN osm_ucast_mgr_t *p_mgr,
				  IN OUT osm_switch_t *p_sw,
				  uint8_t plane)
{
	osm_switch_t *p_dest_sw;
	uint16_t dest_sw_lid;
	osm_advanced_routing_data_t *p_ar_data = &p_sw->ar_data;
	osm_node_t *p_node, *p_dest_node;
	uint8_t routes, port_from_group;
	osm_ar_subgroup_t *p_subgroup, subgroup;
	boolean_t is_group_worthy, is_planarized_sw, is_planarized_dest_sw;
	boolean_t is_prism_sw, is_prism_dest_sw;
	uint8_t port_state;
	uint16_t group_top, group_cap;
	uint16_t ar_group_id;
	uint8_t plft_id = 0; /* Currently only plft 0 */
	osm_graph_t *p_graph;
	osm_vertex_t *p_vertex;
	osm_route_context_t *v2_context;

	if (p_sw->graph_index[plane] < 0) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"SW %s skip routing because it isn't in the plane %d\n",
			osm_switch_name(p_sw), plane);
		return;
		
	}

	p_graph = osm_ucast_mgr_plane_to_graph(p_mgr, plane);
	v2_context = osm_ucast_mgr_plane_to_context(p_mgr, plane);

	p_node = p_sw->p_node;
	group_top = 0;
	group_cap = ib_ar_info_get_group_cap_ho(&p_ar_data->ar_info);
	is_planarized_sw = osm_hierarchy_is_planarized(p_sw);
	is_prism_sw = osm_switch_is_prism_switch(p_sw, NULL);

	foreach_graph_vertex(p_graph, p_vertex) {
		p_dest_sw = osm_graph_vertex_to_switch(p_vertex);
		is_planarized_dest_sw = osm_hierarchy_is_planarized(p_dest_sw);

		if (p_sw == p_dest_sw)
			continue;

		is_prism_dest_sw = osm_switch_is_prism_switch(p_dest_sw, NULL);

		/*
		 * Skip routing between switches on the Black Mamba system with different planes.
		 * The source and the destination switch can belong to the same or different
		 * system.
		 */
		if (is_planarized_sw && !is_prism_sw &&
		    is_planarized_dest_sw && !is_prism_dest_sw &&
		    !(p_sw->plane_mask & p_dest_sw->plane_mask)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"SW %s skip routing to dest sw %s on %s system.\n",
				osm_switch_name(p_sw), osm_switch_name(p_dest_sw),
				p_sw->p_node->node_info.sys_guid ==
				p_dest_sw->p_node->node_info.sys_guid ? "same" : "different");
			continue;
		}

		p_dest_node = p_dest_sw->p_node;
		dest_sw_lid = osm_node_get_base_lid_ho(p_dest_node, 0);

		osm_ar_subgroup_clear(&subgroup);

		routes = ucast_mgr_get_ar_port_mask(v2_context, p_sw, p_dest_sw, &subgroup);
		port_from_group = osm_ar_get_first_port_set_in_subgroup(&subgroup);

		/* check if there is routing to this switch */
		if (routes == 0) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"No routes on plane %u from sw %s to dest sw %s.\n",
				plane, p_sw->p_node->print_desc, p_dest_sw->p_node->print_desc);

			/* try to complete routing via hub switch on that plane */
			routes = ucast_mgr_get_switch_missing_routes(p_mgr, p_sw, p_dest_sw,
								     plane, &subgroup);
			if (!routes) {
				continue;
			}
		}

		/*
		 * There are routes from switch to destination,
		 * take the first port from the group.
		 */
		port_from_group = osm_ar_get_first_port_set_in_subgroup(&subgroup);
		ar_group_id = 0;
		port_state = OSM_AR_IB_LID_STATE_STATIC;
		ar_group_id = osm_ucast_mgr_get_groupid(p_mgr, dest_sw_lid, plft_id);

		is_group_worthy = osm_switch_has_cas(p_dest_sw);

		if (ar_group_id != UCAST_MGR_INVALID_GROUP_ID) {
			if (ar_group_id < group_cap) {
				port_state = OSM_AR_IB_LID_STATE_FREE;
				/* set the group */
				p_subgroup = osm_ar_subgroup_get(p_ar_data, ar_group_id, 0);

				/* add the routes to already existing routes (if any) */
				osm_ar_subgroup_add_subgroup(p_subgroup, &subgroup);
				/* update top */
				group_top = ar_group_id > group_top ? ar_group_id : group_top;
			} /* TODO else ERR/message */
		} else {
			ar_group_id = 0;
		}

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "SW %s (0x%" PRIx64 ") "
			"dest sw %s (0x%" PRIx64 ") lid %u "
			"group_worthy %u port %u ar_group %u state %u "
			"group_cap %u top %u (new %u %u) routes %u (group %s empty) plane %u\n",
			p_node->print_desc, osm_node_get_node_guid_ho(p_node),
			p_dest_sw->p_node->print_desc,
			osm_node_get_node_guid_ho(p_dest_sw->p_node), dest_sw_lid,
			is_group_worthy, port_from_group, ar_group_id, port_state,
			group_cap, ib_ar_info_get_group_top_ho(&p_ar_data->req_ar_info),
			group_top, p_ar_data->group_top,
			routes, osm_ar_is_empty_group(&subgroup) ? "is" : "isn't", plane);

		ucast_add_routing(p_mgr, p_sw, p_dest_sw, port_from_group, port_state, ar_group_id);

	}

	/*
	 * Make sure to set group top on switch at least once.
	 * compare to p_ar_data->group_top, as req_ar_info may have top value as configured
	 * in the subnet.
	 */
	if (group_top > p_ar_data->group_top)
		osm_ar_set_required_group_top(p_ar_data, group_top);
}

static int ucast_mgr_build_routing_par(void ** task, unsigned task_len, void * context,
				   unsigned thread_id)
{
	osm_route_context_t *v2_context = (osm_route_context_t *)context;
	osm_ucast_mgr_t *p_mgr = v2_context->p_mgr;
	osm_switch_t *p_sw;
	osm_switch_t ** sw_arr = (osm_switch_t **)task;
	unsigned i = 0;

	for (i = 0; i < task_len; i++) {
		p_sw = sw_arr[i];
		/* TODO context might be routing engine */
		ucast_mgr_build_routing_on_switch(p_mgr, p_sw, v2_context->plane);
	}
	return 0;
}

void osm_ucast_mgr_post_routing(IN osm_ucast_mgr_t *p_mgr, uint8_t plane)
{
	cl_qmap_t *p_sw_guid_tbl;
	osm_route_context_t *v2_context;

	p_sw_guid_tbl = &p_mgr->p_subn->sw_guid_tbl;
        v2_context = osm_ucast_mgr_plane_to_context(p_mgr, plane);

	cl_mp_thread_pool_qmap_apply(&p_mgr->thread_pool, p_sw_guid_tbl,
				     CL_MP_THREAD_POOL_BATCH_PER_THREAD,
				     ucast_mgr_build_routing_par, v2_context);
}
